/*
 * This program tests for the correct behavior of 
 * gcc/GNU ld.  Things can go wrong for a variety of reasons:
 *
 * 1) if gcc does not use the GNU linker, this program will fail
 *    (probably it won't link)
 *
 * 2) gcc 2.8.[0,1] has a bug.  It allocates unitialized variables
 *    into the bss segment, even if they are declared nocommon and
 *    belonging to named sections.  The bug has been reported
 *    to the FSF and fixed as of today (6/25/1998), but a new
 *    gcc has not been released yet.
 *
 *    Apparently, egcs has the same problem.
 */

#include <stdio.h>

#ifdef CILK_SOLARIS
#define NEED_WORKING_GCC_GNU_LD
#endif

#ifdef CILK_LINUX
#define NEED_WORKING_GCC_GNU_LD
#endif

#ifdef CILK_DUNIX
#define NEED_WORKING_GCC_GNU_LD
#endif


/* do the test only if we are using processes */
#ifdef NEED_WORKING_GCC_GNU_LD

#define CILK_SEGMENT_ATTRIBUTE(s)  __attribute__((section(s), nocommon))
#define DECLARE_SHARED_VARIABLE(before,id,after) \
    before Cilk_##id##_shared after CILK_SEGMENT_ATTRIBUTE("Cilk_shared")
#define USE_SHARED(id) Cilk_##id##_shared

int normal_variable;
DECLARE_SHARED_VARIABLE(int, uninitialized, );
DECLARE_SHARED_VARIABLE(int, initialized, ) = 1;

/* symbols defined by the linker if everything is ok */
extern int _end_Cilk_shared;
extern int _begin_Cilk_shared;

int main()
{
     printf("This is a test program for GCC and the linker.\n");

     /* normal variable in shared segment */
     if (&_begin_Cilk_shared <= &normal_variable &&
	 &normal_variable <= &_end_Cilk_shared)
	  goto bug;

     /* uninitialized variable outside shared segment */
     if (!(&_begin_Cilk_shared <= &USE_SHARED(uninitialized) &&
	 &USE_SHARED(uninitialized) <= &_end_Cilk_shared))
	  goto bug;

     /* initialized variable outside shared segment */
     if (!(&_begin_Cilk_shared <= &USE_SHARED(initialized) &&
	 &USE_SHARED(initialized) <= &_end_Cilk_shared))
	  goto bug;

     printf("As fas as I can tell, everything works correctly.  Good luck!\n");
     return 0;

 bug:
     printf("Something is wrong with gcc and/or the linker.\n"
	    "Either the GNU linker is not installed correctly,\n"
	    "or your gcc has a bug.  gcc-2.8 and egcs are known\n"
	    "to have a bug that prevents Cilk from working.\n"
	    "Please read the Cilk FAQ for advice.\n");
     return 1;
}


#else /* NEED_WORKING_GCC_GNU_LD */
/* nothing to test */
int main()
{
     return 0;
}
#endif
