/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Adapted from Clean ANSI C Parser
 *  Eric A. Brewer, Michael D. Noakes
 *  
 *  $Log: basics.h,v $
 *  Revision 1.1.1.1  1998/03/19 04:43:21  astark
 *  Initial creation of umbrella/all-sets implementation
 *
 *  Revision 1.1.1.1.2.2  1997/02/21 01:26:21  fengmd
 *  Add new cilk2c pragma "ndwarn +/-" to turn on/off warning.
 *
 *  Revision 1.1.1.1.2.1  1997/02/14 23:14:41  fengmd
 *  The Nondeterminator comes into play.
 *
 *  Revision 1.2  1996/12/12 23:27:37  athena
 *  Cilk2c now understands `register int foo asm("r0")' and the like.
 *  Cilk2c understands #pragma segments [+-].
 *  Added flexibility to the rts to accomodate the above.
 *
 *  Revision 1.1  1996/12/10 17:10:47  athena
 *  Got rid of the Cilk_result hack --- required modification
 *  of cilk2c.  Consequently, added cilk2c to the Cilk-5 repository.
 *  Polished Keith's hacks (changed promote into promote_and_detach,
 *  and separated the victim part from the thief part).
 *
 *  Revision 1.3  1996/06/07 17:55:28  athena
 *  Fixed copyrights. I hate this.
 *
 *  Revision 1.2  1996/02/19 22:15:53  randall
 *  Moved _PUSH_FRAME/_POP_FRAME calls around, and reorganized rts macros
 *  a bit.  Removed #include inlining, and added EnterC and ExitC calls to
 *  prevent C headers from being parsed as Cilk.
 *
 *************************************************************************/
/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _BASICS_H_
#define _BASICS_H_


#ifndef __GNUC__
#define inline
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"

#ifdef NO_PROTOTYPES
extern int      fprintf(FILE*, const char *,...);
/* extern int      fputc(char, FILE*); */
extern int      fputs(const char*, FILE*);
extern FILE     *fopen(const char *, const char *);
extern int      fclose(FILE *);
extern int      printf(const char *,...);
extern int	sscanf(const char *, const char *,...);
extern int      _flsbuf(unsigned char, FILE*);
/* extern void     bcopy(char *, char *, int); */
extern int      toupper(int);
extern const void* memcpy(void *,const void*,size_t);
extern int      fflush(FILE *);
#endif


#define GLOBAL
#define PRIVATE static

/* NoReturn indicates that the function never returns (like "exit") */
#ifdef __GNUC__
#define NoReturn volatile void
#else
#define NoReturn void
#endif

/* for generic prototypes (use Generic *) */
typedef void Generic;
typedef void **GenericREF;  /* address of a Generic, for pass by reference */

typedef int Bool;
#define TRUE 1
#define FALSE 0


/* assertion checking */
#undef assert
#ifdef NDEBUG
 #define assert(x) ((void) 0)
#else
 #define assert(x)  ((x) ? (void)0 : (void)Fail(__FILE__, __LINE__, #x))
#endif

#define FAIL(str) \
  Fail(__FILE__, __LINE__, str)
#define UNREACHABLE FAIL("UNREACHABLE")



#define PLURAL(i) (((i) == 1) ? "" : "s")
#define TABSTRING "    "





#define MAX_FILES       256
#define MAX_FILENAME    200
#define MAX_SCOPE_DEPTH 100
#define MAX_OPERATORS   600




/* Basic Typedefs */

typedef struct nodeStruct Node;
typedef struct tablestruct SymbolTable;
typedef int OpType;


typedef struct coord {
    int line;
    short offset;
    short file;
} Coord;

GLOBAL Coord UnknownCoord;
#define IsUnknownCoord(coord)  ((coord).file == UnknownCoord.file)



#define PRINT_COORD(out, c) \
    { if (PrintLineOffset) \
      fprintf(out,"%s:%d:%d", FileNames[(c).file], (int)(c).line, \
              (int)(c).offset); \
    else fprintf(out, "%s:%d", FileNames[(c).file], (int) (c).line); }


#define REFERENCE(var)  ((var)->u.decl.references++)
#define VAR_NAME(var)   ((var)->u.decl.name)



/* Prototypes/definitions from other files */

#include "heap.h"
#include "list.h"
#include "symbol.h"




/* Basic Global Variables */

GLOBAL extern const float VersionNumber;     /* main.c */
GLOBAL extern const char *const VersionDate; /* main.c */
GLOBAL extern const char * Executable;       /* program name, main.c */
GLOBAL extern List *Program;                 /* main.c */
GLOBAL extern int WarningLevel;              /* main.c */
GLOBAL extern int Line, LineOffset, Errors, Warnings;    /* warning.c */
GLOBAL extern unsigned int CurrentFile;      /* c4.l: current file number */
GLOBAL extern char *Filename;                /* c4.l */
GLOBAL extern char *FileNames[MAX_FILES];    /* c4.l: file # to name mapping*/
GLOBAL extern const char *PhaseName;         /* main.c */
GLOBAL extern const char *InputFileRoot;     /* main.c */
GLOBAL extern Bool FileIncludedp[MAX_FILES]; /* c4.l */
GLOBAL extern Bool CurrentIncludedp;         /* c4.l */
GLOBAL extern Bool GenerateSegments;         /* main.c */
GLOBAL extern Bool MemoryChecking;           /* main.c */
GLOBAL extern Bool OptimizeChecking;         /* main.c */
GLOBAL extern Bool MemoryCheckedp;           /* main.c */
GLOBAL extern Bool MemoryWarningp;           /* main.c */

/* ANSI defines the following name spaces (K&R A11.1, pg 227): */
GLOBAL extern SymbolTable *Identifiers, *Labels, *Tags;

/* This table is used to ensure consistency across the translation unit */
GLOBAL extern SymbolTable *Externals;

/* Global Flags */
GLOBAL extern Bool DebugLex;                 /* main.c */
GLOBAL extern Bool PrintLineOffset;          /* main.c */
GLOBAL extern Bool IgnoreLineDirectives;     /* main.c */
GLOBAL extern Bool ANSIOnly;                 /* main.c */
GLOBAL extern Bool FormatReadably;           /* main.c */
GLOBAL extern Bool PrintLiveVars;            /* main.c */

/* Basic Global Procedures */

/* pretty-printing */
GLOBAL void DPN(Node *n);
GLOBAL void DPL(List *list);
GLOBAL void PrintNode(FILE *out, Node *node, int tab_depth);
GLOBAL int PrintConstant(FILE *out, Node *c, Bool with_name);
GLOBAL void PrintCRSpaces(FILE *out, int spaces);
GLOBAL void PrintSpaces(FILE *out, int spaces);
GLOBAL void PrintList(FILE *out, List *list, int tab_depth);
GLOBAL int PrintOp(FILE *out, OpType op);  /* operators.c */
GLOBAL void CharToText(char *array, unsigned char value);
GLOBAL inline int PrintChar(FILE *out, int c);    /* print.c */
GLOBAL int PrintString(FILE *out, const char *string); /* print.c */

/* warning.c */
GLOBAL NoReturn Fail(const char *file, int line, const char *msg);
GLOBAL void SyntaxError(const char *fmt, ...);
GLOBAL void Warning(int level, const char *fmt, ...);
GLOBAL void SyntaxErrorCoord(Coord c, const char *fmt, ...);
GLOBAL void WarningCoord(int level, Coord c, const char *fmt, ...);

/* parsing phase */
GLOBAL int yyparse(void), yylex(void);
GLOBAL char *UniqueString(const char *string);   /* strings.c */
GLOBAL void SetFile(const char *filename, int line);       /* c4.l */

/* verification */
GLOBAL void VerifyParse(List *program);          /* verify-parse.c */

/* semantic check -- sem-check.c */
GLOBAL List *SemanticCheckProgram(List *program);
GLOBAL Node *SemCheckNode(Node *node);
GLOBAL List *SemCheckList(List *list);

/* transform phase -- transform.c */
GLOBAL List *TransformProgram(List *program);
GLOBAL List *LinkProgram(List *program, const char *filename_roots[], int n);

/* output phase -- output.c */
GLOBAL void OutputProgram(FILE *out, List *program);


#endif  /* ifndef _BASICS_H_ */


