/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Rob Miller
 *  
 *  $Log: transform.c,v $
 *  Revision 1.4  1998/07/15 09:31:34  fengmd
 *  Support user-inlet and library, merge changes from the main branch.
 *
 *  Revision 1.3  1998/06/19 05:42:33  fengmd
 *  Check node type before testing _LOCK_CHECK.
 *
 *  Revision 1.2  1998/06/11 13:18:55  fengmd
 *  Report without-lock name along with the without-lock address.
 *
 *  Revision 1.1.1.1  1998/03/19 04:43:23  astark
 *  Initial creation of umbrella/all-sets implementation
 *
 *  Revision 1.1.1.1.2.12  1997/08/29 07:38:19  fengmd
 *  Merge malloc/free checker, race report hashing, garbage collection.
 *
 *  Revision 1.1.1.1.2.11  1997/07/24 02:06:21  randall
 *  Fixed bug in return transform.
 *
 *  Revision 1.1.1.1.2.10  1997/06/16 19:51:45  fengmd
 *  Fix problem of the non-nondeterminator version --- useless frame variable.
 *
 *  Revision 1.1.1.1.2.9  1997/06/16 15:59:10  fengmd
 *  Remerge from the base for Version 5.0.1.
 *
 *  Revision 1.3  1997/03/04 02:07:54  randall
 *  Fixed globalizing of internal static variables.
 *
 *  Revision 1.2  1997/03/04 01:32:51  randall
 *  Get the right attributes on globalized static variables.
 *
 *  Revision 1.1.1.1  1997/02/14 23:06:57  strumpen
 *  Create Cilk Repository
 *
 *  Revision 1.1.1.1  1997/02/14 22:04:05  strumpen
 *  Create cilk repository
 *
 *  Revision 1.7  1997/02/12 19:04:25  randall
 *  Put SYNCHED back in.
 *
 *  Revision 1.6  1997/02/11 00:33:01  athena
 *  Fixed ancient hack in the rts (StackFrameT aliased to StackFrame, etc.)
 *  Changed cilk2c to do The Right Thing.  Adjusted examples.
 *
 *  Revision 1.5  1997/02/07 21:02:46  athena
 *  Fixed behaviour of cilk2c on shared/private variables
 *
 *  Revision 1.4  1997/02/02 22:24:09  randall
 *  Added code to allow spawning via a function pointer.
 *  Made cilk2c accept multi-dimensional incomplete arrays (like int a[][])
 *  in declarations.
 *  Allowed type match between elaborated and unelaborated Sdcls.
 *  Added some BUGS.
 *
 *  Revision 1.3  1997/01/23 02:24:54  bsong
 *  Matteo hacked the compiler to support abort.
 *
 *  Revision 1.2  1996/12/12 23:27:46  athena
 *  Cilk2c now understands `register int foo asm("r0")' and the like.
 *  Cilk2c understands #pragma segments [+-].
 *  Added flexibility to the rts to accomodate the above.
 *
 *  Revision 1.1  1996/12/10 17:11:31  athena
 *  Got rid of the Cilk_result hack --- required modification
 *  of cilk2c.  Consequently, added cilk2c to the Cilk-5 repository.
 *  Polished Keith's hacks (changed promote into promote_and_detach,
 *  and separated the victim part from the thief part).
 *
 *  Revision 1.57  1996/12/09 21:49:54  randall
 *  Added size of return value to signature.
 *
 *  Revision 1.56  1996/11/05 16:15:49  athena
 *  Small change to simplify the Milk rts (compatible with the Cilk rts)
 *
 *  Revision 1.55  1996/09/17 18:10:07  fengmd
 *  restore the change to AddFastStuff and AddSlowStuff.
 *
 *  Revision 1.54  1996/09/16 03:33:29  fengmd
 *  Modify AddFastStuff and AddSlowStuff to remove the additional block structure.
 *
 *  Revision 1.53  1996/09/05 14:51:06  athena
 *  Rationalization of timers
 *
 *  Revision 1.52  1996/08/29 19:55:57  athena
 *  Fixed TIMING/CP confusion
 *
 *  Revision 1.51  1996/08/28 23:11:59  athena
 *  Fixed missing-return bug
 *
 *  Revision 1.50  1996/08/27 19:44:34  randall
 *  Fixed default shared bug.
 *
 *  Revision 1.49  1996/08/27 19:38:02  randall
 *  Made variables shared by default.  Added some checking of private
 *  and shared declarations.
 *
 *  Revision 1.48  1996/08/24 21:18:18  randall
 *  Added support for shared and private declarations of variables.
 *
 *  Revision 1.47  1996/08/16 23:14:24  randall
 *  Made implicit inlet line numbers unknown.
 *
 *  Revision 1.46  1996/08/16 23:10:42  randall
 *  Redid how implicit inlets are done.  They are now a type of spawn
 *  statement.  Reorganized the translations of spawn statements a little.
 *
 *  Revision 1.45  1996/06/06 03:52:30  fengmd
 *  Output IN_FAST_INLET, IN_SLOW_INLET and IN_STANDALONE_INLET.
 *
 *  Revision 1.44  1996/06/06 01:02:19  fengmd
 *  Output preprocessing directives for CILK_WHERE_AM_I.
 *
 *  Revision 1.43  1996/06/03 03:28:25  fengmd
 *  Minor change for implicit inlet.
 *
 *  Revision 1.42  1996/05/29 17:04:45  randall
 *  made signature static, fixed return transformation.
 *
 *  Revision 1.41  1996/05/28 21:23:53  athena
 *  Removed that ugly signature table and associated -link option.
 *
 *  Revision 1.40  1996/05/21 13:06:30  fengmd
 *  Fix bug on implicit inlet to accumulate structure members, check type mismatch.
 *
 *  Revision 1.39  1996/05/16 14:52:26  athena
 *  Added critical path measurements. I am not sure I got it right,
 *  Keith please check.
 *
 *  Revision 1.38  1996/05/15 21:18:41  athena
 *  Added rts hooks to cilk2c (for timing, etc.)
 *
 *  Revision 1.37  1996/05/13 21:04:48  randall
 *  Comment update.
 *
 *  Revision 1.36  1996/05/09 17:48:34  randall
 *  Got const to compile right.
 *
 *  Revision 1.35  1996/05/09 17:32:14  randall
 *  Implemented static variable globalizing.
 *
 *
 ********************************************************************
 *
 * Copyright (c) 1994 MIT Laboratory for Computer Science
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MIT LABORATORY FOR COMPUTER SCIENCE BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the MIT Laboratory for
 * Computer Science shall not be used in advertising or otherwise to
 * promote the sale, use or other dealings in this Software without prior
 * written authorization from the MIT Laboratory for Computer Science.
 * 
 *************************************************************************/
#pragma ident "$Id: transform.c,v 1.4 1998/07/15 09:31:34 fengmd Exp $ Copyright 1994 Massachusetts Institute of Technology"

#include <ctype.h>
#include "ast.h"
GLOBAL SUEtype *make_SUE(NodeType typ, const char *name, List *fields);

/* CilkPrefix: prefix for all preprocessor-generated symbols */
GLOBAL const char *CilkPrefix = "_";

PRIVATE char CilkNameBuf[100];
PRIVATE const char *ApplyPrefix(const char *root)
{
     sprintf(CilkNameBuf, "%s%s", CilkPrefix, root);
     return UniqueString(CilkNameBuf);
}

PRIVATE const char *MakeSyncName(int n)
{
     sprintf(CilkNameBuf, "%s%s%d", CilkPrefix, "sync", n);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeFrameTypeName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_frame", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeSlowProcName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_slow", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}
PRIVATE const char *MakeScopeName(int n)
{
     sprintf(CilkNameBuf, "scope%d", n);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeLinkName(const char *proc)
{
     sprintf(CilkNameBuf, "%s%s_sig", CilkPrefix, proc);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeTempName(void)
{
     static int temp_count = 0;
     sprintf(CilkNameBuf, "%s%s%d", CilkPrefix, "temp", temp_count++);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeInletName(const char *proc, const char *inlet)
{
     sprintf(CilkNameBuf, "%s%s_%s_inlet", CilkPrefix, proc, inlet);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeImplicitInletName(const char *proc, int entry)
{
     sprintf(CilkNameBuf, "%s%s_%d_inlet", CilkPrefix, proc, entry);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *MakeInletArgsTypeName(const char *inlet)
{
     sprintf(CilkNameBuf, "%s_args", inlet);
     return (UniqueString(CilkNameBuf));
}

PRIVATE const char *FrameName, *HeapFrameName, *HeaderName, *DummyReturnName;
PRIVATE const char *EntryName, *IndexName, *SizeName, *FunctionName;
PRIVATE const char *SynchronizeName, *SynchedName;
PRIVATE const char *GenericFrameTypeName;
PRIVATE const char *InitFrameName, *PushFrameName;
PRIVATE const char *XPopFrameResultName, *XPopFrameNoResultName;
PRIVATE const char *SetResultName, *SetNoResultName;
PRIVATE const char *InletArgsName, *InletResultName, *ResumeName;
PRIVATE const char *AbortSlowName, *AbortStandaloneName;
PRIVATE const char *DoubleReturnFastName, *DoubleReturnSlowName;
PRIVATE const char *DoubleReturnStandaloneName;
PRIVATE const char *StartThreadFastName, *StartThreadSlowName, 
     *BeforeSpawnFastName, *BeforeSpawnSlowName, *AfterSpawnFastName, 
     *AfterSpawnSlowName, *AfterSyncSlowName, *BeforeSyncSlowName, 
     *AtSyncFastName, *BeforeReturnFastName, *BeforeReturnSlowName,
     *BeforeReturnInletName, *AtThreadBoundarySlowName;
PRIVATE const char *AccessReadProfileName, *AccessWriteProfileName;
PRIVATE const char *AccessReadProfileName4, *AccessWriteProfileName4;
PRIVATE const char *AccessReadProfileName8, *AccessWriteProfileName8;
PRIVATE const char *AccessReadProfileAddrName, *AccessWriteProfileAddrName;
PRIVATE const char *AccessReadProfileAddrName4, *AccessWriteProfileAddrName4;
PRIVATE const char *AccessReadProfileAddrName8, *AccessWriteProfileAddrName8;
PRIVATE const char *SpawnVarProfileName;

PRIVATE void InitCilkTransform(void)
{
     /* frame names */
     FrameName = ApplyPrefix("frame");
     HeapFrameName = ApplyPrefix("heap_frame");

     /* parts-of-frame names */
     HeaderName = UniqueString("header");
     DummyReturnName = UniqueString("dummy_return");

     /* parts of header */
     EntryName = UniqueString("entry");
     IndexName = UniqueString("index");
     SizeName = UniqueString("size");
     FunctionName = UniqueString("fn");

     /* parts of RTS */
     GenericFrameTypeName = UniqueString("StackFrame");

     /* RTS macros */
     InitFrameName = UniqueString("_INIT_FRAME");
     PushFrameName = UniqueString("_PUSH_FRAME");
     XPopFrameResultName = UniqueString("_XPOP_FRAME_RESULT");
     XPopFrameNoResultName = UniqueString("_XPOP_FRAME_NORESULT");
     SetResultName = UniqueString("_SET_RESULT");
     SetNoResultName = UniqueString("_SET_NORESULT");
     SynchronizeName = UniqueString("_SYNC");
     SynchedName = UniqueString("_SYNCHED");
     AbortSlowName = UniqueString("_ABORT_SLOW");
     AbortStandaloneName = UniqueString("_ABORT_STANDALONE");
     ResumeName = UniqueString("_RESUME");
     DoubleReturnFastName = UniqueString("_DOUBLE_RETURN_FAST");
     DoubleReturnSlowName = UniqueString("_DOUBLE_RETURN_SLOW");
     DoubleReturnStandaloneName = UniqueString("_DOUBLE_RETURN_STANDALONE");
     StartThreadFastName = UniqueString("_START_THREAD_FAST");
     StartThreadSlowName = UniqueString("_START_THREAD_SLOW");
     BeforeSpawnFastName = UniqueString("_BEFORE_SPAWN_FAST");
     BeforeSpawnSlowName = UniqueString("_BEFORE_SPAWN_SLOW");
     AfterSpawnFastName = UniqueString("_AFTER_SPAWN_FAST");
     AfterSpawnSlowName = UniqueString("_AFTER_SPAWN_SLOW");
     AtSyncFastName = UniqueString("_AT_SYNC_FAST");
     AtThreadBoundarySlowName = UniqueString("_AT_THREAD_BOUNDARY_SLOW");
     BeforeSyncSlowName = UniqueString("_BEFORE_SYNC_SLOW");
     AfterSyncSlowName = UniqueString("_AFTER_SYNC_SLOW");
     BeforeReturnFastName = UniqueString("_BEFORE_RETURN_FAST");
     BeforeReturnSlowName = UniqueString("_BEFORE_RETURN_SLOW");
     BeforeReturnInletName = UniqueString("_BEFORE_RETURN_INLET");

     InletArgsName = ApplyPrefix("inletargs");
     InletResultName = ApplyPrefix("inletresult");

     AccessReadProfileName = "_READ_CHECK_VALUE";
     AccessWriteProfileName = "_WRITE_CHECK_VALUE";
     AccessReadProfileName4 = "_READ_CHECK_VALUE4";
     AccessWriteProfileName4 = "_WRITE_CHECK_VALUE4";
     AccessReadProfileName8 = "_READ_CHECK_VALUE8";
     AccessWriteProfileName8 = "_WRITE_CHECK_VALUE8";
     AccessReadProfileAddrName = "_READ_CHECK";
     AccessWriteProfileAddrName = "_WRITE_CHECK";
     AccessReadProfileAddrName4 = "_READ_CHECK4";
     AccessWriteProfileAddrName4 = "_WRITE_CHECK4";
     AccessReadProfileAddrName8 = "_READ_CHECK8";
     AccessWriteProfileAddrName8 = "_WRITE_CHECK8";
     SpawnVarProfileName = "_SPAWNVAR_CHECK";
}

/* true iff transforming a cilk procedure */
PRIVATE int in_procedure = 0;
PRIVATE int in_inlet = 0;
PRIVATE int on_abort_defined = 0;	/* true if on_abort is defined in
					 * * a cilk procedure */
PRIVATE Node *curr_proc;
PRIVATE const char *proc_name;
PRIVATE int sync_count;
typedef enum {
     Fast,			/* generating fast version of procedure */
     Slow,			/* generating slow version of procedure */
     Link,			/* generating linkage information       */
     InletProc			/* generating stand-alone inlet         */
} VersionType;
VersionType version;
List *link_initializers;
List *standalone_inlets;

/* for Nondeterminator */
typedef enum {
     Read,
     Write,
} AccessProfileType;

#define CHECK_READ(node, type) \
  AccessProfile(node, type, Read, 0, 0, 1, '=')
#define CHECK_READ_ADDR(node, type) \
  AccessProfile(node, type, Read, 1, 0, 1, '=')
#define CHECK_READ_MULT(node, type, multiple) \
  AccessProfile(node, type, Read, 0, 0, multiple, '=')
#define CHECK_READ_OPT(node, type) \
  AccessProfile(node, type, Read, 0, 1, 1, '=')
#define CHECK_READ_OPT_MULT(node, type, multiple) \
  AccessProfile(node, type, Read, 0, 1, multiple, '=')

#define CHECK_WRITE(node, type, operator) \
  AccessProfile(node, type, Write, 0, 0, 1, operator)
#define CHECK_WRITE_ADDR(node, type, operator) \
  AccessProfile(node, type, Write, 1, 0, 1, operator)
#define CHECK_WRITE_MULT(node, type, multiple, operator) \
  AccessProfile(node, type, Write, 0, 0, multiple, operator)
#define CHECK_WRITE_OPT(node, type, operator) \
  AccessProfile(node, type, Write, 0, 1, 1, operator)
#define CHECK_WRITE_OPT_MULT(node, type, multiple, operator) \
  AccessProfile(node, type, Write, 0, 1, multiple, operator)

PRIVATE Node *AccessProfile(Node *node, Node *type, 
			    AccessProfileType accesstype, int withaddress,
			    int opt, int multiple, int operator);
PRIVATE Node *SpawnVarProfile(Node *node, Node *type, OpType assign_op);
PRIVATE void AnalyzeFrameReadWrite(Node *node);

PRIVATE Node *MakeFrameType(const char *procname)
{
     return MakeSdcl(EMPTY_TQ,
		     make_SUE(Sdcl, MakeFrameTypeName(procname), NULL));
}
PRIVATE Node *MakeGenericFrameType(void)
{
     return MakeTdef(EMPTY_TQ, GenericFrameTypeName);
}
PRIVATE Node *MakeInletArgsType(const char *inletname)
{
     return MakeSdcl(EMPTY_TQ,
		 make_SUE(Sdcl, MakeInletArgsTypeName(inletname), NULL));
}

PRIVATE Bool IsInletProc(Node *proc)
{
     Node *decl, *fdcl;
     assert(proc->typ == Proc);
     decl = proc->u.proc.decl;
     assert(decl->typ == Decl);
     fdcl = decl->u.decl.type;
     assert(fdcl->typ == Fdcl);
     return (TYPE_QUALS(fdcl->u.fdcl.tq) & T_INLET) ? 1 : 0;
}

/*
 * Figures out the type (Fdcl) of a spawned function.  Pass the
 * function expression (p->u.spawn.name) to this routine.
 * Also works for inlet decls.
 */
PRIVATE Node *GetSpawnFdcl(Node *spawned_fn)
{
  Node *type = NodeDataType(spawned_fn);
  
  if (IsPtrToFunction(type))
    type = type->u.ptr.type;
  
  assert(type->typ == Fdcl);
  return type;
}


PRIVATE Node *TransformNode(Node *node);

PRIVATE List *TransformList(List *list)
{
     List *aptr;

     for (aptr = list; aptr; aptr = Rest(aptr)) {
	  Node *item = FirstItem(aptr);

	  SetItem(aptr, TransformNode(item));
     }

     return list;
}

PRIVATE inline Node *TransformConst(Node *node, ConstNode *u)
{
     return node;
}

PRIVATE inline Node *TransformId(Node *node, idNode *u)
{
     const char *scope;

     if (!u->decl)
	  return node;

     /* rename id to decl's name: implements static variable renaming */
     assert(u->decl->typ == Decl);
     u->text = u->decl->u.decl.name;

     /* get scope name of id's declaration */
     scope = u->decl->u.decl.scope;

     switch (version) {
	 case Fast:
	      if ((u->decl->u.decl.kind == 2) ||
		  (node->memorycheckedp && (u->decl->u.decl.kind == 3))) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   if (node->memorycheckedp && node->nondeterminator)
		     return CHECK_READ(result, NodeCopy(u->decl->u.decl.type, 
							Subtree));
		   else
		     return (result);
	      }
	      break;
	 case Slow:
	      if (u->decl->u.decl.kind == 2 || u->decl->u.decl.kind == 3) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   if (node->memorycheckedp && node->nondeterminator)
		     return CHECK_READ(result, NodeCopy(u->decl->u.decl.type, 
							Subtree));
		   else
		     return (result);
	      }
	      break;
	 case Link:
	      return MakeBinop('.',
			       MakeId(scope),
			       MakeId(u->decl->u.decl.name));
	 case InletProc:
	      if (u->decl->u.decl.kind == 4) {
		   Node *result = MakeBinop(ARROW,
					    MakeId(InletArgsName),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return result;
	      } else if (u->decl->u.decl.kind != 0) {
		   Node *result = MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(scope)),
					    MakeId(u->decl->u.decl.name));
		   SetCoords(result, node->coord, Subtree);
		   return result;
	      }
	      break;
     }
     if (node->memorycheckedp && node->nondeterminator)
	return CHECK_READ(node, NodeCopy(u->decl->u.decl.type, Subtree));
     else
       return node;
}

PRIVATE inline Node *TransformBinop(Node *node, binopNode *u)
{
     u->left = TransformNode(u->left);

     if (IsAssignmentOp(u->op) && node->memorycheckedp &&
	 node->nondeterminator) {
       if ((u->left->typ == Unary) && (u->left->u.unary.op == INDIR)) {
	 u->left->u.unary.expr = CHECK_WRITE_ADDR(u->left->u.unary.expr,
						  NodeCopy(u->type, Subtree),
						  u->op);
       }
       else {
	 u->left = CHECK_WRITE_MULT(u->left,
				    NodeCopy(u->type, Subtree),
				    node->nondeterminator,
				    u->op);
       }
     }

     u->right = TransformNode(u->right);

     if (node->memorycheckedp && node->nondeterminator) {
       if ((u->op == ARROW) || (u->op == '.'))
	 node = CHECK_READ(node, NodeCopy(u->type, Subtree));
       else if (!IsAssignmentOp(u->op))
	 UNREACHABLE;
     }

     return node;
}

PRIVATE inline Node *TransformUnary(Node *node, unaryNode *u)
{
     if (version == Link && u->op == INDIR) {
	  /* 
	   * transform *a to a[0], *(a+n) to a[n], *(n+a) to a[n], 
	   * and *&x to x 
	   */
	  switch (u->expr->typ) {
	      case Id:
		   return (MakeArray(TransformNode(u->expr), List1(MakeConstSint(0))));
	      case Binop:
		   if (u->expr->u.binop.op == '+' || 
		       u->expr->u.binop.op == '-') {
			if (NodeIsConstant(u->expr->u.binop.left))
			     return (MakeArray(TransformNode(u->expr->u.binop.right),
					       List1(TransformNode(u->expr->u.binop.left))));
			if (NodeIsConstant(u->expr->u.binop.right))
			     return (MakeArray(TransformNode(u->expr->u.binop.left),
					       List1(TransformNode(u->expr->u.binop.right))));
		   }
		   break;
	      case Unary:
		   if (u->expr->u.unary.op == ADDRESS)
			return (TransformNode(u->expr->u.unary.expr));
		   break;
	      default:
		   break;
	  }
     }
     u->expr = TransformNode(u->expr);

     if (node->memorycheckedp && node->nondeterminator) {
       if (IsIncDecOp(u->op)) {
	 u->expr = CHECK_WRITE(u->expr,NodeCopy(u->type, Subtree), u->op);
       }
       else if (u->op == INDIR)
	 u->expr = CHECK_READ_ADDR(u->expr, NodeCopy(u->type, Subtree));
       else 
	 UNREACHABLE;
     }

     return node;
}

PRIVATE inline Node *TransformCast(Node *node, castNode *u)
{
     u->type = TransformNode(u->type);
     u->expr = TransformNode(u->expr);

     return node;
}

PRIVATE inline Node *TransformComma(Node *node, commaNode *u)
{
     u->exprs = TransformList(u->exprs);
     return node;
}

PRIVATE inline Node *TransformTernary(Node *node, ternaryNode *u)
{
     u->cond = TransformNode(u->cond);
     u->true = TransformNode(u->true);
     u->false = TransformNode(u->false);

     return node;
}

PRIVATE inline Node *TransformArray(Node *node, arrayNode *u)
{
     u->name = TransformNode(u->name);
     u->dims = TransformList(u->dims);

     if (node->memorycheckedp && node->nondeterminator)
       node = CHECK_READ_MULT(node, NodeCopy(u->type, Subtree), 
			      node->nondeterminator);

     return node;
}

PRIVATE inline Node *TransformCall(Node *node, callNode *u)
{
     PRIVATE Node *name_strip(Node *);

     u->name = TransformNode(u->name);
     u->args = TransformList(u->args);

     if ((u->name->typ == Id) &&
	 u->name->u.id.text &&
	 (strcmp(u->name->u.id.text, "_LOCK_CHECK") == 0))
       u->args = List2(FirstItem(u->args),
		       name_strip(NodeCopy(FirstItem(u->args), Subtree)));

     return node;
}

PRIVATE inline Node *TransformInitializer(Node *node, initializerNode *u)
{
     u->exprs = TransformList(u->exprs);
     return node;
}

PRIVATE inline Node *TransformImplicitCast(Node *node, implicitcastNode *u)
{
     u->expr = TransformNode(u->expr);
     return node;
}

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Node *TransformLabel(Node *node, labelNode *u)
{
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformSwitch(Node *node, SwitchNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);

     return node;
}

PRIVATE inline Node *TransformCase(Node *node, CaseNode *u)
{
     u->stmt = TransformNode(u->stmt);

     return node;
}

PRIVATE inline Node *TransformDefault(Node *node, DefaultNode *u)
{
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformIf(Node *node, IfNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformIfElse(Node *node, IfElseNode *u)
{
     u->expr = TransformNode(u->expr);
     u->true = TransformNode(u->true);
     u->false = TransformNode(u->false);
     return node;
}

PRIVATE inline Node *TransformWhile(Node *node, WhileNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformDo(Node *node, DoNode *u)
{
     u->expr = TransformNode(u->expr);
     u->stmt = TransformNode(u->stmt);
     return node;
}

PRIVATE inline Node *TransformFor(Node *node, ForNode *u)
{
     u->cond = TransformNode(u->cond);
     u->init = TransformNode(u->init);
     u->next = TransformNode(u->next);
     u->stmt = TransformNode(u->stmt);

     if (OptimizeChecking && node->nondeterminator) {
       List *profile;
       List *rvar, *wvar, *rarr, *warr, *list;
       ListMarker marker;
       Node *item;
       
       rvar = (List *) FirstItem(u->rwlist);
       wvar = (List *) FirstItem(Rest(u->rwlist));
       rarr = (List *) FirstItem(Rest(Rest(u->rwlist)));
       warr = (List *) FirstItem(Rest(Rest(Rest(u->rwlist))));
       profile = List1(node);

       IterateList(&marker, rvar);
       while (NextOnList(&marker, (GenericREF) &item)) {
	 profile = ConsItem(
		     CHECK_READ_OPT(item,
				    NodeCopy(item->u.id.decl->u.decl.type,  
					     Subtree)),
		     profile);
       }

       IterateList(&marker, wvar);
       while (NextOnList(&marker, (GenericREF) &item)) {
	 profile = ConsItem(
		     CHECK_WRITE_OPT(item,
				     NodeCopy(item->u.id.decl->u.decl.type,  
					      Subtree),
				     '='),
		     profile);
       }

       IterateList(&marker, rarr);
       while (NextOnList(&marker, (GenericREF) &list)) {
	 item = (Node *) FirstItem(list);
	 profile = ConsItem(
		     CHECK_READ_OPT_MULT(item,
					 NodeCopy(item->u.array.type,  
						  Subtree),
					 ((Node *) FirstItem(Rest(list)))->u.Const.value.i),
		     profile);
       }

       IterateList(&marker, warr);
       while (NextOnList(&marker, (GenericREF) &list)) {
	 item = (Node *) FirstItem(list);
	 profile = ConsItem(
		     CHECK_WRITE_OPT_MULT(item,
					  NodeCopy(item->u.array.type,  
						   Subtree),
					  ((Node *) FirstItem(Rest(list)))->u.Const.value.i,
					  '='),
		     profile);
       }

       return MakeBlock(NULL, NULL, profile);
     }
     else
       return node;
}

PRIVATE inline Node *TransformGoto(Node *node, GotoNode *u)
{
     return node;
}

PRIVATE inline Node *TransformContinue(Node *node, ContinueNode *u)
{
     return node;
}

PRIVATE inline Node *TransformBreak(Node *node, BreakNode *u)
{
     return node;
}

PRIVATE inline Node *TransformReturn(Node *node, ReturnNode *u)
{
     Node *result;
     List *statements = NULL;
     List *decls = NULL;

     if (!in_procedure)
	  return node;

     if (u->needs_sync) {
	  result = MakeBlockCoord(NULL,
				  NULL,
				  List2(MakeSyncCoord(node->coord),
				  MakeReturnCoord(u->expr, node->coord)),
				  node->coord,
				  UnknownCoord);
	  /* copy over analysis results from return to sync node */
	  ((Node *) FirstItem(result->u.Block.stmts))->analysis.livevars = u->livevars;
	  ((Node *) FirstItem(result->u.Block.stmts))->analysis.dirtyvars = u->dirtyvars;
	  return TransformNode(result);
     }

     u->expr = TransformNode(u->expr);

     /* I'm not sure if this condition is right... -KHR */
     if (in_inlet && (version == Fast || version == Slow))
       return node;
     else if (version == Fast)
       {
	 /* return x ==> { T _tmp = x; _BEFORE_RETURN_FAST(); return _tmp; } */
	 /* return ==> { _BEFORE_RETURN_FAST(); return; } */
	 if (u->expr)
	   {
	     const char *tmp = MakeTempName();
	     Node *returntype = curr_proc->u.proc.decl->u.decl.type->u.fdcl.returns;
	     Node *decl = MakeDecl(tmp,
				   EMPTY_TQ,
				   NodeCopy(returntype, Subtree),
				   u->expr,
				   NULL);
	     decls = List1(decl);
	     
	     statements = List2(MakeCall(MakeId(BeforeReturnFastName), NULL),
				MakeReturn(MakeId(tmp)));
	   }
	 else
	   statements = List2(MakeCall(MakeId(BeforeReturnFastName), NULL),
			      node);
       }
     else if (version == Slow) {
	  /* return x  ==>  {_SET_RESULT(x); _SET_VALID; return;} */
	  /* return  ==>  {_SET_VALID; return;} */
	  if (u->expr)
	       statements = List1(MakeCall(MakeId(SetResultName),
					   List1(u->expr)));
	  else
	       statements = List1(MakeCall(MakeId(SetNoResultName), NULL));

	  statements = AppendItem(statements,
			   MakeCall(MakeId(BeforeReturnSlowName), NULL));
	  statements = AppendItem(statements, MakeReturn(NULL));
     }
     else if (version == InletProc) {
	  /* return x  ==>  {*_inletresult = x; return;} */
	  /* return  ==>  return */
	  if (!u->expr)
	       return MakeBlock(NULL, NULL,
				List2(MakeCall(MakeId(BeforeReturnInletName),
					       NULL),
				      node));
	  statements = List3(MakeBinop('=',
				       MakeUnary('*',
						 MakeId(InletResultName)),
				       u->expr),
			     MakeCall(MakeId(BeforeReturnInletName), NULL),
			     MakeReturn(NULL));
     } else
	  FAIL("bad version in TransformReturn");

     result = MakeBlock(NULL, decls, statements);
     SetCoords(result, node->coord, Subtree);
     return result;
}

PRIVATE inline Node *TransformBlock(Node *node, BlockNode *u)
{
     ListMarker decl_marker;
     Node *decl;

     IterateList(&decl_marker, u->decl);
     while (NextOnList(&decl_marker, (GenericREF) & decl)) {
	  decl = TransformNode(decl);
	  SetCurrentOnList(&decl_marker, decl);
     }

     u->stmts = TransformList(u->stmts);

     return node;
}

/*************************************************************************/
/*                                                                       */
/*                             Type nodes                                */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Node *TransformPrim(Node *node, primNode *u)
{
     return node;
}

PRIVATE inline Node *TransformTdef(Node *node, tdefNode *u)
{
     return node;
}

PRIVATE inline Node *TransformPtr(Node *node, ptrNode *u)
{
     u->type = TransformNode(u->type);
     return node;
}

PRIVATE inline Node *TransformAdcl(Node *node, adclNode *u)
{
     u->type = TransformNode(u->type);
     u->dim = TransformNode(u->dim);
     return node;
}

PRIVATE inline Node *TransformFdcl(Node *node, fdclNode *u)
{
     u->args = TransformList(u->args);
     u->returns = TransformNode(u->returns);

     return node;
}

#define StructCheckFields(x) x=x
#define UnionCheckFields(x) x=x
#define AssignEnumValues(x) x=x

PRIVATE inline Node *TransformSdcl(Node *node, sdclNode *u)
{
     if (SUE_ELABORATED(u->tq))
	  StructCheckFields(u->type);
     return node;
}

PRIVATE inline Node *TransformUdcl(Node *node, udclNode *u)
{
     if (SUE_ELABORATED(u->tq))
	  UnionCheckFields(u->type);
     return node;
}

PRIVATE inline Node *TransformEdcl(Node *node, edclNode *u)
{
     if (SUE_ELABORATED(u->tq))
	  AssignEnumValues(u->type);
     return node;
}

/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/

PRIVATE inline Node *TransformDecl(Node *node, declNode *u)
{
     u->type = TransformNode(u->type);
     u->init = TransformNode(u->init);
     u->bitsize = TransformNode(u->bitsize);

     if (in_procedure && (u->kind == 2 ||
			  ((version == Slow || node->memorycheckedp) && 
			   u->kind == 3))) {
	  if (!u->init)
	       return NULL;
	  {
	       /* We do an incredible hack to get a frame-only variable
		* with initialization to compile correctly.
		*     t x = i;
		* compiles to:
		*     int tmp1 = ({t tmp2 = i; f->x = tmp2; tmp1;});
		*/
	       const char *tmp1 = MakeTempName();
	       const char *tmp2 = MakeTempName();
	       Node *newinit, *newdecl, *assign;
	       Node *id = MakeId(u->name);
	       id->u.id.decl = node;

	       if (NodeDataType(node)->typ == Adcl)
		 assign = MakeCall(MakeId(UniqueString("memcpy")),
				   List3(TransformNode(id),
					 MakeId(tmp2),
					 MakeUnary(SIZEOF,
						   MakeId(tmp2))));
	       else
		 assign = MakeBinop('=', TransformNode(id), MakeId(tmp2));

	       newinit = MakeBlock(NULL,
				   List1(MakeDecl(tmp2,
						  u->tq,
						  u->type,
						  u->init,
						  u->bitsize)),
				   List2(assign,
					 MakeId(tmp1)));
	       newdecl = MakeDecl(tmp1,
				  EMPTY_TQ,
				  MakePrim(EMPTY_TQ, Sint),
				  newinit,
				  NULL);
	       SetCoords(newdecl, node->coord, Subtree);
	       return newdecl;
	  }
     }
     /* delete static declarations (they have been globalized) -KHR */
     if (in_procedure && NodeStorageClass(node) == T_STATIC)
	  return NULL;

     /* remove const in slow declaration (to allow it to be loaded from *
      * frame) */
     if (in_procedure && version == Slow)
	  NodeRemoveTq(node->u.decl.type, T_CONST);

     /*
      * Generate attribute for T_SHARED and T_PRIVATE variables
      */
     if (NodeDeclQuals(node) & T_PRIVATE)
       {
	    if (GenerateSegments) 
		 u->attribs = ConsItem(MakeAttrib(UniqueString("section"),
				   MakeId(UniqueString("\"Cilk_private\""))),
			       ConsItem(MakeAttrib(UniqueString("nocommon"),
						   NULL),
					u->attribs));
	    else
		 WarningCoord(2, node->coord,
			      "`private' qualifier not supported on this target\n");
       }
     else if (NodeDeclQuals(node) & T_SHARED ||
	      (GenerateSegments &&
	       NodeDeclLocation(node) == T_TOP_DECL &&
	       NodeStorageClass(node) != T_EXTERN &&
	       NodeStorageClass(node) != T_REGISTER &&
	       NodeStorageClass(node) != T_TYPEDEF &&
	       NodeDataType(node)->typ != Fdcl))
       {
	    if (NodeDeclQuals(node) & T_SHARED && !GenerateSegments)
		 WarningCoord(2, node->coord,
			      "`shared' qualifier not supported on this target\n");
	    else
		 u->attribs = ConsItem(MakeAttrib(UniqueString("section"),
					  MakeId(UniqueString("\"Cilk_shared\""))),
			       ConsItem(MakeAttrib(UniqueString("nocommon"),
						   NULL),
					u->attribs));
       }

     return node;
}

PRIVATE inline Node *TransformAttrib(Node *node, attribNode *u)
{
     return node;
}

PRIVATE void MakeInletProc(Node *node)
{
  Node *decl = node->u.proc.decl;
  const char *name = MakeInletName(proc_name, decl->u.decl.name);
  Node *args_decl, *args_type;
  Node *inlet_type;
  Node *inlet_decl;
  Node *inlet;
  
  if (!IsVoidType(FirstItem(decl->u.decl.type->u.fdcl.args))) {
    args_decl = MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
			 make_SUE(Sdcl,
				  MakeInletArgsTypeName(name),
				  decl->u.decl.type->u.fdcl.args));
    args_type = MakeInletArgsType(name);
  } else {
    args_decl = NULL;
    args_type = MakePrim(EMPTY_TQ, Void);
  }
  
  inlet_type = MakeFdcl(EMPTY_TQ,
			List3(MakeDecl(FrameName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       MakeFrameType(proc_name)),
				       NULL,
				       NULL),
			      MakeDecl(InletArgsName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       args_type),
				       NULL,
				       NULL),
			      MakeDecl(InletResultName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       NodeCopy(decl->u.decl.type->u.fdcl.returns,
							Subtree)),
				       NULL,
				       NULL)),
			MakePrim(EMPTY_TQ, Void));
  
  inlet_decl = MakeDecl(name,
			T_STATIC | T_STANDALONE_INLET,
			inlet_type,
			NULL,
			NULL);
  if (args_decl)
    SetCoords(args_decl, node->coord, Subtree);
  SetCoords(inlet_decl, node->coord, Subtree);
  in_inlet = 1;
  version = InletProc;
  inlet = MakeProc(inlet_decl,
		   TransformNode(node->u.proc.body));
  version = Slow;
  in_inlet = 0;
  standalone_inlets = AppendItem(standalone_inlets, args_decl);
  standalone_inlets = AppendItem(standalone_inlets, inlet);
}

PRIVATE void MakeImplicitInletProc(Node *node)
{
  const char *name = MakeImplicitInletName(proc_name, sync_count+1);  /* +1 because inlet is made before entry number is allocated */
  Node *inlet_type, *inlet_decl, *stmt, *inlet;
  Node *rhs_type, *lhs_type;
  
  assert(node->typ == Spawn);
  assert(node->u.spawn.assign_op != '=');
  /*
   * Generate inlet code for implicit inlet.  It should look
   * something like this:
   * void _foo_inlet(_foo_frame *_frame, int *arg, int *result)
   * { *result += *arg; }
   * Note: frame pointer is not used.
   */
  rhs_type = NodeCopy(GetSpawnFdcl(node->u.spawn.name)->u.fdcl.returns, Subtree);
  lhs_type = NodeCopy(NodeDataType(node->u.spawn.receiver), Subtree);
  
  inlet_type = MakeFdcl(EMPTY_TQ,
			List3(MakeDecl(FrameName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       MakeFrameType(proc_name)),
				       NULL,
				       NULL),
			      MakeDecl(InletArgsName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       rhs_type),
				       NULL,
				       NULL),
			      MakeDecl(InletResultName,
				       EMPTY_TQ,
				       MakePtr(EMPTY_TQ,
					       lhs_type),
				       NULL,
				       NULL)),
			MakePrim(EMPTY_TQ, Void));
  
  inlet_decl = MakeDecl(name,
			T_STATIC | T_STANDALONE_INLET,
			inlet_type,
			NULL,
			NULL);
  stmt = MakeBinop(node->u.spawn.assign_op,
		   MakeUnary('*', MakeId(InletResultName)),
		   MakeUnary('*', MakeId(InletArgsName)));
  inlet = MakeProc(inlet_decl,
		   MakeBlock(NULL,
			     NULL,
			     List1(stmt)));
  
  SetCoords(inlet, UnknownCoord, Subtree);
  standalone_inlets = AppendItem(standalone_inlets, inlet);
}

PRIVATE inline Node *TransformProc(Node *node, procNode *u)
{
     Node *standalone_copy = NULL;

     if (IsInletProc(node)) {
	  assert(in_procedure && !in_inlet);
	  if (version == Slow)
	       standalone_copy = NodeCopy(node, Subtree);
	  in_inlet = 1;
     }
     else {
          /* indicate this starts C code for CILK_WHERE_AM_I */
          NodeAddTq(u->decl, T_C_CODE);
     }
     
     u->decl = TransformNode(u->decl);
     u->body = TransformNode(u->body);

     if (IsInletProc(node)) {
	  in_inlet = 0;

          /* Change the type qualifier into T_SLOW_INLET for CILK_WHERE_AM_I */
          if (version == Fast)
	      NodeAddTq(u->decl, T_FAST_INLET);
          if (version == Slow)
	      NodeAddTq(u->decl, T_SLOW_INLET);

	  /* make standalone inlet */
	  if (version == Slow)
	      MakeInletProc(standalone_copy);
     }

     return node;
}

PRIVATE inline Node *TransformText(Node *node, textNode *u)
{
     return node;
}

PRIVATE List *SaveVariables(List *vars, Node *except)
{
     ListMarker m;
     Node *var;
     List *statements = NULL;

     IterateList(&m, vars);
     while (NextOnList(&m, (GenericREF) & var)) {
	  if (var != except &&
	      (var->u.decl.kind == 1 ||
	       (version == Fast && (!(var->memorycheckedp)) &&
		var->u.decl.kind == 3))) {
	       statements = AppendItem(statements,
				       MakeBinop('=',
						 MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
					      MakeId(var->u.decl.scope)),
					       MakeId(var->u.decl.name)),
					      MakeId(var->u.decl.name)));
	  }
     }
     return (statements);
}

PRIVATE List *RestoreVariables(List *vars)
{
     ListMarker m;
     Node *var;
     List *statements = NULL;

     IterateList(&m, vars);
     while (NextOnList(&m, (GenericREF) & var)) {
	  if ((var->u.decl.kind == 1 ||
	       (version == Fast && (!(var->memorycheckedp)) &&
		var->u.decl.kind == 3))) {
	       statements = AppendItem(statements,
				       MakeBinop('=',
						 MakeId(var->u.decl.name),
						 MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
					      MakeId(var->u.decl.scope)),
					     MakeId(var->u.decl.name))));
	  }
     }
     return (statements);
}

/*
 * This function generates one of two macros
 *
 * _XPOP_FRAME_RESULT(frame, return_value, stuff for slow, result)
 * _XPOP_FRAME_NORESULT(frame, return_value, stuff for slow)
 *
 * depending on the presence of a receiver for the result.
 */

PRIVATE inline Node *MakeXPopFrame(Node *node, spawnNode *u)
{
     Node *frame;
     Node *return_value;

     /* return type of the procedure being compiled, not the procedure *
      * being spawned! */
     Node *returntype = curr_proc->u.proc.decl->u.decl.type->u.fdcl.returns;

     frame = MakeId(FrameName);

     if (version == Slow)
	  return_value = MakeText(UniqueString("/* return nothing */"),
				  FALSE);
     else if (IsScalarType(returntype))
	  return_value = MakeConstSint(0);
     else if (IsVoidType(returntype))
	  return_value = MakeText(UniqueString("/* return nothing */"),
				  FALSE);
     else
	  return_value = MakeBinop(ARROW,
				   MakeId(FrameName),
				   MakeId(DummyReturnName));

     if (u->receiver)
	  return MakeCall(MakeId(XPopFrameResultName),
			  List3(frame,
				return_value,
				u->receiver));
     else
	  return MakeCall(MakeId(XPopFrameNoResultName),
			  List2(frame,
				return_value));
}

PRIVATE Node *MakeSetIndex(Node *receiver_addr)
{
     /* sets index of current frame to be the index of receiver in the *
      * frame. */
     /* _frame->_header.index = (char*) receiver_addr - (char*)_frame */

     return MakeBinop('=',
		      MakeBinop('.',
				MakeBinop(ARROW,
					  MakeId(FrameName),
					  MakeId(HeaderName)),
				MakeId(IndexName)),
		      MakeBinop('-',
				MakeCast(MakePtr(EMPTY_TQ,
					       MakePrim(EMPTY_TQ, Char)),
					 receiver_addr),
				MakeCast(MakePtr(EMPTY_TQ,
					       MakePrim(EMPTY_TQ, Char)),
					 MakeId(FrameName))));
}

PRIVATE Bool SimplePointer(Node *receiver);

/* returns TRUE if receiver's address in the frame can be calculated
 * statically, and if receiver is side-effect-free.  Returns FALSE otherwise.
 */
PRIVATE Bool SimpleReceiver(Node *receiver)
{
     List *dims;

     switch (receiver->typ) {
	 case Id:
	      if (receiver->u.id.decl->u.decl.kind != 0)
		   return TRUE;
	      break;
	 case Binop:
	      if (receiver->u.binop.op == '.')
		   return SimpleReceiver(receiver->u.binop.left);
	      break;
	 case Unary:
	      if (receiver->u.unary.op == INDIR)
		   return SimplePointer(receiver->u.unary.expr);
	      break;
	 case Array:
	      /* simple if all array dimensions are constant */
	      for (dims = receiver->u.array.dims; dims; dims = Rest(dims))
		   if (!NodeIsConstant(FirstItem(dims)))
			return FALSE;
	      return SimplePointer(receiver->u.array.name);
	 default:
	      /* there are probably other cases, but I don't feel like *
	       * dealing * with them now. KHR */
	      break;
     }
     return FALSE;
}
/* returns TRUE if pointer's offset in frame can be calculated statically,
 * and if receiver is side-effect-free.  Returns FALSE otherwise.
 */
PRIVATE Bool SimplePointer(Node *pointer)
{
     switch (pointer->typ) {
	 case Id:
	      if (pointer->u.id.decl->u.decl.type->typ == Adcl)
		   return TRUE;
	      break;
	 case Binop:
	      if (pointer->u.binop.op == '+' || pointer->u.binop.op == '-') {
		   if (NodeIsConstant(pointer->u.binop.right) &&
		       SimplePointer(pointer->u.binop.left))
			return TRUE;
		   if (NodeIsConstant(pointer->u.binop.left) &&
		       SimplePointer(pointer->u.binop.right))
			return TRUE;
	      }
	      break;
	 case Unary:
	      if (pointer->u.unary.op == ADDRESS)
		   return SimpleReceiver(pointer->u.unary.expr);
	      break;
	 case Cast:
	      return SimplePointer(pointer->u.cast.expr);
	 case ImplicitCast:
	      return SimplePointer(pointer->u.implicitcast.expr);
	 default:
	      break;
     }
     return FALSE;
}

/* gcc asm extensions */

PRIVATE inline Node *TransformAsm(Node *node, asmNode *u)
{
     u->template = TransformNode(u->template);
     u->output = TransformList(u->output);
     u->input = TransformList(u->input);
     u->clobbered = TransformList(u->clobbered);
     return node;
}
PRIVATE inline Node *TransformAsmArg(Node *node, asmargNode *u)
{
     u->constraint = TransformNode(u->constraint);
     u->expr = TransformNode(u->expr);
     return node;
}

PRIVATE Bool side_effect_free;

PRIVATE void CheckSideEffectFree(Node *node)
{
     switch (node->typ) {
	 case Binop:
	      if (IsAssignmentOp(node->u.binop.op)) {
		   side_effect_free = FALSE;
		   return;
	      }
	      break;
	 case Unary:
	      if (IsIncDecOp(node->u.unary.op)) {
		   side_effect_free = FALSE;
		   return;
	      }
	      break;
	 case Block:
	 case Call:
	 case Spawn:
	      side_effect_free = FALSE;
	      return;
	 case Label:
	 case Switch:
	 case Case:
	 case Default:
	 case If:
	 case IfElse:
	 case While:
	 case Do:
	 case For:
	 case Goto:
	 case Continue:
	 case Break:
	 case Return:
	 case Sync:
	      FAIL("Statement in spawn arg");
	      break;
	 case Prim:
	 case Tdef:
	 case Adcl:
	 case Fdcl:
	 case Sdcl:
	 case Udcl:
	 case Edcl:
	 case Decl:
	 case Attrib:
	 case Proc:
	      /* don't need to touch types */
	      return;
	 default:
	      break;
     }

#define CHILD(n)   CheckSideEffectFree(n)
     ASTWALK(node, CHILD);
#undef CHILD
}

PRIVATE Bool SideEffectFree(Node *node)
{
     side_effect_free = TRUE;
     CheckSideEffectFree(node);
     return side_effect_free;
}

/* returns a decl of a temporary variable with type <type> */
PRIVATE Node *MakeTempDecl(Node *type)
{
     const char *temp_arg = MakeTempName();
     return MakeDecl(temp_arg,
		     EMPTY_TQ,
		     NodeCopy(type, Subtree),
		     NULL,
		     NULL);
}

PRIVATE inline List *TransformSpawnArgs(List *args,
					List **declarations,
					List **statements)
{
     List *side_effect_free_args;
     Node *arg, *temp_arg, *temp_decl;
     ListMarker marker;

     side_effect_free_args = NULL;
     IterateList(&marker, args);
     while (NextOnList(&marker, (GenericREF) & arg)) {
	  if (!SideEffectFree(arg)) {
	       temp_decl = MakeTempDecl(NodeDataType(arg));
	       temp_arg = MakeId(temp_decl->u.decl.name);

	       *declarations = AppendItem(*declarations,
					  temp_decl);
	       *statements = AppendItem(*statements,
					MakeBinop('=',
					  MakeId(temp_decl->u.decl.name),
						  TransformNode(arg)));
	  } else
	       temp_arg = TransformNode(arg);
	  side_effect_free_args = AppendItem(side_effect_free_args, temp_arg);
     }

     return side_effect_free_args;
}

PRIVATE Node *TransformReceiver(Node *receiver,
				OpType assign_op,
				Node *return_type,
				Node **index,	/* linkage info         */
				Node **assignment,	/* assignment *
							 * statement */
				List **declarations,
				List **statements)
{
     Node *rec, *receiver_tmp, *receiver_addr;
     Node *spawn_result = NULL;
     int receiver_type;		/* 
				 * 0 - no receiver
				 * 1 - register variable
				 * 2 - fixed location in frame
				 *       (SimpleReceiver) 
				 * 3 - unknown location
				 *  or side-effecting receiver */
     /* figure out receiver type */
     if (!receiver)
	  receiver_type = 0;
     else if (receiver->typ == Id &&
	      (receiver->u.id.decl->u.decl.kind == 1 ||
	     (version == Fast && (!(receiver->memorycheckedp)) &&
	      receiver->u.id.decl->u.decl.kind == 3)))
	  receiver_type = 1;
     else if (SimpleReceiver(receiver))
	  receiver_type = 2;
     else
	  receiver_type = 3;

     /* generate index part of link information - slow version only. */
     if (version == Slow)
	  switch (receiver_type) {
	      case 0:		/* no receiver */
		   *index = MakeConstSint(0);
		   break;
	      case 1:		/* register variable */
		   *index = MakeCall(MakeId("offsetof"),
				     List2(MakeFrameType(proc_name),
					   MakeBinop('.',
			       MakeId(receiver->u.id.decl->u.decl.scope),
			     MakeId(receiver->u.id.decl->u.decl.name))));
		   break;
	      case 2:		/* fixed loc in frame */
		   rec = NodeCopy(receiver, Subtree);
		   version = Link;
		   rec = TransformNode(rec);
		   version = Slow;
		   *index = MakeCall(MakeId("offsetof"),
				     List2(MakeFrameType(proc_name),
					   rec));
		   break;
	      case 3:		/* unknown location */
		   *index = MakeConstSint(-1);
		   break;
	      default:
		   FAIL("bad receiver type");
		   break;
	  }

     /* make <spawn_result>, <assignment> */
     switch (receiver_type) {
	 case 0:		/* no receiver */
	      spawn_result = NULL;
	      *assignment = NULL;
	      break;
	 case 1:		/* register variable */
	      if (assign_op == '=') {
		   spawn_result = receiver;
		   *assignment = NULL;
		   break;
	      }
	      /* otherwise, fall through */
	 case 2:		/* fixed loc in frame */
	      receiver_tmp = MakeTempDecl(NodeCopy(return_type, Subtree));
	      *declarations = AppendItem(*declarations, receiver_tmp);
	      spawn_result = MakeId(receiver_tmp->u.decl.name);
	      rec = NodeCopy(receiver, Subtree);
	      rec->nondeterminator = FALSE;
	      *assignment = MakeBinop(assign_op,
				      TransformNode(rec),
				      MakeId(receiver_tmp->u.decl.name));
	      break;
	 case 3:		/* unknown location */
	      receiver_tmp = MakeTempDecl(NodeCopy(return_type, Subtree));
	      *declarations = AppendItem(*declarations, receiver_tmp);
	      receiver_addr = MakeTempDecl(MakePtr(EMPTY_TQ,
						NodeDataType(receiver)));
	      *declarations = AppendItem(*declarations, receiver_addr);
	      rec = NodeCopy(receiver, Subtree);
	      *statements = AppendItem(*statements,
				       MakeBinop('=',
				      MakeId(receiver_addr->u.decl.name),
					 TransformNode(MakeUnary(ADDRESS,
							    rec))));
	      spawn_result = MakeId(receiver_tmp->u.decl.name);
	      *assignment = MakeBinop(assign_op,
				      MakeUnary('*',
				     MakeId(receiver_addr->u.decl.name)),
				      MakeId(receiver_tmp->u.decl.name));
	      *statements = AppendItem(*statements,
				       MakeSetIndex(MakeId(receiver_addr->u.decl.name)));
	      break;
     }
     return spawn_result;
}

PRIVATE void TransformSpawn2(Node *node, spawnNode *u,
			     List **declarations, List **statements)
{
     Node *index, *assignment, *rec;
     Node *return_type = GetSpawnFdcl(u->name)->u.fdcl.returns;

     assert(in_procedure && !in_inlet);

     /* reserve a sync number */
     sync_count++;

     rec = u->receiver;
     u->receiver = TransformReceiver(u->receiver, u->assign_op, return_type, &index,
				     &assignment, declarations, statements);
     assert(SideEffectFree(u->name)); /* need this for now -KHR */
     u->name = TransformNode(u->name);
     u->args = TransformSpawnArgs(u->args, declarations, statements);

     /* set up link information */
     if (version == Slow && !u->in_inlet) {
          Node *size, *inlet, *argsize, *argindex;
	  if (u->receiver)
	    size = MakeUnary(SIZEOF, NodeCopy(return_type, Subtree));
	  else
	    size = MakeConstSint(0);
	  
	  if(u->assign_op == '=')
	    {
	      inlet = MakeConstSint(0);
	      argsize = MakeConstSint(0);
	      argindex = MakeConstSint(0);
	    }
	  else
	    {
	      inlet = MakeId(MakeImplicitInletName(proc_name, sync_count));
	      argsize = MakeUnary(SIZEOF, NodeCopy(return_type, Subtree));
	      argindex = MakeConstSint(0);
	    }
	  
	  link_initializers =
	    AppendItem(link_initializers,
		       MakeInitializer(List5(size,
					     index,
					     inlet,
					     argsize,
					     argindex)));
     }
     /* make entry setting statement */
     *statements = AppendItem(*statements,
			      MakeBinop('=',
					MakeBinop('.',
						  MakeBinop(ARROW,
						       MakeId(FrameName),
						     MakeId(HeaderName)),
						  MakeId(EntryName)),
					MakeConstSint(sync_count)));

     /* save live, dirty variables */
     *statements = JoinLists(*statements,
			  SaveVariables(node->analysis.dirtyvars, NULL));

     /* Insert _BEFORE_SPAWN_[FAST|SLOW] hook */
     *statements =
	 AppendItem(*statements,
		  MakeCall(MakeId(version == Slow ? BeforeSpawnSlowName :
				  BeforeSpawnFastName),
			   NULL)
	 );

     /* push frame */
     *statements = AppendItem(*statements,
			      MakeCall(MakeId(PushFrameName),
				       List1(MakeId(FrameName))));

     /* do call */
     if (u->receiver) {
	  *statements = AppendItem(*statements,
				   MakeBinop('=',
					     u->receiver,
					     MakeCall(u->name,
						      u->args)));
	  if (node->memorycheckedp && (!(u->in_inlet))) {
	    /* inform the Nondeterminator that this is a spawn var */
	    *statements = AppendItem(*statements,
				     SpawnVarProfile(TransformNode(rec),
						     NodeCopy(return_type,
							      Subtree),
						     u->assign_op));
	  }
     }
     else
	  *statements = AppendItem(*statements,
				   MakeCall(u->name,
					    u->args));

     /* Extended pop: pop, check for steal, locking (in some unspecified * 
      * order) */
     *statements = AppendItem(*statements,
			      MakeXPopFrame(node, u));

     /* Insert _AFTER_SPAWN_[FAST|SLOW] hook */
     *statements =
	 AppendItem(*statements,
		    MakeCall(MakeId(version == Slow ? AfterSpawnSlowName :
				    AfterSpawnFastName),
			     NULL)
	 );

     /* save spawn result somewhere, if necessary */
     if (assignment != NULL)
	  *statements = AppendItem(*statements, assignment);
}

PRIVATE inline Node *TransformSpawn(Node *node, spawnNode *u)
{
     List *declarations = NULL;
     List *statements = NULL;
     Node *result;
     Coord coord = u->receiver ? u->receiver->coord : node->coord;

     /* make standalone implicit inlet */
     if (version == Slow && u->assign_op != '=')
       MakeImplicitInletProc(node);

     /* do guts of transformation */
     TransformSpawn2(node, u, &declarations, &statements);

     /* delay the call of Cilk_lreturn_spbags after the inlet is called */
     if (MemoryChecking && u->receiver)
       statements = AppendItem(statements,
			       MakeCall(MakeId("_ND_LRETURN_SPBAGS"),
					List1(MakeId(FrameName))));
       
     /* Add entry recovery statements to slow version */
     if (version == Slow) {
	  statements = AppendItem(statements,
				  MakeIf(MakeConstSint(0),
					 MakeBlock(NULL, NULL,
			     ConsItem(MakeLabel(MakeSyncName(sync_count),
						NULL),
			   RestoreVariables(node->analysis.livevars)))));
	  statements = AppendItem(statements,
				  MakeCall(MakeId(AtThreadBoundarySlowName),
						  NULL));
     }

     /* make block of stuff to replace spawn statement */
     result = MakeBlock(NULL, declarations, statements);
     SetCoords(result, coord, Subtree);
     return result;
}

PRIVATE inline Node *TransformSync(Node *node, syncNode *u)
{
     Node *result;
     List *statements;
     List *outerstatements;

     assert(in_procedure && !in_inlet);

     /* reserve a sync number */
     sync_count++;

     if (version == Fast)
       {
	 result = MakeCall(MakeId(AtSyncFastName), NULL);
	 SetCoords(result, node->coord, Subtree);
	 return result;
       }

     /* set up dummy link information */
     if (version == Slow) {
	  link_initializers =
	      AppendItem(link_initializers,
			 MakeInitializer(List5(MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0),
					       MakeConstSint(0))));
     }
     /* 
      * generate sync code.  Sync code goes something like this:   
      * { 
      *     _BEFORE_SYNC_SLOW(); 
      *     _frame->_header.entry = 3;
      *     <save dirty variables> 
      *     if (_SYNC) { 
      *          _POP_FRAME(_frame); 
      *          return; 
      *    _sync3:; 
      *          <restore live variables> 
      *    } 
      *    _AFTER_SYNC_SLOW();
      * } 
      */

     outerstatements = List1(MakeCall(MakeId(BeforeSyncSlowName), NULL));

     /* set entry */
     outerstatements = AppendItem(outerstatements,
			     MakeBinop('=',
				  MakeBinop('.',
					    MakeBinop(ARROW,
						      MakeId(FrameName),
						      MakeId(HeaderName)),
					    MakeId(EntryName)),
				  MakeConstSint(sync_count)));

     /* save live, dirty variables */
     outerstatements = JoinLists(outerstatements,
			  SaveVariables(node->analysis.dirtyvars, NULL));

     /* return */
     statements = List1(
			     MakeReturn(NULL)
	  );

     /* sync label */
     statements = AppendItem(statements,
			     MakeLabel(MakeSyncName(sync_count),
				       NULL));

     /* restore variables */
     statements = JoinLists(statements,
			    RestoreVariables(node->analysis.livevars));

     outerstatements = AppendItem(outerstatements,
				  MakeIf(MakeId(SynchronizeName),
				     MakeBlock(NULL, NULL, statements)));

     outerstatements = AppendItem(outerstatements,
			      MakeCall(MakeId(AfterSyncSlowName), NULL));
     outerstatements = AppendItem(outerstatements,
				  MakeCall(MakeId(AtThreadBoundarySlowName),
						  NULL));

     result = MakeBlock(NULL, NULL, outerstatements);

     SetCoords(result, node->coord, Subtree);
     return result;
}

PRIVATE Node *TransformInletCall(Node *node, inletcallNode *u)
{
     List *declarations = NULL;
     List *statements = NULL;
     Node *result;
     Node *spawnarg;
     List *otherargs;
     List *formals, *actuals;
     Node *temp_arg, *temp_decl;
     Node *index, *assignment;
     Coord coord = u->receiver ? u->receiver->coord : node->coord;
     Bool void_spawn;

     assert(in_procedure && !in_inlet);

     /* get spawn argument and any other arguments */
     assert(u->args != NULL);
     spawnarg = FirstItem(u->args);
     assert(spawnarg->typ == Spawn);
     void_spawn = IsVoidType(spawnarg);
     otherargs = Rest(u->args);

     u->receiver = TransformReceiver(u->receiver, '=', NodeDataType(spawnarg), &index,
				&assignment, &declarations, &statements);
     u->name = TransformNode(u->name);
     otherargs = TransformSpawnArgs(otherargs, &declarations, &statements);

     /* set up link information */
     if (version == Slow) {
	  Node *size, *inlet, *argindex, *argsize;
	  const char *scope;
	  const char *name;

	  /* calculate size entry - size of spawn's return value */
	  if (void_spawn)
	       size = MakeConstSint(0);
	  else
	       size = MakeUnary(SIZEOF, NodeDataType(spawnarg));

	  /* make inlet pointer */
	  name = MakeInletName(proc_name, u->name->u.id.decl->u.decl.name);
	  inlet = MakeId(name);

	  /* make argsize and argindex */
	  formals = GetSpawnFdcl(u->name)->u.fdcl.args;
	  if (formals != NULL && !IsVoidType(FirstItem(formals))) {
	       scope = ((Node *) FirstItem(formals))->u.decl.scope;
	       assert(scope != NULL);
	       argsize = MakeUnary(SIZEOF,
				   MakeInletArgsType(name));
	       argindex = MakeCall(MakeId("offsetof"),
				   List2(MakeFrameType(proc_name),
					 MakeId(scope)));
	  } else {
	       argsize = MakeConstSint(0);
	       argindex = MakeConstSint(0);
	  }
	  link_initializers =
	      AppendItem(link_initializers,
			 MakeInitializer(List5(size,
					       index,
					       inlet,
					       argsize,
					       argindex)));
     }
     /* save other arguments besides spawn argument to frame */
     actuals = otherargs;
     if (void_spawn)
	  formals = GetSpawnFdcl(u->name)->u.fdcl.args;
     else
	  formals = Rest(GetSpawnFdcl(u->name)->u.fdcl.args);
     for (; actuals != NULL; actuals = Rest(actuals), formals = Rest(formals)) {
	  Node *actual = FirstItem(actuals);
	  Node *formal = FirstItem(formals);
	  formal = MakeBinop('.',
			     MakeBinop(ARROW,
				       MakeId(FrameName),
				       MakeId(formal->u.decl.scope)),
			     MakeId(formal->u.decl.name));
	  statements = AppendItem(statements,
				  MakeBinop('=',
					    formal,
					    actual));
     }

     /* transform spawn argument */
     if (!void_spawn) {
	  temp_decl = MakeTempDecl(NodeDataType(spawnarg));

	  /* treat this variable as a register variable, regardless of *
	   * type */
	  temp_decl->u.decl.kind = 1;
	  declarations = AppendItem(declarations,
				    temp_decl);

	  temp_arg = MakeId(temp_decl->u.decl.name);
	  temp_arg->u.id.decl = temp_decl;

	  spawnarg->u.spawn.receiver = temp_arg;
	  otherargs = ConsItem(temp_arg, otherargs);
     } else
	  spawnarg->u.spawn.receiver = NULL;

     TransformSpawn2(spawnarg, &spawnarg->u.spawn, &declarations, &statements);

     /* tell the Nondeterminator that we are entering an inlet */
     if (MemoryChecking)
       statements = AppendItem(statements,
			       MakeCall(MakeId("_ND_ENTER_INLET"), 
					List1(MakeCast(
						MakePtr(EMPTY_TQ,
						  MakeGenericFrameType()),
						MakeId(FrameName)))));

     /* do call of inlet */
     if (u->receiver)
	  statements = AppendItem(statements,
				  MakeBinop('=',
					    u->receiver,
					    MakeCall(u->name,
						     otherargs)));
     else
	  statements = AppendItem(statements,
				  MakeCall(u->name,
					   otherargs));

     /* tell the Nondeterminator that we are leaving an inlet */
     if (MemoryChecking)
       statements = AppendItem(statements,
			       MakeCall(MakeId("_ND_LEAVE_INLET"), NULL));

     /* save inlet result somewhere, if necessary */
     if (assignment != NULL)
	  statements = AppendItem(statements, assignment);

     /* delay the call of Cilk_lreturn_spbags after the inlet is called */
     if (MemoryChecking)
       statements = AppendItem(statements,
			       MakeCall(MakeId("_ND_LRETURN_SPBAGS"),
					List1(MakeId(FrameName))));
       
     /* Add entry recovery statements to slow version */
     if (version == Slow) {
	  statements = AppendItem(statements,
				  MakeIf(MakeConstSint(0),
					 MakeBlock(NULL, NULL,
			     ConsItem(MakeLabel(MakeSyncName(sync_count),
						NULL),
			   RestoreVariables(node->analysis.livevars)))));
	  statements = AppendItem(statements,
				  MakeCall(MakeId(AtThreadBoundarySlowName),
						  NULL));
     }


     /* make block of stuff to replace inlet_call statement */
     result = MakeBlock(NULL, declarations, statements);
     SetCoords(result, coord, Subtree);
     return result;
}

PRIVATE inline Node *TransformResume(Node *node, resumeNode *u)
{
     List *statements;
     Node *result;

     assert(in_procedure && in_inlet);

     u->expr = TransformNode(u->expr);

     switch (version) {
	 case Fast:
	 case Slow:
	      return MakeReturnCoord(u->expr, node->coord);
	 case InletProc:
	      statements = List1(MakeCall(MakeId(ResumeName), NULL));
	      if (u->expr)
		   statements = AppendItem(statements,
					   MakeBinop('=',
						     MakeUnary('*',
						MakeId(InletResultName)),
						     u->expr));
	      statements = AppendItem(statements,
				      MakeReturn(NULL));
	      result = MakeBlock(NULL,
				 NULL,
				 statements);
	      SetCoords(result, node->coord, Subtree);
	      return result;
	      break;
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformAbort(Node *node, abortNode *u)
{
     Node *result;

     assert(in_procedure);

     switch (version) {
	 case Fast:
	      return MakeTextCoord(UniqueString("/* abort */;"), FALSE, node->coord);
	 case Slow:
	      result = MakeCall(MakeId(AbortSlowName), NULL);
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 case InletProc:
	      result = MakeCall(MakeId(AbortStandaloneName), NULL);
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformDoubleReturn(Node *node, doublereturnNode *u)
{
     Node *result;

     assert(in_procedure && in_inlet);

     switch (version) {
	 case Fast:
	      result = MakeCall(MakeId(DoubleReturnFastName),
				MakeNewList(u->expr));
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 case Slow:
	      result = MakeCall(MakeId(DoubleReturnSlowName),
				MakeNewList(u->expr));
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 case InletProc:
	      result = MakeCall(MakeId(DoubleReturnStandaloneName),
				MakeNewList(u->expr));
	      SetCoords(result, node->coord, Subtree);
	      return result;
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformOnAbort(Node *node, onabortNode *u)
{
     assert(in_procedure && !in_inlet);
     if (on_abort_defined) {
	  SyntaxErrorCoord(node->coord, "on_abort is redefined in a cilk procedure");
	  return node;
     } else {
	  Node *decl = MakeDecl("_on_abort",
				EMPTY_TQ,
				MakeFdcl(EMPTY_TQ,
					 MakeNewList(PrimVoid),
			     MakePrimCoord(EMPTY_TQ, Void, node->coord)),
				NULL,
				NULL);
	  Node *stmt;

	  /* temporary turn off in_procedure to disable Cilk constructs in 
	   * 
	   * * on_abort */
	  in_procedure = 0;
	  stmt = TransformNode(u->stmt);
	  in_procedure = 1;

	  /* turn on on_abort_defined to disable other on_abort in the *
	   * same procedure */
	  on_abort_defined = 1;
	  return MakeProc(decl, stmt);
     }
}

PRIVATE inline Node *TransformSynched(Node *node, synchedNode *u)
{
     assert(in_procedure && !in_inlet);

     switch (version) {
	 case Fast:
	      return MakeConstSint(1);
	 case Slow:
	 case InletProc:
	      return MakeId(SynchedName);
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE inline Node *TransformSpawnCounter(Node *node, spawncounterNode *u)
{
     assert(in_procedure && !in_inlet);
     assert(0);  /* removed spawn counter -KHR */

     switch (version) {
	 case Fast:
	      return MakeConstSint(0);
	 case Slow:
	 case InletProc:
	      return MakeId(SynchronizeName);
	 default:
	      FAIL("Unknown version");
     }

     return node;
}

PRIVATE Node *MakeSlowDecl(const char *name)
{
     Node *arg = MakeDecl(FrameName,
			  T_FORMAL_DECL,
			  MakePtr(EMPTY_TQ, MakeFrameType(name)),
			  NULL,
			  NULL);
     return MakeDecl(MakeSlowProcName(name),
		     T_STATIC | T_TOP_DECL | T_SLOW_PROCEDURE,
		     MakeFdcl(EMPTY_TQ, List1(arg), MakePrim(EMPTY_TQ, Void)),
		     NULL,
		     NULL);
}

PRIVATE void SetDecl(Node *decl, List *register_vars, List *spawn_vars,
		     int scope_count)
{
     TypeQual sc = NodeStorageClass(decl);
     if (sc == T_EXTERN || sc == T_STATIC || sc == T_TYPEDEF ||
	 decl->u.decl.type->typ == Fdcl)
	  return;		/* leave as kind 0 */

     if (scope_count > 0 && DECL_LOCATION(decl->u.decl.tq) == T_FORMAL_DECL) {
	  /* inlet argument variable */
	  decl->u.decl.kind = 4;
     } else if (FindItem(register_vars, decl)) {
	  /* decl is for a register variable. */
	  if (FindItem(spawn_vars, decl))
	       decl->u.decl.kind = 3;
	  else
	       decl->u.decl.kind = 1;
     } else {
	  /* decl is for a frame-only variable. */
	  decl->u.decl.kind = 2;
     }

     decl->u.decl.scope = MakeScopeName(scope_count);
}

PRIVATE List *FindAllScopes(Node *node, List *declscopes)
{
     if (node->typ == Block && node->u.Block.decl)
	  declscopes = AppendItem(declscopes, node->u.Block.decl);
     else if (node->typ == Proc && IsInletProc(node)) {
	  /* add function arguments to scope list */
	  Node *type = node->u.proc.decl->u.decl.type;
	  assert(type->typ == Fdcl);

	  declscopes = AppendItem(declscopes, type->u.fdcl.args);
	  return declscopes;
	  /* return keeps FindAllScopes from looking inside local *
	   * procedure */
     }
     /* transform children */
#define CHILD(n)  declscopes = FindAllScopes(n, declscopes)
     ASTWALK(node, CHILD)
#undef CHILD

	 return declscopes;
}

PRIVATE List *FindSideEffectedVariables(Node *node, List *sofar)
{
     switch (node->typ) {
	 case Binop:
	      if (IsAssignmentOp(node->u.binop.op) &&
		  node->u.binop.left->typ == Id)
		   sofar = ConsItem(node->u.binop.left->u.id.decl, sofar);
	      break;
	 case Unary:
	      if (IsIncDecOp(node->u.unary.op) &&
		  node->u.unary.expr->typ == Id)
		   sofar = ConsItem(node->u.unary.expr->u.id.decl, sofar);
	      break;
	 default:
	      break;
     }

#define CODE(n)   sofar = FindSideEffectedVariables(n, sofar)
     ASTWALK(node, CODE)
#undef CODE

	 return sofar;
}

PRIVATE List *FindSpawnVariables(Node *node, List *sofar)
{
     if (node->typ == Spawn) {
	  if (node->u.spawn.receiver &&
	      node->u.spawn.receiver->typ == Id)
	       return ConsItem(node->u.spawn.receiver->u.id.decl, sofar);
     } else if (node->typ == InletCall) {
	  if (node->u.inletcall.receiver &&
	      node->u.inletcall.receiver->typ == Id)
	       return ConsItem(node->u.inletcall.receiver->u.id.decl, sofar);
     } else if (node->typ == Proc && IsInletProc(node)) {
	  /* need to find all variables modified by the inlet.  These * *
	   * variables need to be in the frame for the slow version. */
	  return JoinLists(sofar,
			   FindSideEffectedVariables(node, NULL));
     }
     /* transform children */
#define CHILD(n)  sofar = FindSpawnVariables(n, sofar)
     ASTWALK(node, CHILD)
#undef CHILD

	 return sofar;
}

PRIVATE void AnalyzeVariables(Node *node)
{
     ListMarker marker;
     Node *decl;
     int scope_count;

     /* find out which variables can be stored in registers */
     List *register_vars = RegisterVariables(node, NULL);
     List *spawn_vars = FindSpawnVariables(node, NULL);

     /* find all local scopes */
     List *scopes = FindAllScopes(node, NULL);
     scopes = ConsItem(node->u.proc.decl->u.decl.type->u.fdcl.args, scopes);

     /* iterate over all local variables in the procedure */
     scope_count = 0;
     while (scopes) {
	  IterateList(&marker, FirstItem(scopes));
	  while (NextOnList(&marker, (GenericREF) & decl)) {
	       if (decl->typ == Decl)
		    SetDecl(decl, register_vars, spawn_vars, scope_count);
	  }
	  scopes = Rest(scopes);
	  scope_count++;
     }

     /* make frame references as global */
     if (node->memorycheckedp)
       AnalyzeFrameReadWrite(node);
}

PRIVATE Node *ExpandTypedefs(Node *node)
{
     Node *junk;
     Node *new = NodeCopy(node, NodeOnlyExact);

#define CHILD(n) n = ExpandTypedefs(n)
     ASTWALK(new, CHILD)
#undef CHILD

     /* expand locally declared typedefs */
	 if (new->typ == Tdef) {
	  if (LookupSymbol(Identifiers, new->u.tdef.name, (GenericREF) & junk)) {
	       if (NodeStorageClass(junk) == T_TYPEDEF &&
		   junk->u.decl.type == new->u.tdef.type)
		    return new;	/* external declaration is OK */
	  }
	  /*
	   * external declaration is not the same as the local one - must
	   * expand the typedef. 
	   */
	  /* 
	   * tdef.type is not visited when doing an ASTWALK! could we
	   * get into circularity problems?  I don't think so.  -KHR 
	   */
	  new->u.tdef.type = ExpandTypedefs(new->u.tdef.type);

	  /*
	   * is this the right thing to do with tdef's type qualifiers? -
	   *  KHR 
	   */
	  NodeAddTq(new->u.tdef.type, new->u.tdef.tq);
	  return new->u.tdef.type;
     } else
	  return new;
}

PRIVATE Node *MakeFrame(procNode *u)
{
     List *scopes, *scope;
     int scope_count;
     List *fields;
     ListMarker marker1, marker2;
     Node *result;
     Node *type;

     /* 
      * make frame:
      *
      * typedef struct foo_frame {
      *   StackFrameT _header;
      *   { ... args ... } _scope0;
      *   { ... scope 1 vars ...} _scope1;
      *   ...
      *   { ... scope <n> vars ...} _scope<n>;
      * }
      *
      * for each different type of inlet, we also have a scope that
      * holds its argument list.
      * { ... args of inlet 1 ... } _scope<x>;
      * { ... args of inlet <n> ... } _scope<y>;
      *
      * The inlet scopes are mixed in with the regular scopes in no
      * particular order.  We could put all of the inlet argument
      * scopes in a oneof because only one is being used at a time.
      */
     fields = List1(MakeDecl(HeaderName,
			     EMPTY_TQ | T_SU_DECL,
			     MakeGenericFrameType(),
			     NULL,
			     NULL));

     /* if proc return type is not an arithmetic type, put it in frame */
     type = u->decl->u.decl.type->u.fdcl.returns;
     if (!IsVoidType(type) && !IsScalarType(type)) {
	  fields = AppendItem(fields,
			      MakeDecl(DummyReturnName,
				       EMPTY_TQ | T_SU_DECL,
				       NodeCopy(type, Subtree),
				       NULL,
				       NULL));
     }
     /* add scopes to frame */
     scopes = FindAllScopes(u->body, NULL);
     scopes = ConsItem(u->decl->u.decl.type->u.fdcl.args, scopes);

     scope_count = 0;
     IterateList(&marker1, scopes);
     while (NextOnList(&marker1, (GenericREF) & scope)) {
	  List *scopefields = NULL;
	  Node *decl;

	  IterateList(&marker2, scope);
	  while (NextOnList(&marker2, (GenericREF) & decl)) {
	       if (decl->typ != Decl)
		    continue;
	       if (decl->u.decl.kind == 0)
		    continue;
	       SetCurrentOnList(&marker2, NodeCopy(decl, Subtree));
	       NodeRemoveTq(decl, T_REGISTER);
	       NodeRemoveTq(decl->u.decl.type, T_CONST);
	       decl->u.decl.type = ExpandTypedefs(decl->u.decl.type);
#if 0
	       /* This isn't really working yet... -KHR */
	       /*
		* for array decls such as "int a[] = {1,2,3};", we must
		* calculate the implicit dimension of array a by counting
		* the number of initializers, and store this value in
		* the declaration in the frame.
		*/
	       if (decl->u.decl.type->typ == Adcl &&
		   decl->u.decl.init != NULL &&
		   decl->u.decl.type->u.adcl.dim->typ == ImplicitCast &&
		   decl->u.decl.type->u.adcl.dim->u.implicitcast.expr == NULL)
		 {
		   Node *init = decl->u.decl.init;
		   int dim;
		   assert(init->typ == Initializer);
		   dim = ListLength(init->u.initializer.exprs);
		   decl->u.decl.type->u.adcl.dim = MakeConstSint(dim);
		 }
#endif
	       decl->u.decl.init = NULL;
	       scopefields = AppendItem(scopefields, decl);
	  }

	  if (scopefields)
	       fields = AppendItem(fields,
				   MakeDecl(MakeScopeName(scope_count),
					    EMPTY_TQ | T_SU_DECL,
				    MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
				      make_SUE(Sdcl, NULL, scopefields)),
					    NULL,
					    NULL));
	  scope_count++;
     }

     /* finally, make struct declaration for frame */
     result = MakeDecl(NULL,
		       EMPTY_TQ | T_TOP_DECL,
		       MakeSdcl(EMPTY_TQ | T_SUE_ELABORATED,
	make_SUE(Sdcl, MakeFrameTypeName(u->decl->u.decl.name), fields)),
		       NULL,
		       NULL);
     SetCoords(result, UnknownCoord, Subtree);
     return result;
}

PRIVATE inline Node *MakeLinkage(Node *node)
{
     Node *linkage;

     /* make ProcInfo declaration */
     linkage = MakeDecl(MakeLinkName(proc_name),
			T_STATIC,
			MakeAdcl(EMPTY_TQ,
				 MakeTdef(EMPTY_TQ,
					  UniqueString("ProcInfo")),
				 NULL),
			MakeInitializer(link_initializers),
			NULL);
     SetCoords(linkage, UnknownCoord, Subtree);
     return (linkage);
}

/*
 * Add code to allocate frame to beginning of fast body.  For example:
 *    _foo_frame *_frame;
 *    _START_THREAD_FAST();
 *    _INIT_FRAME(_frame, sizeof(_foo_frame), _foo_sig);
 *    (_INIT_FRAME does, among other things,
 *           _frame->_header.size = sizeof(_foo_frame);
 *           _frame->_header.sig = _foo_sig; )
 *
 *    ... transfer non-register arguments to frame ...
 */
PRIVATE Node *AddFastStuff(Node *body, List *formals)
{
     List *stmts;
     ListMarker marker;
     Node *decl;

     stmts = List1(MakeCall(MakeId(StartThreadFastName),
			    NULL));
     stmts = AppendItem(stmts,
			MakeCall(MakeId(InitFrameName),
				 List3(MakeId(FrameName),
				       MakeUnary(SIZEOF,
					       MakeFrameType(proc_name)),
				  MakeId(MakeLinkName(proc_name)))));

     /* move any frame-only formals to frame */
     IterateList(&marker, formals);
     while (NextOnList(&marker, (GenericREF) & decl)) {
	  if (decl->u.decl.kind == 2) {
	       const char *name = decl->u.decl.name;
	       const char *scope = decl->u.decl.scope;

	       /* 
		* we need to do something different for array arguments
	        * -KHR 
		*/
	       stmts = AppendItem(stmts,
				  MakeBinop('=',
					    MakeBinop('.',
						      MakeBinop(ARROW,
						       MakeId(FrameName),
							  MakeId(scope)),
						      MakeId(name)),
					    MakeId(name)));
	  }
     }

     /* append code of procedure here */
     stmts = AppendItem(stmts, body);

     return MakeBlock(NULL,
		      List1(MakeDecl(FrameName,
				     EMPTY_TQ | T_BLOCK_DECL,
				     MakePtr(EMPTY_TQ,
					     MakeFrameType(proc_name)),
				     NULL,
				     NULL)),
		      stmts);
}

/* 
 * Add case statement and argument unpacking to beginning of slow body.
 * For example:
 *    int n;       < declaration for each register formal >
 *    _START_THREAD_SLOW();
 *    switch (_frame->_header.entry) {
 *      case 1: goto _sync1;
 *      case 2: goto _sync2;
 *      case 3: goto _sync3;
 *    }
 *    n = _frame->n;   < initialization for each register formal >
 */
Node *AddSlowStuff(Node *body, List *formals)
{
     List *cases = NULL;
     List *regdecls = NULL;
     ListMarker marker;
     Node *decl;

     /* make cases for case statement */
     for (; sync_count; sync_count--) {
	  cases = ConsItem(MakeCase(MakeConstSint(sync_count),
			     MakeGoto(MakeLabel(MakeSyncName(sync_count),
						NULL)),
				    NULL),
			   cases);
     }

     /* get declarations for register formals */
     IterateList(&marker, formals);
     while (NextOnList(&marker, (GenericREF) & decl)) {
	  if (decl->u.decl.kind == 1)
	       regdecls = AppendItem(regdecls, decl);
     }

     return MakeBlock(NULL,
		      regdecls,
		      ConsItem(
			     MakeCall(MakeId(StartThreadSlowName), NULL),
				   ConsItem(
						MakeSwitch(MakeBinop('.',
							 MakeBinop(ARROW,
						       MakeId(FrameName),
						     MakeId(HeaderName)),
						      MakeId(EntryName)),
					    MakeBlock(NULL, NULL, cases),
							   cases),
				   AppendItem(RestoreVariables(regdecls),
					      body))));
}

Node *GlobalizeSUE(SUEtype * type)
{
     Node *junk;
     assert(type->complete);

     /* choose a globally unique name for this SUE */
     if (LookupSymbol(Tags, type->name, (GenericREF) & junk)) {
	  char buf[40];
	  sprintf(buf, "%s%.16s", CilkPrefix, type->name);
	  type->name = InsertUniqueSymbol(Tags, type, buf);
     }
     switch (type->typ) {
	 case Sdcl:
	      return MakeSdcl(T_SUE_ELABORATED, type);
	 case Udcl:
	      return MakeUdcl(T_SUE_ELABORATED, type);
	 case Edcl:
	      return MakeEdcl(T_SUE_ELABORATED, type);
	 default:
	      FAIL("bad type in GlobalizeSUE");
	      return NULL;
     }
}

List *MoveSUEToTop(Node *node, List *sofar)
{
     Node *new = NULL;

#define CODE(n)   sofar = MoveSUEToTop(n, sofar)
     ASTWALK(node, CODE)
#undef CODE

	 switch (node->typ) {
	 case Sdcl:
	      if (SUE_ELABORATED(node->u.sdcl.tq)) {
		   new = GlobalizeSUE(node->u.sdcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Udcl:
	      if (SUE_ELABORATED(node->u.udcl.tq)) {
		   new = GlobalizeSUE(node->u.udcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Edcl:
	      if (SUE_ELABORATED(node->u.edcl.tq)) {
		   new = GlobalizeSUE(node->u.edcl.type);
		   NodeRemoveTq(node, T_SUE_ELABORATED);
	      }
	      break;
	 case Decl:
	      if (NodeStorageClass(node) == T_STATIC &&
		  NodeDeclLocation(node) == T_BLOCK_DECL) {
		   Node *junk;
		   /* choose a globally unique name for this static *
		    * variable */
		   if (LookupSymbol(Identifiers, node->u.decl.name, (GenericREF) & junk)) {
			node->u.decl.name = InsertUniqueSymbol(Identifiers, node, node->u.decl.name);
		   }
		   else {
		        InsertSymbol(Identifiers, node->u.decl.name, node, NULL);
		   }
		   new = node;
		   NodeSetDeclLocation(new, T_TOP_DECL);
		   new = TransformDecl(new, &new->u.decl);
		   
		   /* 
		    * Note: the static variables will be renamed (if necessary)
		    * by TransformId, and the local static declaration will be
		    * removed by TransformDecl.    -KHR
		    */
	      }
	      break;
	 default:
	      break;
     }

     if (new) {
	  /*
	   * make sure we expand any typedefs of declarations moved to 
	   * the top level.   -KHR 
	   */
	  new = ExpandTypedefs(new);
	  SetCoords(new, UnknownCoord, Subtree);
	  sofar = AppendItem(sofar, new);
     }
     return sofar;
}

PRIVATE inline List *TransformCilkProc(Node *node, procNode *u)
{
     Node *fastnode, *slownode;
     Node *decl;
     List *result;

     /* analyze procedure variables */
     AnalyzeVariables(node);

     /* start with empty text node - just enforces coord start */
     result = List1(MakeTextCoord("", FALSE, node->coord));

     /* move structures/unions/enums, static variables to outer scope */
     result = JoinLists(result, MoveSUEToTop(node, NULL));

     /* make frame declaration for procedure */
     result = AppendItem(result, MakeFrame(u));

     /* write down some global variables for other routines to read */
     proc_name = u->decl->u.decl.name;
     curr_proc = node;

     fastnode = NodeCopy(node, Subtree);
     slownode = NodeCopy(node, Subtree);

     /* add implicit sync and return to slow and fast bodies, if needed */
     if (node->u.proc.needs_sync) {
	  slownode->u.proc.body->u.Block.stmts =
	      AppendItem(slownode->u.proc.body->u.Block.stmts,
	      MakeSyncCoord(slownode->u.proc.body->u.Block.right_coord));
	  fastnode->u.proc.body->u.Block.stmts =
	      AppendItem(fastnode->u.proc.body->u.Block.stmts,
	      MakeSyncCoord(fastnode->u.proc.body->u.Block.right_coord));
     }
     if (node->u.proc.needs_return) {
	  slownode->u.proc.body->u.Block.stmts =
	      AppendItem(slownode->u.proc.body->u.Block.stmts,
			 MakeReturnCoord(NULL,
			    slownode->u.proc.body->u.Block.right_coord));
	  fastnode->u.proc.body->u.Block.stmts =
	      AppendItem(fastnode->u.proc.body->u.Block.stmts,
			 MakeReturnCoord(NULL,
			    fastnode->u.proc.body->u.Block.right_coord));
     }
     /* transform fast body */
     sync_count = 0;
     version = Fast;
     fastnode->u.proc.decl = TransformNode(fastnode->u.proc.decl);
     in_procedure = 1;
     fastnode->u.proc.body = TransformNode(fastnode->u.proc.body);
     in_procedure = 0;
     on_abort_defined = 0;

     /* transform slow body */
     sync_count = 0;
     version = Slow;
     /* add info for entry 0 */
     link_initializers = 
       List1(
	     MakeInitializer(
			     List5(
				   IsVoidType(u->decl->u.decl.type->u.fdcl.returns) ? 
				   MakeConstSint(0) : 
				   MakeUnary(SIZEOF, NodeCopy(u->decl->u.decl.type->u.fdcl.returns, Subtree)),
				   MakeUnary(SIZEOF, MakeFrameType(proc_name)),
				   MakeId(MakeSlowProcName(proc_name)),
				   MakeConstSint(0),
				   MakeConstSint(0)
				   )
			     )
	     );
     standalone_inlets = NULL;
     slownode->u.proc.decl = TransformNode(slownode->u.proc.decl);
     in_procedure = 1;
     slownode->u.proc.body = TransformNode(slownode->u.proc.body);
     in_procedure = 0;
     on_abort_defined = 0;

     /* add stuff to beginning of procedure */
     fastnode->u.proc.body = AddFastStuff(fastnode->u.proc.body,
			fastnode->u.proc.decl->u.decl.type->u.fdcl.args);
     slownode->u.proc.body = AddSlowStuff(slownode->u.proc.body,
			slownode->u.proc.decl->u.decl.type->u.fdcl.args);

     /* make slow decl */
     /* 
      * cilk int foo(int x, int y)  ==>  void _foo_slow(_foo_frame *_frame)
      */
     decl = MakeSlowDecl(proc_name);
     SetCoords(decl, slownode->coord, Subtree);
     slownode->u.proc.decl = decl;

     /* set fast decl flag */
     NodeAddTq(fastnode->u.proc.decl, T_FAST_PROCEDURE);

     /* make slow procedure declaration */
     result = AppendItem(result, 
			 MakeSlowDecl(u->decl->u.decl.name));

     /* add in standalone inlets */
     result = JoinLists(result, standalone_inlets);

     /* add linkage stuff */
     result = AppendItem(result, MakeLinkage(node));

     result = AppendItem(result, fastnode);
     result = AppendItem(result, slownode);

     return result;
}

PRIVATE Node *TransformNode(Node *node)
{
     if (node == NULL)
	  return NULL;
#define CODE(name, node, union) return Transform##name(node, union)
     ASTSWITCH(node, CODE)
#undef CODE
	 UNREACHABLE;
     return NULL;
}

GLOBAL List *TransformTopDecl(Node *node)
{
     List *result;

     if (!node)
	  return NULL;
     if (node->typ == Proc &&
	 node->u.proc.decl->u.decl.type->typ == Fdcl &&
	 node->u.proc.decl->u.decl.type->u.fdcl.tq & T_PROCEDURE) {
	  /* cilk procedure */
	  result = TransformCilkProc(node, &node->u.proc);
     } else
	  result = List1(TransformNode(node));

     return (result);
}

GLOBAL List *TransformPass1(List *program)
{
     ListMarker marker;
     Node *node;
     List *newprogram = NULL;

     IterateList(&marker, program);
     while (NextOnList(&marker, (GenericREF) & node)) {
	  newprogram = JoinLists(newprogram, TransformTopDecl(node));
     }

     return newprogram;
}

/*
 * TransformProgram should convert a type-checked source language tree
 * into a standard C tree.  Type information does not need to be preserved,
 * but all new Decl nodes must be properly annotated with their DECL_LOCATION
 * (e.g. top-level, block, structure field, etc.).
 */

GLOBAL List *TransformProgram(List *program)
{
     InitCilkTransform();
     program = TransformPass1(program);
     program = ConsItem(MakeText("#include <cilk-cilk2c.h>\n", TRUE),
			program);
     return program;
}

PRIVATE int operator_type_number(OpType assign_op, Node *type)
{
  unsigned int operator = 0;
  unsigned int type_num;

  /* commutative operators have same op_num */
  switch (assign_op) {
  case PLUSassign:
  case MINUSassign:
  case PREINC:
  case PREDEC:
  case POSTINC:
  case POSTDEC:
    operator = 0;
    break;
  case MULTassign:
    if (IsFloatingType(type))
      operator = 1;
    else
      operator = 2;
    break;
  case DIVassign:
    operator = 1;
    break;
  case ANDassign:
  case ERassign:
  case ORassign:
    operator = 2;
    break;
  case LSassign:
  case RSassign:
  case MODassign:
  case '=':
    operator = 3;
    break;
  default: 
    FAIL("bad operator in operator_number");
  }
  
  if (IsIntegralType(type))
    type_num = 0;
  else if (IsFloatingType(type))
    type_num = 1;
  else if (IsPointerType(type))
    type_num = 2;
  else
    type_num = 3;

  assert(operator < 4);
  assert(type_num < 4);
  return ((type_num << 2) + operator);
}

PRIVATE void ClearCoord(Node *node)
{
  node->coord = UnknownCoord;
#define CHILD(n) ClearCoord(n)
  ASTWALK(node, CHILD)
#undef CHILD
}  

PRIVATE Node *name_strip(Node *node)
{
  Node *anode;

#define CHILD(n)  n = name_strip(n)
  ASTWALK(node, CHILD)
#undef CHILD

  if ((node->typ == Unary) && (node->u.unary.op == INDIR)) {
    anode = node->u.unary.expr;
    if (anode->typ == Cast) {
      anode = anode->u.cast.expr;
      if ((anode->typ == Call) && (anode->u.call.name->typ == Id)) {
	if ((strcmp(anode->u.call.name->u.id.text, 
		    AccessReadProfileName) == 0) ||
	    (strcmp(anode->u.call.name->u.id.text, 
		    AccessReadProfileName4) == 0) ||
	    (strcmp(anode->u.call.name->u.id.text, 
		    AccessReadProfileName8) == 0) ||
	    (strcmp(anode->u.call.name->u.id.text, 
		    AccessWriteProfileName) == 0) ||
	    (strcmp(anode->u.call.name->u.id.text, 
		    AccessWriteProfileName4) == 0) ||
	    (strcmp(anode->u.call.name->u.id.text, 
		    AccessWriteProfileName8) == 0))
	  return FirstItem(anode->u.call.args);
	else if ((strcmp(anode->u.call.name->u.id.text, 
			 AccessReadProfileAddrName) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text,
			 AccessReadProfileAddrName4) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text,
			 AccessReadProfileAddrName8) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text, 
			 AccessWriteProfileAddrName) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text,
			 AccessWriteProfileAddrName4) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text,
			 AccessWriteProfileAddrName8) == 0) ||
		 (strcmp(anode->u.call.name->u.id.text,
			 SpawnVarProfileName) == 0))
	  return MakeUnary(INDIR, FirstItem(anode->u.call.args));
      }
    }
  }
  else if ((node->typ == Binop) && (node->u.binop.op == '.') &&
	   (node->u.binop.left->typ == Binop) && 
	   (node->u.binop.left->u.binop.op == ARROW) &&
	   (node->u.binop.left->u.binop.left->typ == Id) &&
	   (strcmp(node->u.binop.left->u.binop.left->u.id.text, "_frame") == 0))
    return node->u.binop.right;

  return node;
}

PRIVATE Node *AccessProfile(Node *node, Node *type, AccessProfileType accesstype, int withaddress, int opt, int multiple, int operator)
{
     Node *call = NULL, *atype, *op_type_node, *var_node;
     int size;
     /*
     if (!(node->memorycheckedp))
         return node;
     */
     if (!node)
         return node;

     /* for array, we are taking address of the array, no memory access.
	for example, int a[2], and any use of a is no access of memory */
     if (type->typ == Adcl)
       return node;

     /* if the real type of node is an array (though typedef), we convert
	this into a pointer type. 
	for example, typedef int aa[10]; typedef aa *bb; bb y; then *y.
	however, the cast type is unchanged, so *y=*((aa *)(y,sizeof(int *))).
	The sizeof is correct for function call with argument *y because
	we pass pointer rather than the whole array.
	But the sizeof would be wrong if *y is used elsewhere (I couldn't
	think of any other legal use (*y), so I leave it unconsidered */
     atype = NodeCopy(NodeDataType(type), Subtree);
     assert(atype);
     if (atype->typ == Adcl)
       atype = MakePtr(EMPTY_TQ, atype->u.adcl.type);
     if (atype->typ == Ptr) {
       assert(atype->u.ptr.type);
     }
     ClearCoord(atype);

     assert(type);
     if (type->typ == Ptr) {
       assert(type->u.ptr.type);
     }
     ClearCoord(type);

     size = NodeSizeof(node, atype) * multiple;
     op_type_node = MakeConstSint(operator_type_number(operator, atype));

     var_node = name_strip(NodeCopy(node, Subtree));

     switch (accesstype) {
     case Read:
       if (size == sizeof(int)) {
	 call = MakeId(withaddress ? 
		       AccessReadProfileAddrName4 :
		       AccessReadProfileName4);
	 call = MakeCall(call, List3(node, op_type_node, var_node));
       }
       else if (size == sizeof(double)) {
	 call = MakeId(withaddress ?
		       AccessReadProfileAddrName8 :
		       AccessReadProfileName8);
	 call = MakeCall(call, List3(node, op_type_node, var_node));
       }
       else {
	 call = MakeId(withaddress ?
		       AccessReadProfileAddrName :
		       AccessReadProfileName);
	 call = MakeCall(call, List4(node, MakeConstSint(size), op_type_node,
				     var_node));
       }
       break;
     case Write:
       if (size == sizeof(int)) {
	 call = MakeId(withaddress ? 
		       AccessWriteProfileAddrName4 :
		       AccessWriteProfileName4);
	 call = MakeCall(call, List3(node, op_type_node, var_node));
       }
       else if (size == sizeof(double)) {
	 call = MakeId(withaddress ?
		       AccessWriteProfileAddrName8 :
		       AccessWriteProfileName8);
	 call = MakeCall(call, List3(node, op_type_node, var_node));
       }
       else {
	 call = MakeId(withaddress ?
		       AccessWriteProfileAddrName :
		       AccessWriteProfileName);
	 call = MakeCall(call, List4(node, MakeConstSint(size), op_type_node,
				     var_node));
       }
       break;
     default:
       FAIL("bad access type in AccessProfile");
     }

     if (opt == 0) {
       call = MakeCast(MakePtr(EMPTY_TQ, type), call);
       call->parenthesized = 1;
       if (!withaddress) {
	 call = MakeUnary(INDIR, call);
	 call->u.unary.type = type;
	 call->parenthesized = 1;
       }
       call->coord = node->coord;
     }
     return call;
}

PRIVATE Node *SpawnVarProfile(Node *node, Node *type, OpType assign_op)
{
     Node *var_node;

     /*
     if (node->memorycheckedp)
         return node;
     */

     if (!node)
         return node;

     assert(type);
     type = NodeDataType(type);
     assert(type);

     var_node = name_strip(NodeCopy(node, Subtree));

     return MakeCall(MakeId(SpawnVarProfileName),
                     List5(node,
                           MakeConstSint(NodeSizeof(node, type)),
			   MakeConstSint(
			     operator_type_number(assign_op, type)),
			   MakeCast(MakePtr(EMPTY_TQ,
					    MakeGenericFrameType()),
				    MakeId(FrameName)),
			   var_node));
}

/*
 * similar routine as AnalyzeGlobalReadWrite in analyze.c.
 * use bottom-up tree-search to find all frame reference.
 * called in AnalyzeVariables after the kind of Id is analyzed.
 * node->nondeterminator is TRUE if node potentially accesses shared location.
 */
PRIVATE void AnalyzeFrameReadWrite(Node *node)
{
  switch (node->typ) {
  case Id:
    if (node->u.id.decl && ((node->u.id.decl->u.decl.kind == 2) || 
			    (node->u.id.decl->u.decl.kind == 3)))
      node->nondeterminator = TRUE;
    break;
  default:
    break;
  }

#define CODE(n)   AnalyzeFrameReadWrite(n)
  ASTWALK(node, CODE)
#undef CODE
    
  switch (node->typ) {
  case Array:
    if (node->u.array.name->nondeterminator) {
      node->nondeterminator = TRUE;
      node->u.array.name->nondeterminator = FALSE;
    }
    break;
  case Binop:
    if (IsAssignmentOp(node->u.binop.op) && 
	node->u.binop.left->nondeterminator) {
      node->nondeterminator = TRUE;
      node->u.binop.left->nondeterminator = FALSE;
    }
    else if ((node->u.binop.op == '.') && 
	     node->u.binop.left->nondeterminator) {
      node->nondeterminator = TRUE;
      node->u.binop.left->nondeterminator = FALSE;
    }
    break;
  case Unary:
    if (IsIncDecOp(node->u.unary.op) && node->u.unary.expr->nondeterminator) {
      /*
      if ((node->u.unary.expr->typ == Id) &&
	  (node->u.unary.expr->u.id.decl->u.decl.kind == 3))
	node->u.unary.expr->nondeterminator = FALSE;
      */
      node->nondeterminator = TRUE;
      node->u.unary.expr->nondeterminator = FALSE;
    }
    else if (node->u.unary.op == ADDRESS) {
      node->nondeterminator = FALSE;
      node->u.unary.expr->nondeterminator = FALSE;
    }
    break;
  case Spawn:
    if (node->u.spawn.receiver && node->u.spawn.receiver->nondeterminator) {
      node->u.spawn.receiver->nondeterminator = FALSE;
    }
    break;
  case InletCall:
    if (node->u.inletcall.receiver && 
	node->u.inletcall.receiver->nondeterminator) {
      node->u.inletcall.receiver->nondeterminator = FALSE;
    }
    break;
  default:
    break;
  }
}
