/*
 * $Id: cilk-nd-cilk2c.h,v 1.5 1998/07/15 09:31:38 fengmd Exp $
 */

/*
 * $Log: cilk-nd-cilk2c.h,v $
 * Revision 1.5  1998/07/15 09:31:38  fengmd
 * Support user-inlet and library, merge changes from the main branch.
 *
 * Revision 1.4  1998/06/11 22:42:38  fengmd
 * Stripped version of Nondeterminator, without including cilk.c, scheduler etc.
 *
 * Revision 1.3  1998/06/11 13:18:58  fengmd
 * Report without-lock name along with the without-lock address.
 *
 * Revision 1.2  1998/05/31 13:53:53  fengmd
 * New run-time options to play: -all-sets, -umbrella
 *
 * Revision 1.1.1.1  1998/03/19 04:43:30  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.2.4  1997/09/05 05:16:01  fengmd
 * FREE_ID changes to 2.  Fix cilk2c for Cilk_free var problem.
 *
 * Revision 1.1.2.3  1997/08/29 07:38:37  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:40  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifdef CILK_ND

/*
 * checking protype, called from cilk2c'ed Cilk program
 */
#define _READ_CHECK(p, size, op_type, v) \
  Cilk_internal_read_check(p, size, op_type, __LINE__, __FILE__, #v)
#define _READ_CHECK4(p, op_type, v) \
  Cilk_internal_read_check4(p, op_type, __LINE__, __FILE__, #v)
#define _READ_CHECK8(p, op_type, v) \
  Cilk_internal_read_check8(p, op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK(p, size, op_type, v) \
  Cilk_internal_write_check(p, size, op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK4(p, op_type, v) \
  Cilk_internal_write_check4(p, op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK8(p, op_type, v) \
  Cilk_internal_write_check8(p, op_type, __LINE__, __FILE__, #v)

#define _READ_CHECK_VALUE(p, size, op_type, v) \
  Cilk_internal_read_check(&(p), size, op_type, __LINE__, __FILE__, #v)
#define _READ_CHECK_VALUE4(p, op_type, v) \
  Cilk_internal_read_check4(&(p), op_type, __LINE__, __FILE__, #v)
#define _READ_CHECK_VALUE8(p, op_type, v) \
  Cilk_internal_read_check8(&(p), op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK_VALUE(p, size, op_type, v) \
  Cilk_internal_write_check(&(p), size, op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK_VALUE4(p, op_type, v) \
  Cilk_internal_write_check4(&(p), op_type, __LINE__, __FILE__, #v)
#define _WRITE_CHECK_VALUE8(p, op_type, v) \
  Cilk_internal_write_check8(&(p), op_type, __LINE__, __FILE__, #v)

#define _SPAWNVAR_CHECK(p, size, op_type, f, v) \
  Cilk_internal_spawnvar_check(&(p), size, op_type, f, __LINE__, __FILE__, #v)

#define _MALLOC_CHECK(p)      Cilk_nd_malloc(p, __LINE__, __FILE__, "malloc")
#define _FREE_CHECK(p, v)     Cilk_nd_free(p, __LINE__, __FILE__, #v)
#define _MEMALIGN_CHECK(p, s) Cilk_nd_memalign(p, s, __LINE__, __FILE__, "memalign")
#define _CALLOC_CHECK(p, s)   Cilk_nd_calloc(p, s, __LINE__, __FILE__, "calloc")
#define _REALLOC_CHECK(p, s)  Cilk_nd_realloc(p, s, __LINE__, __FILE__, "realloc")
#define _VALLOC_CHECK(p)      Cilk_nd_valloc(p, __LINE__, __FILE__, "valloc")

#define _LOCK_CHECK(p, v)     Cilk_nd_lock(p, #v)
#define _UNLOCK_CHECK(p)      Cilk_nd_unlock(p)

/*
 * Stuffs stolen from cilk-cilk2c.h, simplified
 */

#define _PUSH_FRAME(f)

#define _INIT_FRAME(f,s,signat) { \
     volatile StackFrame **_t; \
     Cilk_alloc_proc_id_spbags(); \
     f = Cilk_create_frame(s, signat); \
     _t = USE_PRIVATE(cache)->tail; \
     CILK_COMPLAIN((StackFrame **)_t < USE_PRIVATE(cache)->stack + \
		   USE_SHARED_RO(stackdepth),\
		   (USE_SHARED_RO(stack_overflow_msg)));\
     *_t = (StackFrame *)f; \
     USE_PRIVATE(cache)->tail = _t + 1;\
     Cilk_init_spbags((StackFrame *)f); \
}

#define _SYNC 1

#define _SYNCHED 0

#define _ABORT_STANDALONE()

#define _ABORT_SLOW()

#define _SET_RESULT(x)

#define _SET_NORESULT()

#define _XPOP_FRAME_RESULT(f, r, result)

#define _XPOP_FRAME_NORESULT(f, r) \
{ \
     volatile StackFrame **_t;\
     _t = USE_PRIVATE(cache)->tail; \
     Cilk_lreturn_spbags((StackFrame *)f, *_t); \
}

#define _BEFORE_SPAWN_FAST() \
     WHEN_CILK_DEBUG(if (USE_SHARED_RO(num_current_locks) != 0) \
		        Cilk_locked_spawn_warn();)

#define _BEFORE_SPAWN_SLOW() \
     WHEN_CILK_DEBUG(if (USE_SHARED_RO(num_current_locks) != 0) \
		        Cilk_locked_spawn_warn();)

#define _AT_THREAD_BOUNDARY_SLOW()

#define _AFTER_SPAWN_FAST()

#define _AFTER_SPAWN_SLOW()

#define _START_THREAD_FAST()

#define _START_THREAD_SLOW()

#define _BEFORE_RETURN_FAST() \
      Cilk_destroy_frame((StackFrame *)_frame, sizeof(*_frame)); \
      --USE_PRIVATE(cache)->tail;

#define _BEFORE_RETURN_SLOW() \
      Cilk_destroy_frame((StackFrame *)_frame, sizeof(*_frame)); \
      --USE_PRIVATE(cache)->tail;

#define _BEFORE_RETURN_INLET()

#define _BEFORE_SYNC_SLOW()

#define _AFTER_SYNC_SLOW()

#define _AT_SYNC_FAST() \
      Cilk_sync_spbags((StackFrame *)_frame);

#define _ND_ENTER_INLET(f) Cilk_enter_inlet(f);

#define _ND_LEAVE_INLET() Cilk_leave_inlet();

#define _ND_LRETURN_SPBAGS(f) \
{ \
     volatile StackFrame **_t;\
     _t = USE_PRIVATE(cache)->tail; \
     Cilk_lreturn_spbags((StackFrame *)f, *_t); \
}

#endif
