/*
 * $Id: cilk-nd.h,v 1.17 1998/07/16 05:56:37 fengmd Exp $
 */

/*
 * $Log: cilk-nd.h,v $
 * Revision 1.17  1998/07/16 05:56:37  fengmd
 * Move fake_lock and publish from cilk.h to here.
 *
 * Revision 1.16  1998/07/15 09:31:39  fengmd
 * Support user-inlet and library, merge changes from the main branch.
 *
 * Revision 1.15  1998/07/14 19:37:21  astark
 * PUBLISH changes, recycle_mem option
 *
 * Revision 1.14  1998/07/06 21:52:35  astark
 * changed reporting of umbrella violations
 * report current access plus one for each lock held, using current lock name
 *
 * Revision 1.13  1998/06/18 20:30:18  astark
 * removed -nd-heap option, grow disjoint set automatically
 *
 * Revision 1.12  1998/06/12 17:44:53  fengmd
 * Include Ien's heuristics.
 *
 * Revision 1.11  1998/06/11 13:18:59  fengmd
 * Report without-lock name along with the without-lock address.
 *
 * Revision 1.10  1998/06/03 02:10:37  fengmd
 * Reduce the default nd-heap size.
 *
 * Revision 1.9  1998/06/02 21:40:08  athena
 * Minor fix
 *
 * Revision 1.8  1998/06/01 14:35:20  fengmd
 * Incorporate the change of Cilk_lockvar and Cilk_lockvar_basetype.
 *
 * Revision 1.7  1998/05/31 13:53:54  fengmd
 * New run-time options to play: -all-sets, -umbrella
 *
 * Revision 1.6  1998/05/29 21:40:46  astark
 * added publish functionality
 *
 * Revision 1.5  1998/03/20 05:26:30  astark
 * added missing extern
 *
 * Revision 1.4  1998/03/20 03:36:06  astark
 * rewrote ALL_SETS free_check
 * changed UMBRELLA to use current_locks list instead of current_lockset array
 *
 * Revision 1.3  1998/03/19 23:11:33  astark
 * locked_spawn check
 *
 * Revision 1.2  1998/03/19 20:07:33  astark
 * fixed SET_LIVE
 *
 * Revision 1.1.1.1  1998/03/19 04:43:30  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.2.14  1997/09/05 05:16:03  fengmd
 * FREE_ID changes to 2.  Fix cilk2c for Cilk_free var problem.
 *
 * Revision 1.1.2.13  1997/09/04 23:17:45  astark
 * Fixed commutativity bug by introducing NO_INLET_ID
 *
 * Revision 1.1.2.12  1997/09/03 00:56:21  astark
 * Fixed race hashing to use filename and line num as well as var
 *
 * Revision 1.1.2.11  1997/09/02 11:34:56  fengmd
 * ND_VERBOSE defaults to 0.
 *
 * Revision 1.1.2.10  1997/08/29 07:38:38  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.2.9  1997/06/06 16:25:53  fengmd
 * Add the Linux port of the Nondeterminator.
 *
 * Revision 1.1.2.8  1997/03/21 22:29:14  fengmd
 * Change name from spbag to spbags.
 *
 * Revision 1.1.2.7  1997/03/13 18:48:07  fengmd
 * Declare pointer constant type to avoid gcc complaining.
 *
 * Revision 1.1.2.6  1997/03/04 00:17:21  fengmd
 * ND_VERBOSE defaults to 0.
 *
 * Revision 1.1.2.5  1997/03/04 00:06:58  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.4  1997/03/03 21:02:11  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:32  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:41  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * macros to define the address range of the static region
 */
#if CILK_SOLARIS
extern int _begin_Cilk_shared;
extern int _end_Cilk_shared;
#define STATIC_ARENA_START      &_begin_Cilk_shared
#define STATIC_ARENA_END        &_end_Cilk_shared

#elif CILK_SOLARIS_THREADS
extern int _etext;
extern int _end;
#define STATIC_ARENA_START      &_etext
#define STATIC_ARENA_END        &_end

#elif CILK_IRIX_THREADS
extern _fdata;
extern _end;
#define STATIC_ARENA_START      &_fdata
#define STATIC_ARENA_END        &_end

#elif CILK_LINUX
extern int _begin_Cilk_shared;
extern int _end_Cilk_shared;
#define STATIC_ARENA_START      &_begin_Cilk_shared
#define STATIC_ARENA_END        &_end_Cilk_shared
#ifndef MAP_FAILED
#define MAP_FAILED		((void *) -1)
#endif

#else
#warning The Nondeterminator has only been tested on Solaris, IRIX and LINUX
#warning Please define STATIC_ARENA_START and STATIC_ARENA_END in cilk-nd.h
#endif

/* 
 * macro to display the shadow address range
 */
#define ND_VERBOSE 0

/* 
 * macros to get shadow addresses
 */
#if (CILK_SOLARIS || CILK_SOLARIS_THREADS || CILK_LINUX)

/* for UMBRELLA */
#define STATIC_LOCKS(x) \
  ((ManyLocksT **) (((unsigned int) (x)) + USE_SHARED_RO(static_locks_offset)))
#define HEAP_LOCKS(x) \
  ((ManyLocksT **) (((unsigned int) (x)) - USE_SHARED_RO(heap_size)))
#define STATIC_ACCESSOR(x) \
  ((DisjointSetMemberT *) (((unsigned int) (x)) + USE_SHARED_RO(static_accessor_offset)))
#define HEAP_ACCESSOR(x) \
  ((DisjointSetMemberT *) (((unsigned int) (x)) - USE_SHARED_RO(both_plus_heap_size)))

/* for ALL_SETS */
#define STATIC_LOCKLISTS(x) \
  ((LockSetListT **) (((unsigned int) (x)) + USE_SHARED_RO(static_locks_offset)))
#define HEAP_LOCKLISTS(x) \
  ((LockSetListT **) (((unsigned int) (x)) - USE_SHARED_RO(heap_size)))

#elif CILK_IRIX_THREADS

#define SHADOW_SPACE_SHIFT 0x30000000

/* for UMBRELLA */
#define STATIC_LOCKS(x) \
  ((ManyLocksT **) (((unsigned int) (x)) + USE_SHARED_RO(static_locks_offset) + SHADOW_SPACE_SHIFT))
#define HEAP_LOCKS(x) \
  ((ManyLocksT **) (((unsigned int) (x)) - USE_SHARED_RO(heap_size) + SHADOW_SPACE_SHIFT))
#define STATIC_ACCESSOR(x) \
  ((DisjointSetMemberT *) (((unsigned int) (x)) + USE_SHARED_RO(static_accessor_offset) + SHADOW_SPACE_SHIFT))
#define HEAP_ACCESSOR(x) \
  ((DisjointSetMemberT *) (((unsigned int) (x)) - USE_SHARED_RO(both_plus_heap_size) + SHADOW_SPACE_SHIFT))

/* for ALL_SETS */
#define STATIC_LOCKLISTS(x) \
  ((LockSetListT **) (((unsigned int) (x)) + USE_SHARED_RO(static_locks_offset) + SHADOW_SPACE_SHIFT))
#define HEAP_LOCKLISTS(x) \
  ((LockSetListT **) (((unsigned int) (x)) - USE_SHARED_RO(heap_size) + SHADOW_SPACE_SHIFT))
#endif

/* 
 * macro to get shadow address of var and info
 * from shadow address of reader and writer
 */
#define SHADOW_VAR(p)       \
 ((char **) (((unsigned int) (p)) - USE_SHARED_RO(both_size2)))
#define SHADOW_INFO(x)      \
 ((LineAndFileT *) (((unsigned int) (x)) - USE_SHARED_RO(both_size)))

/*
 * macro to extract pieces of shadow info
 *
 * representation of LiveAndWithoutT:
 *   <without func> <live bit>
 * representation of LineAndFileT;
 *   <line num -- 24 bits> <file id -- 8 bits>
 */
#define GET_LIVE(x) \
     ((char) (((unsigned int) (x)) & 1))
#define GET_WITHOUT_FUNC(x) \
     ((DisjointSetMemberT) (((unsigned int) (x)) >> 1))
#define SET_LIVE(law, l) \
     ((LiveAndWithoutT) ((l) ? (((unsigned int) (law)) | 0x00000001) : \
                               (((unsigned int) (law)) & 0xFFFFFFFE)))
#define SET_WITHOUT_FUNC(law, wf) \
     ((LiveAndWithoutT) ((((unsigned int) (wf)) << 1) | \
                         (((unsigned int) (law)) & 1)))
#define MAKE_LIVE_WITHOUT(l, wf) \
     ((LiveAndWithoutT) ((((unsigned int) (wf)) << 1) | (l)))
#define GET_LINE(x) \
     (((unsigned int) (x)) >> 8)
#define GET_FILE(x) \
     (((unsigned int) (x)) & 0x000000FF)
#define MAKE_LINE_FILE(l,f) \
     ((LineAndFileT) ((((unsigned int) (l)) << 8) | ((unsigned int) (f))))

#define RLOCK     ((Cilk_lockvar_basetype *) 1)
#define NO_LOCK   ((Cilk_lockvar_basetype *) NULL)
#define PLUS_LOCK ((Cilk_lockvar_basetype *) 2)
#define AND_LOCK  ((Cilk_lockvar_basetype *) 3)
#define DIV_LOCK  ((Cilk_lockvar_basetype *) 4)
#define USER_INLET_LOCK ((Cilk_lockvar_basetype *) 5)

#define FAKE_LOCK_NAME(x) ((x == RLOCK) ? "fake_read_lock" : \
			   ((x == NO_LOCK) ? "empty_lock" : \
			    ((x == PLUS_LOCK) ? "fake_plus_lock" : \
			     ((x == AND_LOCK) ? "fake_and_lock" : \
			      ((x == DIV_LOCK) ? "fake_div_lock" : \
			       ((x == USER_INLET_LOCK) ? "fake_inlet_lock" : \
				(Cilk_die("unknown lock for FAKE_LOCK_NAME"), \
				 "")))))))

/* 
 * SP-bags constants
 */
#define S_SET        ((DisjointSetMemberT) 1)
#define P_SET        ((DisjointSetMemberT) 0)
#define FREE_ID      ((DisjointSetMemberT) 0)
#define NO_INLET_ID  ((DisjointSetMemberT) 2) /* must be distinct from S_SET and FREE_ID */
#define FIRST_THREAD ((DisjointSetMemberT) 3)

/* 
 * the least significant 28 bits are used for procedure id, the high-order
 * 4 bits are used for storing operator and type 
 */
#define PROC_ID_BIT     28
#define MAX_PROC_NUMBER (1UL<<PROC_ID_BIT)
#define PROC_ID(x)      (((unsigned int) x) & (MAX_PROC_NUMBER - 1))
#define OP_TYPE(x)      \
  ((((unsigned int) x) & (~(MAX_PROC_NUMBER - 1))) >> PROC_ID_BIT)
#define OP_TYPE_PROC_ID(proc_id, op_type) ((op_type << PROC_ID_BIT) + proc_id)
#define DEFAULT_ND_HEAPSIZE (1UL<<16)  /* number of IDs initially allocated for
					  disjoint set structures */

/* 
 * operator takes 2 lower-order bits, and the high-order 2 bits for type 
 */
#define OPERATOR_BIT      2
#define MAX_OPERATOR      (1UL<<OPERATOR_BIT)
#define CHECK_OPERATOR(x) ((x) & (MAX_OPERATOR - 1))
#define TYPE_BIT          2
#define CHECK_TYPE(x)     (((x) & (~(MAX_OPERATOR - 1))) >> OPERATOR_BIT)

/* 
 * operator value, from 0 to 2^OPERATOR_BIT - 1
 */
#define PLUSassign 0
#define DIVassign  1
#define ANDassign  2
#define EQassign   3

/* 
 * type value, from 0 to 2^TYPE_BIT - 1
 */
#define INTtype    0
#define FLOATtype  1
#define POINTtype  2
#define OTHERtype  3

#define FREEOPTYPE ((INTtype << OPERATOR_BIT) || EQassign)

typedef unsigned int LineAndFileT;
typedef unsigned int LiveAndWithoutT;
     
/* 
 * structure to store list of file name 
 */
typedef struct FileListT { 
  struct FileListT *next; 
  char *filename; 
  int index;
} FileListT;

/* 
 * indices to use for initial condition versus file name 
 */
#ifdef CILK_ND
#define INIT_INDEX 0
#define START_FILE_INDEX 1
#endif

/* 
 * structure for race hash table 
 */
typedef struct RaceHashT {
  struct RaceHashT *next;
  char *current_name, *shadow_name;
  int current_info, shadow_info;
} RaceHashT;

#define RACE_HASH_SIZE 10000

/*
 * structures for locks shadow space in UMBRELLA algorithm
 */
typedef struct LockInfoT
{
  Cilk_lockvar_basetype *plock;
  LiveAndWithoutT lwithout;
  LineAndFileT linefile;
  char *pname;
} LockInfoT;

typedef struct ManyLocksT
{
  int num_lockinfos;
  LockInfoT *plockinfo;
  char race_reported;
} ManyLocksT;

/*
 * structures for locks shadow space in ALL_SETS algorithm
 */
#define LOCK_HASH_SIZE ((1U * MEGABYTE) - 1)

typedef unsigned int HashT;

typedef struct LockSetT {
  struct LockSetT *next;
  struct LockSetT *prev;
  Cilk_lockvar_basetype *lock;
  char *name;
} LockSetT;

typedef struct LockSetHashT {
  unsigned int size;
  Cilk_lockvar_basetype *locks[1];
} LockSetHashT;

typedef struct LockSetListT {
  struct LockSetListT *next;
  HashT hash_index;
  DisjointSetMemberT proc_id;
  LineAndFileT linefile;
  char *pname;
} LockSetListT;

/*
 * variables defined in nondeterminator.c, and used by Cilk rts
 */
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_end, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, heap_size, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, both_size, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, both_size2, );
DECLARE_SHARED_RO_VARIABLE(extern size_t, both_plus_heap_size, );
DECLARE_SHARED_RO_VARIABLE(extern int, first_stop, );
DECLARE_SHARED_RO_VARIABLE(extern int, float_comm, );
DECLARE_SHARED_RO_VARIABLE(extern int, recycle_mem, );
DECLARE_SHARED_RO_VARIABLE(extern int, nd_arena, );
DECLARE_SHARED_RO_VARIABLE(extern int, all_sets, );
DECLARE_SHARED_RO_VARIABLE(extern int, umbrella, );
DECLARE_SHARED_RO_VARIABLE(extern DisjointSetMemberT, current_proc_id, );
DECLARE_SHARED_RO_VARIABLE(extern int, num_current_locks, );
DECLARE_SHARED_RO_VARIABLE(extern LockSetT, current_locks, );
DECLARE_SHARED_RO_VARIABLE(extern int, race_linenum, );
DECLARE_SHARED_RO_VARIABLE(extern char*, race_filename, );
DECLARE_SHARED_RO_VARIABLE(extern char*, race_name, );

/* 
 * SP-bags action functions, called from cilk2c'ed Cilk program or rts
 */
extern void Cilk_init_spbags(StackFrame *f);
extern void Cilk_lreturn_spbags(StackFrame *f,volatile StackFrame *child);
extern void Cilk_sync_spbags(StackFrame *f);
extern void Cilk_alloc_proc_id_spbags();
extern void Cilk_enter_inlet(StackFrame *f);
extern void Cilk_leave_inlet();
extern inline void DestroyManyLocks(ManyLocksT *manyLocksp);
extern inline void DestroyLockInfoTs(int old_num_lockinfos, LockInfoT *oldLockInfop);
extern inline void inc_ref_count(DisjointSetMemberT proc_id);
extern inline void dec_ref_count(DisjointSetMemberT proc_id);
extern inline void update_ref_count(DisjointSetMemberT old_proc_id,
				    DisjointSetMemberT new_proc_id);
extern void Cilk_locked_spawn_warn(void);
extern inline void Cilk_publish(void *p, int sz);
extern inline void save_shadow_info(DisjointSetMemberT *race_party);
extern inline int find_filename(char *filename);

/*
 * checking protype, called with gcc on cilk2c'ed Cilk program
 */
extern volatile void *Cilk_internal_read_check(volatile void *p, size_t size, 
			 int op_type, int linenum, char *filename, char *name);
extern volatile void *Cilk_internal_read_check4(volatile void *p, int op_type, 
			 int linenum, char *filename, char *name);
extern volatile void *Cilk_internal_read_check8(volatile void *p, int op_type, 
			 int linenum, char *filename, char *name);
extern volatile void *Cilk_internal_write_check(volatile void *p, size_t size, 
			 int op_type, int linenum, char *filename, char *name);
extern volatile void *Cilk_internal_write_check4(volatile void *p, int op_type,
			 int linenum, char *filename, char *name);
extern volatile void *Cilk_internal_write_check8(volatile void *p, int op_type,
			 int linenum, char *filename, char *name);
extern void Cilk_internal_spawnvar_check(volatile void *p, size_t size, 
			 int op_type, StackFrame *f, int linenum,
			 char *filename, char *name);
extern volatile void *Cilk_internal_free_check(volatile void *p, size_t size, 
			 int op_type, int linenum, char *filename, char *name);

extern void *Cilk_nd_malloc(size_t size, int linenum, char *filename, 
			    char *name);
extern void Cilk_nd_free(void *p, int linenum, char *filename, 
			 char *name);
extern void *Cilk_nd_memalign(size_t alignment, size_t size, int linenum, 
		       char *filename, char *name);
extern void *Cilk_nd_calloc(size_t nelem, size_t elsize, int linenum, 
			    char *filename, char *name);
extern void *Cilk_nd_realloc(void *s, size_t size, int linenum, char *filename,
		      char *name);
extern void *Cilk_nd_valloc(size_t size, int linenum, char *filename, 
			    char *name);

/*
 * checking protype called in Cilk program, transformed in cilk2c.
 * no need to implement.
 */
#ifdef CILK_ND
extern void Cilk_read_check(void *p, size_t size);
extern void Cilk_write_check(void *p, size_t size);
inline void Cilk_nd_lock(Cilk_lockvar_basetype *lock, char *name);
inline void Cilk_nd_unlock(Cilk_lockvar_basetype *lock);
#else
#define Cilk_read_check(p, size)
#define Cilk_write_check(p, size)
#endif

inline void save_info_var(LockSetListT *l);
HashT locks_to_hash_index();
void clean_shadow_space_all_sets(LockSetListT **accessor_lock, int size);

#ifdef CILK_ND
/* maximum number of locks that may be held at any one time */
#define CILK_MAX_LOCKS 100     
#define Cilk_fake_lock Cilk_lock
#define Cilk_fake_unlock Cilk_unlock
#define CILK_DECLARE_FAKE_LOCKVAR(a) Cilk_lockvar a
#define PUBLISH(p)            Cilk_publish(p, sizeof(*(p)))
#define PUBLISH_ARRAY(p, sz)  Cilk_publish(p, sizeof(*(p)) * (sz))
/* 
 * current lockset for umbrella algorithm.  Modified by Cilk_lock, Cilk_unlock
 */
DECLARE_SHARED_RO_VARIABLE(extern int, num_current_locks, );
#else
#define Cilk_fake_lock(a)
#define Cilk_fake_unlock(a)
#define CILK_DECLARE_FAKE_LOCKVAR(a) 
#define PUBLISH(p)
#define PUBLISH_ARRAY(p, sz) 
#endif
