/*
 * $Id: nondeterminator.c,v 1.22 1998/07/17 05:32:52 fengmd Exp $
 */

/*
 * $Log: nondeterminator.c,v $
 * Revision 1.22  1998/07/17 05:32:52  fengmd
 * The option -nproc is ineffective for the Nondeterminator.
 *
 * Revision 1.21  1998/07/15 09:31:41  fengmd
 * Support user-inlet and library, merge changes from the main branch.
 *
 * Revision 1.20  1998/07/14 19:37:23  astark
 * PUBLISH changes, recycle_mem option
 *
 * Revision 1.19  1998/07/06 21:52:37  astark
 * changed reporting of umbrella violations
 * report current access plus one for each lock held, using current lock name
 *
 * Revision 1.18  1998/06/18 20:30:19  astark
 * removed -nd-heap option, grow disjoint set automatically
 *
 * Revision 1.17  1998/06/17 20:37:04  fengmd
 * Merge new changes from the main branch, support thread version.
 *
 * Revision 1.16  1998/06/12 17:44:55  fengmd
 * Include Ien's heuristics.
 *
 * Revision 1.15  1998/06/12 05:47:30  fengmd
 * Remove spawnvar_check_*.
 *
 * Revision 1.14  1998/06/11 22:42:40  fengmd
 * Stripped version of Nondeterminator, without including cilk.c, scheduler etc.
 *
 * Revision 1.13  1998/06/11 13:19:01  fengmd
 * Report without-lock name along with the without-lock address.
 *
 * Revision 1.12  1998/06/01 14:35:22  fengmd
 * Incorporate the change of Cilk_lockvar and Cilk_lockvar_basetype.
 *
 * Revision 1.11  1998/05/31 13:53:55  fengmd
 * New run-time options to play: -all-sets, -umbrella
 *
 * Revision 1.10  1998/05/29 21:40:47  astark
 * added publish functionality
 *
 * Revision 1.9  1998/04/20 18:38:21  astark
 * changed malloc/free in C file to say "Unknown C file"
 *
 * Revision 1.7  1998/03/31 10:40:51  fengmd
 * Improve ALL_SETS.
 *
 * Revision 1.6  1998/03/20 05:27:35  astark
 * added out of memory checks at all mallocs
 * removed repeated freeing and reallocation of header struct for locks shadow
 * space in umbrella
 *
 * Revision 1.5  1998/03/20 03:45:28  astark
 * free/free condition assert
 *
 * Revision 1.4  1998/03/20 03:36:06  astark
 * rewrote ALL_SETS free_check
 * changed UMBRELLA to use current_locks list instead of current_lockset array
 *
 * Revision 1.3  1998/03/19 23:11:19  astark
 * locked_spawn check
 *
 * Revision 1.2  1998/03/19 09:12:06  fengmd
 * Change back the way to handle implicit inlet via lock.
 *
 * Revision 1.1.1.1  1998/03/19 04:43:30  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.2.14  1997/09/05 05:16:04  fengmd
 * FREE_ID changes to 2.  Fix cilk2c for Cilk_free var problem.
 *
 * Revision 1.1.2.13  1997/09/04 23:24:28  astark
 * Fixed typo in internal error message
 *
 * Revision 1.1.2.12  1997/09/04 23:17:47  astark
 * Fixed commutativity bug by introducing NO_INLET_ID
 *
 * Revision 1.1.2.11  1997/09/04 02:25:52  astark
 * Changed free race message
 *
 * Revision 1.1.2.10  1997/09/03 00:56:23  astark
 * Fixed race hashing to use filename and line num as well as var
 *
 * Revision 1.1.2.9  1997/08/29 07:38:40  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.1.2.8  1997/03/21 22:30:00  fengmd
 * USE ASSERT instead of CHECK. CHECK could not be masked by CILK_DEBUG.
 *
 * Revision 1.1.2.7  1997/03/13 18:48:09  fengmd
 * Declare pointer constant type to avoid gcc complaining.
 *
 * Revision 1.1.2.6  1997/03/04 00:07:01  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.5  1997/03/03 21:02:14  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.4  1997/02/18 21:10:05  fengmd
 * More message shown when rts error occurs.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:35  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:43  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

#ifdef CILK_ND

/* region of shared memory */
DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_end, );

/* SP-bags algorithm data structure */
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, parent, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, rank, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, ref_count, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, next_proc_id, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, current_proc_id, );
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT, collected_proc_id, );
DECLARE_SHARED_RO_VARIABLE(int, first_stop, )=0;
DECLARE_SHARED_RO_VARIABLE(int, float_comm, )=0;
DECLARE_SHARED_RO_VARIABLE(int, recycle_mem, )=0;
DECLARE_SHARED_RO_VARIABLE(int, nd_arena, );
DECLARE_SHARED_RO_VARIABLE(int, all_sets, ) = 1;
DECLARE_SHARED_RO_VARIABLE(int, umbrella, ) = 0;
DECLARE_SHARED_RO_VARIABLE(int, in_inlet, ) = 0;
DECLARE_SHARED_RO_VARIABLE(StackFrame*, inlet_parent_frame, );
DECLARE_SHARED_RO_VARIABLE(int, nd_library_inited, ) = 0;

/* we need race_party in both versions to distinguish free/malloc bugs */
DECLARE_SHARED_RO_VARIABLE(DisjointSetMemberT*, race_party, );

/* variables needed to calculate shadow space offsets
   could be made into constants for efficiency */
DECLARE_SHARED_RO_VARIABLE(size_t, static_accessor_offset, );
DECLARE_SHARED_RO_VARIABLE(size_t, static_locks_offset, );
DECLARE_SHARED_RO_VARIABLE(size_t, static_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, heap_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_size, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_size2, );
DECLARE_SHARED_RO_VARIABLE(size_t, both_plus_heap_size, );
extern int _edata;

DECLARE_SHARED_RO_VARIABLE(FileListT*, filelist, );
DECLARE_SHARED_RO_VARIABLE(int, race_linenum, );
DECLARE_SHARED_RO_VARIABLE(char*, race_filename, );
DECLARE_SHARED_RO_VARIABLE(char*, race_name, );
DECLARE_SHARED_RO_VARIABLE(RaceHashT*, race_table, [RACE_HASH_SIZE]);

DECLARE_SHARED_RO_VARIABLE(int, num_current_locks, )=0;
DECLARE_SHARED_RO_VARIABLE(LockSetT, current_locks, );

/* used in ALL_SETS */
DECLARE_SHARED_RO_VARIABLE(HashT, current_lock_hash, );
DECLARE_SHARED_RO_VARIABLE(HashT, lock_hashed, );
DECLARE_SHARED_RO_VARIABLE(LockSetHashT**, lock_hash_table, );
DECLARE_SHARED_RO_VARIABLE(LockSetHashT, lock_hash_table_entry, );

/* variables for standalone Nondeterminator (independent from rts) */
DECLARE_SHARED_RO_VARIABLE(int, active_size, );
DECLARE_SHARED_RO_VARIABLE(int, dump_core, );
DECLARE_SHARED_RO_VARIABLE(size_t, stackdepth, );
DECLARE_SHARED_RO_VARIABLE(void *, infofile, );
DECLARE_SHARED_RO_VARIABLE(const char *, assertion_failed_msg, );
DECLARE_SHARED_RO_VARIABLE(const char *, stack_overflow_msg, );
DECLARE_PRIVATE_VARIABLE(ClosureCache *, cache ,);
DECLARE_SHARED_RO_VARIABLE(static ClosureCache *, caches, );
DECLARE_PRIVATE_VARIABLE(int, _Self, );
DECLARE_SHARED_VARIABLE(int, done,);
#ifdef CILK_LIBRARY_VERSION
DECLARE_SHARED_VARIABLE(int, exit,);
#endif
int CILK_MAGIC_NAME;

/* dummy staff to avoid compiler complaining */
DECLARE_PRIVATE_VARIABLE(Cilk_time, user_work,);
DECLARE_PRIVATE_VARIABLE(Cilk_time, user_critical_path,);
DECLARE_PRIVATE_VARIABLE(Cilk_time, last_cp_time,);

static void grow_disjoint_set();

/* 
 * allocate a procedure ID from garbage-collected pool, or generated a new 
 */
static DisjointSetMemberT alloc_proc_id()
{
  DisjointSetMemberT i;

  if (USE_SHARED_RO(collected_proc_id)) {
    i = USE_SHARED_RO(collected_proc_id);
    USE_SHARED_RO(collected_proc_id) = USE_SHARED_RO(parent)[i];
  }
  else {
    i = USE_SHARED_RO(next_proc_id);
    if (i >= USE_SHARED_RO(nd_arena))
      grow_disjoint_set();
    USE_SHARED_RO(next_proc_id)++;
    USE_SHARED_RO(ref_count)[i] = 0;
  }
  return i;
}

/* 
 * garbage-collect a procedure ID 
 */
static void collect_proc_id(DisjointSetMemberT i)
{
  /* CILK_ASSERT(i >= FIRST_THREAD); */
  if (i >= FIRST_THREAD) {
    USE_SHARED_RO(parent)[i] = USE_SHARED_RO(collected_proc_id);
    USE_SHARED_RO(collected_proc_id) = i;
  }
}

/* 
 * Disjoint-set operations 
 */
inline static DisjointSetMemberT make_set()
{
  register DisjointSetMemberT i;

  i = USE_SHARED_RO(current_proc_id);
  USE_SHARED_RO(parent)[i] = (DisjointSetMemberT) S_SET;
  USE_SHARED_RO(rank)[i] = 0;
  USE_SHARED_RO(ref_count)[i]++;
  return i;
}

inline static DisjointSetMemberT Union(DisjointSetMemberT x, 
                                       DisjointSetMemberT y)
{
  register DisjointSetMemberT rx, ry;

  rx = USE_SHARED_RO(rank)[x];
  ry = USE_SHARED_RO(rank)[y];
  if (rx > ry) {
    USE_SHARED_RO(parent)[y] = x;
    if (y >= FIRST_THREAD)
      USE_SHARED_RO(ref_count)[x]++;
    return x;
  }
  else {
    USE_SHARED_RO(parent)[x] = y;
    if (x >= FIRST_THREAD)
      USE_SHARED_RO(ref_count)[y]++; 
   if (rx == ry)
      USE_SHARED_RO(rank)[y]++;
    return y;
  }
}

inline static DisjointSetMemberT find_set(DisjointSetMemberT proc_id)
{
  register DisjointSetMemberT p = proc_id, pp, ppp;

  if (p > S_SET) {
    if ((p = USE_SHARED_RO(parent)[p]) > S_SET) {
      if ((p = USE_SHARED_RO(parent)[p]) > S_SET) {
	do {
	  pp = p;
	  p = USE_SHARED_RO(parent)[p];
	} while (p > S_SET);
	do {
	  ppp = USE_SHARED_RO(parent)[proc_id];
	  if (USE_SHARED_RO(ref_count)[proc_id]) {
	    USE_SHARED_RO(parent)[proc_id] = pp;
	    USE_SHARED_RO(ref_count)[pp]++;
	    USE_SHARED_RO(ref_count)[ppp]--;
	  }
	  else
	    collect_proc_id(proc_id); 
	  proc_id = ppp;
	} while (proc_id != pp);
      }
    }
  }
  return p;
}

/*
 * incremental garbage collection of procedure ID
 */
inline void inc_ref_count(DisjointSetMemberT proc_id)
{
  USE_SHARED_RO(ref_count)[proc_id]++;
}

inline void dec_ref_count(DisjointSetMemberT proc_id)
{
  if ((--USE_SHARED_RO(ref_count)[proc_id]) == 0) {
    USE_SHARED_RO(ref_count)[USE_SHARED_RO(parent)[proc_id]]--;
    collect_proc_id(proc_id);
  }
}

/* 
 * update the procedure ID, and adjust the reference count at the same time 
 */
inline void update_ref_count(DisjointSetMemberT old_proc_id,
			     DisjointSetMemberT new_proc_id)
{
  if (old_proc_id != new_proc_id) {
    dec_ref_count(old_proc_id);
    inc_ref_count(new_proc_id);
  }
}

/*
 * make a proc id for inlet
 */
inline static DisjointSetMemberT make_inlet_id(StackFrame *f)
{
  register DisjointSetMemberT i;

  i = f->inlet_id = alloc_proc_id();
  USE_SHARED_RO(rank)[i] = 0;
  USE_SHARED_RO(parent)[f->p_set = Union(f->p_set, i)] = P_SET;
  return i;
}

/* 
 * test whether two operations are communitive or not 
 */
inline static Cilk_lockvar_basetype *get_inlet_lock(unsigned int op_type)
{
  register unsigned int op, typ;

  op = CHECK_OPERATOR(op_type);
  typ = CHECK_TYPE(op_type);

  switch (op) {
  case PLUSassign :
    return (typ == OTHERtype ||
	    (typ == FLOATtype && !USE_SHARED_RO(float_comm))) ? 
	      NO_LOCK : PLUS_LOCK;
  case ANDassign :
    return (typ == INTtype ||
	    (typ == FLOATtype && USE_SHARED_RO(float_comm))) ? 
	      AND_LOCK : NO_LOCK;
  case DIVassign :
    return (typ == FLOATtype && USE_SHARED_RO(float_comm)) ? 
      DIV_LOCK : NO_LOCK;
  case OTHERtype : 
    return NO_LOCK;
  default :
    CILK_CHECK(typ > OTHERtype, ("wrong type in is_commutative\n"));
    return 0;
  }
}

/*
 * search the filename in record filename-list, if not found, save it
 */
inline int find_filename(char *filename)
{
  FileListT *fl;
  int index;

  for (fl = USE_SHARED_RO(filelist), index = START_FILE_INDEX; 
       fl; 
       fl = fl->next, index++) {
    if (filename == fl->filename)
      return fl->index;
  }
  if ((fl = (FileListT *) malloc(sizeof(FileListT))) == NULL)
    Cilk_die("malloc failure in find_filename");
  fl->next = USE_SHARED_RO(filelist);
  fl->filename = filename;
  fl->index = index;
  USE_SHARED_RO(filelist) = fl;
  return index;
}

/*
 * get the filename via index
 */
inline static char *get_filename(int index)
{
  FileListT *fl;

  for (fl = USE_SHARED_RO(filelist); 
       fl && (fl->index != index); 
       fl = fl->next);
  CILK_ASSERT(fl);
  return fl->filename;
}

/*
 * add a nested lock into the current lock list
 */
inline static void Cilk_acquire_lock_check(Cilk_lockvar_basetype *lock,
					   char *name)
{
  LockSetT *p, *new_l;

  for (p = USE_SHARED_RO(current_locks).next; 
       (p != &(USE_SHARED_RO(current_locks))) && (lock > p->lock); 
       p = p->next);
  if (lock == p->lock)
    Cilk_die("sequential deadlock --- grabbing acquired locks\n");
  else {
    if ((new_l = (LockSetT *) malloc(sizeof(LockSetT))) == NULL)
      Cilk_die("malloc failed for lock set of the Nondeterminator\n");
    new_l->next = p;
    new_l->prev = p->prev;
    new_l->lock = lock;
    new_l->name = name;
    p->prev->next = new_l;
    p->prev = new_l;
    USE_SHARED_RO(num_current_locks)++;
    if (USE_SHARED_RO(all_sets)) {
      USE_SHARED_RO(current_lock_hash) += (HashT) lock;
      USE_SHARED_RO(lock_hashed) = 0;
    }
  }
}      

/*
 * remove a nested lock from the current lock list
 */
inline static void Cilk_release_lock_check(Cilk_lockvar_basetype *lock)
{
  LockSetT *p;
  
  for (p = USE_SHARED_RO(current_locks).next; 
       (p != &(USE_SHARED_RO(current_locks))) && (lock > p->lock); 
       p = p->next);
  if (lock == p->lock) {
    p->prev->next = p->next;
    p->next->prev = p->prev;
    free(p);
    USE_SHARED_RO(num_current_locks)--;
    if (USE_SHARED_RO(all_sets)) {
      USE_SHARED_RO(current_lock_hash) -= (HashT) lock;
      USE_SHARED_RO(lock_hashed) = 0;
    }
  }
  else
    Cilk_die("search failure in Cilk_release_lock_check\n");
}

/*
 * routines for UMBRELLA
 */
inline static int find_lock_in_current_lockset(Cilk_lockvar_basetype *plock)
{
  /* returns 1 if plock is in current_lockset.  0 otherwise */
  LockSetT *p;

  for (p = USE_SHARED_RO(current_locks).next; 
       (p != &(USE_SHARED_RO(current_locks))) && (plock > p->lock); 
       p = p->next);
  if (p->lock == plock) return 1;
  return 0;
}

inline static LockInfoT *find_lock_in_locksinfo(int num_lockinfos, 
						LockInfoT *lockInfop,
						Cilk_lockvar_basetype *plock)
{
  /* Searches contiguous LockInfo's for plock */
  int i;
  
  for (i=0;i<num_lockinfos;i++) {
    if (plock <= lockInfop->plock) {
      if (plock == lockInfop->plock)
	return lockInfop;
      else
	return NULL;
    }
    lockInfop++;
  }
  return NULL;
}

inline void DestroyLockInfoTs(int num_lockinfos, LockInfoT *lockInfop)
{
  int i;
  LockInfoT *copyLockInfop = lockInfop;

  if (lockInfop == NULL) return;
  
  for (i=0;i<num_lockinfos;i++) {
    dec_ref_count(GET_WITHOUT_FUNC(lockInfop->lwithout));
    lockInfop++;
  }
  free(copyLockInfop);
}  

inline void DestroyManyLocks(ManyLocksT *manyLocksp)
{
  if (manyLocksp != NULL) {
    int i;

    for (i=0;i<manyLocksp->num_lockinfos;i++)
      dec_ref_count(GET_WITHOUT_FUNC(manyLocksp->plockinfo[i].lwithout));
    free(manyLocksp->plockinfo);
    free(manyLocksp);
  }
}

/*
 * return 1 if the current held lock set is the subset of lock set in
 * lock_hash_table[h]; otherwise return 0; routine for ALL-SETS.
 */
inline static int is_current_lock_a_subset(HashT h)
{
  register LockSetT *p;
  register int i;
  register int size;
  register Cilk_lockvar_basetype **locks;
  LockSetHashT *e;
  
  e = USE_SHARED_RO(lock_hash_table)[h];
  size = e->size;
  if (USE_SHARED_RO(num_current_locks) > size)
    return 0;
  locks = e->locks;
  for (i = 0, p = USE_SHARED_RO(current_locks).next; p->lock; p = p->next) {
    while ((i < size) && (p->lock != locks[i]))
      i++;
    if (i == size)
      return 0;
  }
  return 1;
}

/*
 * return 1 if the current held lock set is the superset of lock set in
 * lock_hash_table[h]; otherwise return 0; routine for ALL-SETS.
 */
inline static int is_current_lock_a_superset(HashT h)
{
  register LockSetT *p;
  register int i;
  register int size;
  register Cilk_lockvar_basetype **locks;
  LockSetHashT *e;
  
  e = USE_SHARED_RO(lock_hash_table)[h];
  size = e->size;
  if (USE_SHARED_RO(num_current_locks) < size)
    return 0;
  locks = e->locks;
  for (i = 0, p = USE_SHARED_RO(current_locks).next; i < size; i++) {
    while (p->lock && locks[i] != p->lock)
      p = p->next;
    if (!(p->lock))
      return 0;
  }
  return 1;
}

/*
 * return 1 if the current held lock set is the same as the lock set in
 * lock_hash_table[h]; otherwise return 0; routine for ALL-SETS.
 */
inline static int is_same_as_current_locks(HashT h)
{
  register LockSetT *p;
  register int i;
  register int size;
  register Cilk_lockvar_basetype **locks;
  LockSetHashT *e;
  
  e = USE_SHARED_RO(lock_hash_table)[h];
  size = e->size;
  if (USE_SHARED_RO(num_current_locks) != size)
    return 0;
  locks = e->locks;
  for (i = 0, p = USE_SHARED_RO(current_locks).next; 
       i < size; 
       i++, p = p->next)
    if (p->lock != locks[i])
      return 0;
  return 1;
}

/*
 * return 1 if the intersection between the current held lock set with
 * the lock set in lock_hash_table[h] is empty; otherwise return 0;
 * routine for ALL-SETS.
 */
inline static int is_intersection_empty(HashT h)
{
  register LockSetT *p;
  register int i;
  register int size;
  register Cilk_lockvar_basetype **locks;
  LockSetHashT *e;
  
  e = USE_SHARED_RO(lock_hash_table)[h];
  size = e->size;
  locks = e->locks;
  for (i = 0, p = USE_SHARED_RO(current_locks).next; p->lock && (i < size);)
    if (p->lock == locks[i])
      return 0;
    else if (p->lock < locks[i])
      p = p->next;
    else
      i++;
  return 1;
}

/*
 * convert the current held lock set into a hash table index;
 * routine for ALL-SETS.
 */
HashT locks_to_hash_index()
{
  register LockSetT *p;
  register HashT hash = 0, i;
  register Cilk_lockvar_basetype **locks;
  LockSetHashT *e;

  if (USE_SHARED_RO(num_current_locks) == 0)
    return 0;
  if (USE_SHARED_RO(lock_hashed))
    return USE_SHARED_RO(lock_hashed);
  hash = ((USE_SHARED_RO(current_lock_hash) >> 2) % LOCK_HASH_SIZE) + 1;
  for (i = hash; 
       USE_SHARED_RO(lock_hash_table)[i] && (!is_same_as_current_locks(i));) {
    i++;
    if (i > LOCK_HASH_SIZE)
      i = 1;
    if (i == hash)
      Cilk_die("hash table for lock sets is full\n");
  }
  hash = i;
  if (USE_SHARED_RO(lock_hash_table)[hash] == NULL) {
    if ((e = USE_SHARED_RO(lock_hash_table)[hash] = 
	 (LockSetHashT *)malloc(sizeof(LockSetHashT)+
	    sizeof(Cilk_lockvar_basetype *) * 
	    (USE_SHARED_RO(num_current_locks)-1))) 
	== NULL)
      Cilk_die("malloc failed for lock set hash of the Nondeterminator\n");
    e->size = USE_SHARED_RO(num_current_locks);
    locks = e->locks;
    for (i = 0, p = USE_SHARED_RO(current_locks).next; 
	 i < USE_SHARED_RO(num_current_locks); 
	 i++, p = p->next)
      locks[i] = p->lock;
    USE_SHARED_RO(lock_hashed) = hash;
  }
  return hash;
}

inline void save_info_var(LockSetListT *l)
{
   l->linefile = MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
				find_filename(USE_SHARED_RO(race_filename)));
   l->pname = USE_SHARED_RO(race_name);
}

/*
 * clean shadow space, free lock list
 */
void clean_shadow_space_umbrella(DisjointSetMemberT *accessorp,
				 ManyLocksT **accessor_lock, int size)
{
  int word_size;
  ManyLocksT *l, *l_next;

  size = (size + sizeof(int) - 1) & (~(sizeof(int) -1));
  word_size = size >> 2;

  for (; word_size; word_size--) {
    DestroyManyLocks(*accessor_lock);
    dec_ref_count(*accessorp);
    *accessorp = S_SET;
    *accessor_lock = NULL;
    accessor_lock++;
    accessorp++;
  }
}

void clean_shadow_space_all_sets(LockSetListT **accessor_lock, int size)
{
  int word_size;
  LockSetListT *l, *l_next;

  size = (size + sizeof(int) - 1) & (~(sizeof(int) -1));
  word_size = size >> 2;

  for (; word_size; word_size--) {
    for (l = *accessor_lock; l && (l != (LockSetListT *) S_SET); l = l_next) {
      l_next = l->next;
      free(l);
    }
    *accessor_lock = (LockSetListT *)S_SET;
    accessor_lock++;
  }
}

/*
 * race report
 */
static int race_hash(const LineAndFileT lf1, char *v1,
		     const LineAndFileT lf2, char *v2, const int add)
{
  int index;
  RaceHashT *race, **prace;

  index = ((((int) v1 + (int) v2) >> 2) + lf1 + lf2) % RACE_HASH_SIZE;
  prace = USE_SHARED_RO(race_table) + index;
  while ((race = *prace) != NULL) {
    if (((race->current_name == v1) &&
	 (race->shadow_name == v2) &&
	 (race->current_info == lf1) &&
	 (race->shadow_info == lf2)) ||
	((race->current_name == v2) &&
	 (race->shadow_name == v1) &&
	 (race->current_info == lf2) &&
	 (race->shadow_info == lf1)))
      return 0;
    prace = &(race->next);
  }
  if (add) {
    if ((*prace = (RaceHashT *)malloc(sizeof(RaceHashT))) == NULL)
      Cilk_die("malloc failure in race_hash");
    (*prace)->current_name = v1;
    (*prace)->current_info = lf1;
    (*prace)->shadow_name = v2;
    (*prace)->shadow_info = lf2;
    (*prace)->next = NULL;
  }
  return 1;
}

static int race_hash_accessorp_current(const void *accessorp)
{
  /* this is a hack for now.  It is not necessarily correct to hash only
     on the current thread and accessor; the function should check the
     other involved threads, i.e. without-IDs, as well. */
  LineAndFileT shadow_info;
  char *shadow_name;

  if (USE_SHARED_RO(umbrella)) {
    shadow_info = *SHADOW_INFO(accessorp);
    shadow_name = *SHADOW_VAR(accessorp);
  }
  else { /* ALL_SETS */
    shadow_info = ((LockSetListT *) accessorp)->linefile;
    shadow_name = ((LockSetListT *) accessorp)->pname;
  }
  return race_hash(shadow_info, 
		   shadow_name,
		   MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
				  find_filename(USE_SHARED_RO(race_filename))),
		   USE_SHARED_RO(race_name),
		   1);
}

static int race_hash_accessorp_lockinfop(const void *accessorp,
					 const LockInfoT *lockInfop)
{
  LineAndFileT shadow_info;
  char *shadow_name;

  if (USE_SHARED_RO(umbrella)) {
    shadow_info = *SHADOW_INFO(accessorp);
    shadow_name = *SHADOW_VAR(accessorp);
  }
  else { /* ALL_SETS */
    shadow_info = ((LockSetListT *) accessorp)->linefile;
    shadow_name = ((LockSetListT *) accessorp)->pname;
  }
  return race_hash(shadow_info, shadow_name, lockInfop->linefile,
                   lockInfop->pname, 0);
}

static int race_hash_current_lockinfop(const LockInfoT *lockInfop)
{
  return race_hash(lockInfop->linefile, 
		   lockInfop->pname,
		   MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
				  find_filename(USE_SHARED_RO(race_filename))),
		   USE_SHARED_RO(race_name),
		   0);
}

static void report_race_umbrella(volatile void *p, 
				 const DisjointSetMemberT *accessorp,
				 ManyLocksT *manyLocksp,
				 const DisjointSetMemberT my_proc_id,
				 const int num_in_common)
{
  int hashfound = 0;
  int i, i2;
  LockInfoT *lockInfop, *lockInfop2;
  LockSetT *locksetp;
  LineAndFileT linefile;

  if (GET_FILE(*(SHADOW_INFO(accessorp))) == INIT_INDEX) {
    /* this is the access to never-allocated memory case */
    
    /* set the shadow var to NULL for the hash table entry */
    *SHADOW_VAR(accessorp) = NULL;
    if ((hashfound = race_hash_accessorp_current(accessorp))) {
      fprintf(stderr, "---------------------------------------\n");
      fprintf(stderr, "Access to non-allocated memory @ 0x%p", p);
      fprintf(stderr, ": `%s' (%s:%d)\n", USE_SHARED_RO(race_name),
	      USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
    }
  }
  else {
    /* ordinary race case */

    /* no common lock */
    if (num_in_common == 0) {
      if ((hashfound = race_hash_accessorp_current(accessorp))) {
 	fprintf(stderr, "---------------------------------------\n");

	if ((*accessorp) == FREE_ID || my_proc_id == FREE_ID)
	  fprintf(stderr, "Free:  ");

	fprintf(stderr, "Data race @ 0x%p\n", p);
	fprintf(stderr, "   `%s' (%s:%d) with\n", *(SHADOW_VAR(accessorp)),
		get_filename(GET_FILE(*(SHADOW_INFO(accessorp)))),
		GET_LINE(*(SHADOW_INFO(accessorp))));
	fprintf(stderr, "   `%s' (%s:%d)\n", USE_SHARED_RO(race_name),
		USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
      }
    }
    /* if only one current lock is held by accessor, 
       we can determine if a real race occurred by checking if that lock's 
       without-ID is in a P_SET */
    else if ((USE_SHARED_RO(num_current_locks) == 1) &&
	     (lockInfop = find_lock_in_locksinfo(manyLocksp->num_lockinfos,
			    manyLocksp->plockinfo,
			    (USE_SHARED_RO(current_locks).next)->lock)) &&
	     (find_set(GET_WITHOUT_FUNC(lockInfop->lwithout)) == P_SET)) {
      if ((hashfound = race_hash_accessorp_current(accessorp))) {
 	fprintf(stderr, "---------------------------------------\n");

	if ((*accessorp) == FREE_ID || my_proc_id == FREE_ID)
	  fprintf(stderr, "Free:  ");

	fprintf(stderr, "Data race @ 0x%p\n", p);
	fprintf(stderr, "   `%s' (%s:%d) with\n", lockInfop->pname,
		get_filename(GET_FILE(lockInfop->linefile)),
		GET_LINE(lockInfop->linefile));
	fprintf(stderr, "   `%s' (%s:%d)\n", USE_SHARED_RO(race_name),
		USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
	manyLocksp->race_reported = 1;
      }
    }
    else if (!(manyLocksp->race_reported)) {
      /* now report umbrella violation only if no true race already found in 
	 current umbrella */
      
      /* check if the umbrella contains any pairwise race that has been 
	 seen before */
      for(i = 0, lockInfop = manyLocksp->plockinfo, hashfound = 1; 
	  (i < manyLocksp->num_lockinfos) && hashfound; 
	  i++, lockInfop++) {
	/* check with accessor */
	if (!(hashfound = race_hash_accessorp_lockinfop(accessorp, lockInfop)))
	    break;
	/* check with current */
	if (!(hashfound = race_hash_current_lockinfop(lockInfop)))
	    break;
	/* check with others */
	for(i2 = 0, lockInfop2 = manyLocksp->plockinfo; 
	    i2 < i; 
	    i2++, lockInfop2++) {
	  if (!(hashfound = race_hash(lockInfop->linefile, lockInfop->pname,
				      lockInfop2->linefile, lockInfop2->pname,
				      0)))
	    break;
	}
      }
	
      if(hashfound) {
	if ((hashfound = race_hash_accessorp_current(accessorp))) {
	  fprintf(stderr, "---------------------------------------\n");

	  if ((*accessorp) == FREE_ID || my_proc_id == FREE_ID)
	    fprintf(stderr, "Free:  ");

	  fprintf(stderr, "Umbrella discipline violation @ 0x%p: ", p);
	  fprintf(stderr, "`%s' (%s:%d)\n", USE_SHARED_RO(race_name),
		  USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));

	  /* for each lock currently held, give an access that did not hold 
	     that lock */
	  
	  for (locksetp = USE_SHARED_RO(current_locks).next;
	       locksetp != &USE_SHARED_RO(current_locks);
	       locksetp = locksetp->next) {
	    if (locksetp->lock == RLOCK) {
	      /* it would be nice to handle this case better */
	      fprintf(stderr, "     Parallel write: ");
	    }
	    else if ((locksetp->lock == PLUS_LOCK) || 
		     (locksetp->lock == AND_LOCK)  || 
		     (locksetp->lock == DIV_LOCK)  ||
		     (locksetp->lock == USER_INLET_LOCK)) {
	      fprintf(stderr, "     Non-inlet or noncommutative inlet: ");
	    }
	    else {
	      fprintf(stderr, "     Access without lock `%s' (0x%p): ",
		      locksetp->name, locksetp->lock);
	    }

	    /* num_in_common != 0 means that manyLocksp can't be NULL */
	    CILK_ASSERT(manyLocksp); 
					  
	    if ((locksetp->lock != USER_INLET_LOCK) && 
		(lockInfop = find_lock_in_locksinfo(manyLocksp->num_lockinfos,
						    manyLocksp->plockinfo,
						    locksetp->lock))
		  != NULL) {
	      /* report without id */
	      fprintf(stderr, "`%s' (%s:%d)\n",
		      lockInfop->pname, 
		      get_filename(GET_FILE(lockInfop->linefile)),
		      GET_LINE(lockInfop->linefile));
	    }
	    else {
	      /* report accessor */
	      linefile = *(SHADOW_INFO(accessorp));
	      fprintf(stderr, "`%s' (%s:%d)\n",
		      *(SHADOW_VAR(accessorp)),
		      get_filename(GET_FILE(linefile)),
		      GET_LINE(linefile));
	    }
	  }
	}
      }
    }
  }
  
  if (hashfound) fprintf(stderr, "---------------------------------------\n");
  
  if (USE_SHARED_RO(first_stop)) 
    Cilk_die("The first data race is found, and the program is terminated\n");
}

static void report_race_all_sets(volatile void *p, 
				 LockSetListT *lockSetListp,
				 const DisjointSetMemberT my_proc_id)
{
  int hashfound = 0;

  if ((!lockSetListp) || GET_FILE(lockSetListp->linefile) == INIT_INDEX) {
    /* this is the access to never-allocated memory case */
    
    /* set the shadow var to NULL for the hash table entry */
    if (lockSetListp) lockSetListp->pname = NULL;
    if ((!lockSetListp) || 
	(hashfound = race_hash_accessorp_current(lockSetListp))) {
      fprintf(stderr, "---------------------------------------\n");
      fprintf(stderr, "Access to non-allocated memory @ 0x%p", p);
      fprintf(stderr, ": `%s' (%s:%d)\n", USE_SHARED_RO(race_name),
	      USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
    }
  }
  else {
    /* ordinary race case */

    if ((hashfound = race_hash_accessorp_current(lockSetListp))) {

      fprintf(stderr, "---------------------------------------\n");

      if (lockSetListp->proc_id == FREE_ID || my_proc_id == FREE_ID)
	fprintf(stderr, "Free:  ");

      fprintf(stderr, "Data race @ 0x%p\n", p);
      fprintf(stderr, "   `%s' (%s:%d) with\n", lockSetListp->pname,
	      get_filename(GET_FILE(lockSetListp->linefile)),
	      GET_LINE(lockSetListp->linefile));
      fprintf(stderr, "   `%s' (%s:%d)\n", USE_SHARED_RO(race_name),
	      USE_SHARED_RO(race_filename), USE_SHARED_RO(race_linenum));
    }
  }

  if ((!lockSetListp) || hashfound) 
    fprintf(stderr, "---------------------------------------\n");
  
  if (USE_SHARED_RO(first_stop)) 
    Cilk_die("The first data race is found, and the program is terminated\n");
}


/*
 * main race-checking routines
 */
static void accessor_check_umbrella(volatile void *p, 
				    DisjointSetMemberT *accessorp,
				    ManyLocksT **manyLockspp,
				    DisjointSetMemberT my_proc_id, 
				    int op_type, StackFrame *f)
{
  /* registers? */
  DisjointSetMemberT accessor = *accessorp;
  DisjointSetMemberT without;
  ManyLocksT *manyLocksp = *manyLockspp;
  LockInfoT *lockInfop;
  LockInfoT *currLockInfop;
  LockInfoT *matchLockInfop;
  LockInfoT *oldLockInfop;
  int i, any_alive = 0, any_common = 0;
  LiveAndWithoutT lwithout;
  ManyLocksT *newManyLocksp;
  LockSetT *plockSet;
  int old_num_lockinfos;
  
  if (find_set(accessor) == P_SET) {
    if (manyLocksp != NULL) {
      /* if manyLocksp is NULL, there is automatically a race */
      lockInfop = manyLocksp->plockinfo;
      for (i=0; i<manyLocksp->num_lockinfos; i++) {
	/* for each c in LOCKS[l] with alive true */
	lwithout = lockInfop->lwithout;
	if (GET_LIVE(lwithout)) {
	  /* if c in H */
	  if (find_lock_in_current_lockset(lockInfop->plock)) {
	    if ((USE_SHARED_RO(in_inlet) && accessor != f->inlet_id) ||
		find_set(GET_WITHOUT_FUNC(lwithout)) == P_SET) {
	      any_common++;
	      /* kill lock due to without function */
	      lockInfop->lwithout = SET_LIVE(lwithout, 0);
	    }
	    else {
	      /* the lock is left alive */			    
	      any_alive = 1;
	    }
	  }
	  else {
	    /* c not in H.  kill lock, set without function for reporting */
	    lockInfop->lwithout = MAKE_LIVE_WITHOUT(0, my_proc_id);
	    inc_ref_count(my_proc_id);
	    
	    lockInfop->linefile =
	      MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
			     find_filename(USE_SHARED_RO(race_filename)));
	    lockInfop->pname = USE_SHARED_RO(race_name);
	  }
	}
	else { 
	  /* if lock is already dead, see if without-ID should be replaced */
	  if(!find_lock_in_current_lockset(lockInfop->plock)) {
	    if(find_set(GET_WITHOUT_FUNC(lwithout)) == S_SET) {
	      lockInfop->lwithout = MAKE_LIVE_WITHOUT(0, my_proc_id);
	      inc_ref_count(my_proc_id);

	      lockInfop->linefile =
		MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
			       find_filename(USE_SHARED_RO(race_filename)));
	      lockInfop->pname = USE_SHARED_RO(race_name);
	    }
	  }
	  else
	    any_common++;
	}
	lockInfop++;
      } /* end for loop */
    } /*end (manyLocksp != NULL) */
    
    if (!any_alive) {
	report_race_umbrella(p, accessorp, manyLocksp, my_proc_id, any_common);
    }
  }
  else {
    /* ACCESSOR[l] is in an S-bag */
    
    if (USE_SHARED_RO(num_current_locks) > 0)	{
      /* allocate a new ManyLocks structure for as many locks as are in H */
      if (manyLocksp != NULL) {
	newManyLocksp = manyLocksp;
	old_num_lockinfos = manyLocksp->num_lockinfos;
	oldLockInfop = manyLocksp->plockinfo;
      }
      else {
	if ((newManyLocksp = (ManyLocksT *)malloc(sizeof(ManyLocksT)))
	    == NULL)
	  Cilk_die("malloc failure in accessor_check_umbrella");
	old_num_lockinfos = 0;
	oldLockInfop = NULL;
      }
      newManyLocksp->num_lockinfos = USE_SHARED_RO(num_current_locks);
      if ((newManyLocksp->plockinfo = 
	   (LockInfoT *)malloc(USE_SHARED_RO(num_current_locks)
			       * sizeof(LockInfoT)))
	  == NULL)
	  Cilk_die("malloc failure in accessor check");
      newManyLocksp->race_reported = 0;

      currLockInfop = newManyLocksp->plockinfo;
      
      for (plockSet = USE_SHARED_RO(current_locks).next; 
	   plockSet != &USE_SHARED_RO(current_locks);
	   plockSet = plockSet->next) {
	currLockInfop->plock = plockSet->lock;
	if ((matchLockInfop = find_lock_in_locksinfo(old_num_lockinfos,
						     oldLockInfop,
						     currLockInfop->plock))
	    != NULL) {
	  /* use the without function from LOCKS */
	  without = GET_WITHOUT_FUNC(matchLockInfop->lwithout);
	  currLockInfop->lwithout = MAKE_LIVE_WITHOUT(1, without);
	  inc_ref_count(without);
	  currLockInfop->linefile = matchLockInfop->linefile;
	  currLockInfop->pname = matchLockInfop->pname;
	}
	else {
	  /* ACCESSOR[l] is the without function. At the first access,
	     ACCESSOR[l] will contain whoever did the allocation of l
	     (the proc id if malloced on the heap; S_SET if static) */
	  currLockInfop->lwithout = MAKE_LIVE_WITHOUT(1, accessor);
	  inc_ref_count(accessor);
	  currLockInfop->linefile = *(SHADOW_INFO(accessorp));
	  currLockInfop->pname = *(SHADOW_VAR(accessorp));
	}
	currLockInfop++;
      } /* end for loop */
    }
    else {
      /* no locks currently held */
      newManyLocksp = manyLocksp;;
      if (newManyLocksp != NULL) {
	old_num_lockinfos = manyLocksp->num_lockinfos;
	oldLockInfop = manyLocksp->plockinfo;
	newManyLocksp->plockinfo = NULL;
	newManyLocksp->num_lockinfos = 0;
      }
      else {
	old_num_lockinfos = 0;
	oldLockInfop = NULL;
      }
    }
    
    /* put the new ManyLocks structure in place of the old one */
    DestroyLockInfoTs(old_num_lockinfos, oldLockInfop);
    *manyLockspp = newManyLocksp;
    
    /* overwrite ACCESSOR[l] */
    *accessorp = my_proc_id;
    update_ref_count(accessor, my_proc_id);
    *(SHADOW_INFO(accessorp)) = 
      MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
		     find_filename(USE_SHARED_RO(race_filename)));
    *(SHADOW_VAR(accessorp)) = USE_SHARED_RO(race_name);
  }
}

static void accessor_check_all_sets(volatile void *p, 
				    LockSetListT **lockSetListpp,
				    DisjointSetMemberT my_proc_id, 
				    int op_type, StackFrame *f)
{
  LockSetListT *l, *l_prec, *l_next;
  int redundant = 0;

  if (!f && !(*lockSetListpp))
    report_race_all_sets(p, *lockSetListpp, my_proc_id);
  else if (*lockSetListpp == (LockSetListT *) S_SET)
    *lockSetListpp = (LockSetListT *) NULL;
  for (l = *lockSetListpp, l_prec = NULL; l; l = l_next) {
    l_next = l->next;
    if (find_set(l->proc_id) == P_SET) {
      if ((USE_SHARED_RO(in_inlet) && (l->proc_id != f->inlet_id)) ||
	  is_intersection_empty(l->hash_index))
	report_race_all_sets(p, l, my_proc_id);
      if (is_current_lock_a_superset(l->hash_index))
	redundant = 1;
    }
    else {
      if (is_current_lock_a_subset(l->hash_index)) {
	if (l_prec)
	  l_prec->next = l_next;
	else
	  *lockSetListpp = l_next;
	free(l);
      }
      else
	l_prec = l;
    }
  }
  if (!redundant) {
    if ((l = (LockSetListT *) malloc(sizeof(LockSetListT))) == NULL)
      Cilk_die("malloc failed for lock list of the Nondeterminator\n");
    l->next = *lockSetListpp;
    l->hash_index = locks_to_hash_index();
    l->proc_id = my_proc_id;
    USE_SHARED_RO(ref_count)[my_proc_id]++;
    save_info_var(l);
    *lockSetListpp = l;
  }
}

static void free_check_umbrella(volatile void *p, 
				DisjointSetMemberT *accessorp,
				ManyLocksT **manyLockspp,
				DisjointSetMemberT my_proc_id, 
				int op_type)
{
  /* registers? */
  DisjointSetMemberT accessor = *accessorp;
  ManyLocksT *manyLocksp = *manyLockspp;
  LockInfoT *lockInfop;
  int i, any_alive = 0;
  LiveAndWithoutT lwithout;

  if (find_set(accessor) == P_SET) {
    if (manyLocksp != NULL) {
      /* if manyLocksp is NULL, there is automatically a race */
      lockInfop = manyLocksp->plockinfo;
      for (i=0; i<manyLocksp->num_lockinfos; i++) {
	/* for each c in LOCKS[l] with alive true */
	lwithout = lockInfop->lwithout;
	if (GET_LIVE(lwithout)) {
	  /* if c in H */
	  if (find_lock_in_current_lockset(lockInfop->plock)) {
	    if (find_set(GET_WITHOUT_FUNC(lwithout)) == S_SET) {
	      any_alive = 1;
	      break; /* in the free case, as soon as we know there is no race,
			we can quit this case immediately, because we are going
			to overwrite accessor anyway. */
	    }
	  }
	}
	lockInfop++;
      } /* end for loop */
    } /*end (manyLocksp != NULL) */

    if (!any_alive)
      report_race_umbrella(p, accessorp, manyLocksp, my_proc_id, 0);
  }

  /* Always overwrite accessor when doing a free.  
     The locks info can be deleted, 
     because later accesses to freed memory can't be protected by locks. */
  DestroyManyLocks(manyLocksp);
  *manyLockspp = NULL;
  /* overwrite ACCESSOR[l] */
  *accessorp = my_proc_id;
  dec_ref_count(accessor);   /* don't need ref count for FREE_ID */
  *(SHADOW_INFO(accessorp)) = 
    MAKE_LINE_FILE(USE_SHARED_RO(race_linenum),
		   find_filename(USE_SHARED_RO(race_filename)));
  *(SHADOW_VAR(accessorp)) = USE_SHARED_RO(race_name);
}

static void free_check_all_sets(volatile void *p, 
				LockSetListT **lockSetListpp,
				DisjointSetMemberT my_proc_id, 
				int op_type)
{
  LockSetListT *l, *l_prec, *l_next;
  int redundant = 0;

  CILK_ASSERT(*lockSetListpp); 
  /* free/free problem should already be caught by rts malloc */

  if (*lockSetListpp == (LockSetListT *) S_SET)
    *lockSetListpp = (LockSetListT *) NULL;
  
  for (l = *lockSetListpp, l_prec = NULL; l; l = l_next) {
    l_next = l->next;
    if (find_set(l->proc_id) == P_SET) {
      if (is_intersection_empty(l->hash_index))
	report_race_all_sets(p, l, my_proc_id);
    }
  }

  /* only the FREE_ID needs to be left in the shadow space, 
     since later accesses can't be protected by locks */
  clean_shadow_space_all_sets(lockSetListpp, 1);
  if ((l = (LockSetListT *) malloc(sizeof(LockSetListT))) == NULL)
    Cilk_die("malloc failed for lock list of the Nondeterminator\n");
  l->next = NULL;
  l->hash_index = 0;
  l->proc_id = my_proc_id;
  save_info_var(l);
  *lockSetListpp = l;
}

/*
 * check the read access in region [p, p+size).
 */
inline volatile void *Cilk_internal_read_check(volatile void *p, size_t size, 
					       int op_type, int linenum, 
					       char *filename, char *name)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  Cilk_acquire_lock_check(RLOCK, FAKE_LOCK_NAME(RLOCK));
  Cilk_internal_write_check(p, size, op_type, linenum, filename, name);
  Cilk_release_lock_check(RLOCK);
  return p;
}

/*
 * check the read access in region [p, p+4).
 * this is to optimize the common case of "read an integer or float".
 */
inline volatile void *Cilk_internal_read_check4(volatile void *p, int op_type, 
						int linenum, char *filename, 
						char *name)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  Cilk_acquire_lock_check(RLOCK, FAKE_LOCK_NAME(RLOCK));
  Cilk_internal_write_check4(p, op_type, linenum, filename, name);
  Cilk_release_lock_check(RLOCK);
  return p;
}   

/*
 * check the read access in region [p, p+8).
 * this is to optimize the common case of "read a double".
 */
inline volatile void *Cilk_internal_read_check8(volatile void *p, int op_type, 
						int linenum, char *filename, 
						char *name)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  Cilk_acquire_lock_check(RLOCK, FAKE_LOCK_NAME(RLOCK));
  Cilk_internal_write_check8(p, op_type, linenum, filename, name);
  Cilk_release_lock_check(RLOCK);
  return p;
}   
    
/*
 * check the write access in region [p, p+size)
 */
volatile void *Cilk_internal_write_check(volatile void *p, size_t size, 
					 int op_type, int linenum, 
					 char *filename, char *name)
{
  register int *pp;
  register DisjointSetMemberT my_proc_id;
  register StackFrame* f;

#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  if (USE_SHARED_RO(in_inlet)) {
    f = USE_SHARED_RO(inlet_parent_frame);
    my_proc_id = (f->inlet_id == NO_INLET_ID) ? make_inlet_id(f) : f->inlet_id;
  }
  else {
    f = NULL;
    my_proc_id = USE_SHARED_RO(current_proc_id);
  }

  USE_SHARED_RO(race_linenum) = linenum; 
  USE_SHARED_RO(race_filename) = filename;
  USE_SHARED_RO(race_name) = name;

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      if (USE_SHARED_RO(umbrella))
	for (; size; size -= 1, pp++) {
	  accessor_check_umbrella(pp, HEAP_ACCESSOR(pp), HEAP_LOCKS(pp), 
				  my_proc_id, op_type, f);
	}
      else { /* ALL_SETS */
	for (; size; size -= 1, pp++) {
	  accessor_check_all_sets(pp, HEAP_LOCKLISTS(pp), my_proc_id, op_type,
				  f);
	}
      }
      return p;
    }
  }
  
  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      if (USE_SHARED_RO(umbrella))
	for (; size; size -= 1, pp++) {
	  accessor_check_umbrella(pp, STATIC_ACCESSOR(pp), STATIC_LOCKS(pp), 
				  my_proc_id, op_type, f);
	}
      else { /* ALL_SETS */
	for (; size; size -= 1, pp++) {
	  accessor_check_all_sets(pp, STATIC_LOCKLISTS(pp), 
				  my_proc_id, op_type, f);
	}
      }
      return p;
    }
  }

  return p;
}

/*
 * check the write access in region [p, p+4).
 * this is to optimize the common case of "write an integer or float".
 */
volatile void *Cilk_internal_write_check4(volatile void *p, int op_type, 
					  int linenum, char *filename, 
					  char *name)
{
  register DisjointSetMemberT my_proc_id;
  register StackFrame* f;

#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  if (USE_SHARED_RO(in_inlet)) {
    f = USE_SHARED_RO(inlet_parent_frame);
    my_proc_id = (f->inlet_id == NO_INLET_ID) ? make_inlet_id(f) : f->inlet_id;
  }
  else {
    f = NULL;
    my_proc_id = USE_SHARED_RO(current_proc_id);
  }

  USE_SHARED_RO(race_linenum) = linenum; 
  USE_SHARED_RO(race_filename) = filename;
  USE_SHARED_RO(race_name) = name;

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (USE_SHARED_RO(umbrella))
	accessor_check_umbrella(p, HEAP_ACCESSOR(p), HEAP_LOCKS(p), 
				my_proc_id, op_type, f);
      else /* ALL_SETS */
	accessor_check_all_sets(p, HEAP_LOCKLISTS(p), my_proc_id, op_type, f);
      return p;
    }
  }

  if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      if (USE_SHARED_RO(umbrella))
	accessor_check_umbrella(p, STATIC_ACCESSOR(p), STATIC_LOCKS(p), 
				my_proc_id, op_type, f);
      else /* ALL_SETS */
	accessor_check_all_sets(p, STATIC_LOCKLISTS(p), my_proc_id, op_type, 
				f);
      return p;
    }
  }

  return p;
}

/*
 * check the write access in region [p, p+8).
 * this is to optimize the common case of "write a double".
 */
volatile void *Cilk_internal_write_check8(volatile void *p, int op_type, 
					  int linenum, char *filename, 
					  char *name)
{
  register volatile void *pp;
  register DisjointSetMemberT my_proc_id;
  register StackFrame* f;

#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return p;
#endif

  if (USE_SHARED_RO(in_inlet)) {
    f = USE_SHARED_RO(inlet_parent_frame);
    my_proc_id = (f->inlet_id == NO_INLET_ID) ? make_inlet_id(f) : f->inlet_id;
  }
  else {
    f = NULL;
    my_proc_id = USE_SHARED_RO(current_proc_id);
  }

  USE_SHARED_RO(race_linenum) = linenum; 
  USE_SHARED_RO(race_filename) = filename;
  USE_SHARED_RO(race_name) = name;

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p+4 < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (USE_SHARED_RO(umbrella)) {
	accessor_check_umbrella(p, HEAP_ACCESSOR(p), HEAP_LOCKS(p), 
				my_proc_id, op_type, f);
	pp = p + 4;
	accessor_check_umbrella(p, HEAP_ACCESSOR(p), HEAP_LOCKS(p), 
				my_proc_id, op_type, f);
      }
      else { /* ALL_SETS */
	accessor_check_all_sets(p, HEAP_LOCKLISTS(p), my_proc_id, op_type, f);
	pp = p + 4;
	accessor_check_all_sets(p, HEAP_LOCKLISTS(p), my_proc_id, op_type, f);
      }
      return p;
    }
  }

  if (p >= (void *) STATIC_ARENA_START) {
    if (p+4 < (void *) STATIC_ARENA_END) {
      if (USE_SHARED_RO(umbrella)) {
	accessor_check_umbrella(p, STATIC_ACCESSOR(p), STATIC_LOCKS(p), 
				my_proc_id, op_type, f);
      	pp = p + 4;
	accessor_check_umbrella(pp, STATIC_ACCESSOR(pp), STATIC_LOCKS(pp), 
				my_proc_id, op_type, f);
      }
      else { /* ALL_SETS */
	accessor_check_all_sets(p, STATIC_LOCKLISTS(p), my_proc_id, op_type,
				f);
	pp = p + 4;
	accessor_check_all_sets(pp, STATIC_LOCKLISTS(pp), my_proc_id, op_type,
				f);
      }
      return p;
    }
  }

  return p;
}

/* 
 * similar to Cilk_internal_write_check, check the value assigned by
 * spawned procedures 
 */
void Cilk_internal_spawnvar_check(volatile void *p, size_t size, int op_type, 
				  StackFrame *f, int linenum, char *filename, 
				  char *name)
{
  /* could be optimized since it is known that the only lock that may be held
     is the inlet lock */
  
  register int *pp;
  Cilk_lockvar_basetype *inlet_lock;
  DisjointSetMemberT my_proc_id = (f->inlet_id == NO_INLET_ID) ? 
    make_inlet_id(f) : f->inlet_id;

  USE_SHARED_RO(in_inlet) = 1;

  if ((inlet_lock = get_inlet_lock(op_type)) != NULL) {
    Cilk_acquire_lock_check(inlet_lock, FAKE_LOCK_NAME(inlet_lock));
  }

  USE_SHARED_RO(race_linenum) = linenum; 
  USE_SHARED_RO(race_filename) = filename;
  USE_SHARED_RO(race_name) = name;

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp+size-1 < (int *) USE_SHARED_RO(heap_arena_end)) {
      if (USE_SHARED_RO(umbrella)) {
	for (; size; size -= 1, pp++) {
	  accessor_check_umbrella(pp, HEAP_ACCESSOR(pp), HEAP_LOCKS(pp), 
				  my_proc_id, op_type, f);
	}
      }
      else { /* ALL_SETS */
	for (; size; size -= 1, pp++) {
	  accessor_check_all_sets(pp, HEAP_LOCKLISTS(pp), 
				  my_proc_id, op_type, f);
	}
      }

      if (inlet_lock != NULL) Cilk_release_lock_check(inlet_lock);

      USE_SHARED_RO(in_inlet) = 0;

      return;
    }
  }

  if (pp >= (int *) STATIC_ARENA_START) {
    if (pp+size-1 < (int *) STATIC_ARENA_END) {
      if (USE_SHARED_RO(umbrella)) {
	for (; size; size -= 1, pp++) {
	  accessor_check_umbrella(pp, STATIC_ACCESSOR(pp), STATIC_LOCKS(pp),
				 my_proc_id, op_type, f);
	}
      }
      else { /* ALL_SETS */
	for (; size; size -= 1, pp++) {
	  accessor_check_all_sets(pp, STATIC_LOCKLISTS(pp),
				  my_proc_id, op_type, f);
	}
      }
	
      if (inlet_lock != NULL) Cilk_release_lock_check(inlet_lock);

      USE_SHARED_RO(in_inlet) = 0;

      return;
    }
  }

  CILK_CHECK(pp, ("spawn receiver address %p out of range\n", pp));
  return;
}

/* 
 * similar to Cilk_internal_write_check, except using FREE_ID as 
 * current_proc_id.  Also, p is guaranteed to be on the heap 
 */
volatile void *Cilk_internal_free_check(volatile void *p, size_t size, 
					int op_type, int linenum, 
					char *filename, char *name)
{
  register int *pp;
  
  USE_SHARED_RO(race_linenum) = linenum; 
  USE_SHARED_RO(race_filename) = filename;
  USE_SHARED_RO(race_name) = name;

  size = (((size + sizeof(int) - 1) & (~(sizeof(int) -1))) / sizeof(int));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  /* pp must point into the heap */
  CILK_ASSERT(pp >= (int *)USE_SHARED_RO(heap_arena_start));
  CILK_ASSERT(pp+size-1 < (int *)USE_SHARED_RO(heap_arena_end));

  if (USE_SHARED_RO(umbrella)) {
    for (; size; size -= 1, pp++) {
      /*
       * if the shadow space contains FREE_ID, that is not a race since 
       * that represents the extra memory returned as part of the malloc 
       * bucket 
       */
      if (*HEAP_ACCESSOR(pp) == FREE_ID) break;
      free_check_umbrella(pp, HEAP_ACCESSOR(pp), HEAP_LOCKS(pp), FREE_ID, 
			  op_type);
      /* really only want to do the parallel case here, 
	 just to check for races */
    }
  }
  else { /* ALL_SETS */
    for (; size; size -= 1, pp++) {
      /*
       * if the shadow space contains FREE_ID, that is not a race since 
       * that represents the extra memory returned as part of the malloc 
       * bucket 
       */
      if (*HEAP_LOCKLISTS(pp) == FREE_ID) break;
      free_check_all_sets(pp, HEAP_LOCKLISTS(pp), FREE_ID, op_type);
    }
  }

  return p;
}
    
/*
 * allocate a procedure ID to a new spawned procedure
 * Cilk_alloc_proc_id_spbags is called in INIT_FRAME before frame is allocated
 */
inline void Cilk_alloc_proc_id_spbags()
{
  USE_SHARED_RO(current_proc_id) = alloc_proc_id();
}

/*
 * initialize the SP-bag per instance of the procedure.
 * Cilk_init_spbags is called in INIT_FRAME when a procedure starts
 */
inline void Cilk_init_spbags(StackFrame *f)
{
  f->proc_id = f->s_set = make_set();
  f->p_set = P_SET;
  f->inlet_id = NO_INLET_ID;
}

/*
 * union the S-bag of the child into the P-bag of the parent.
 * Cilk_lreturn_spbags is called in _XPOP_FRAME_RESULT or _XPOP_FRAME_NORESULT
 */
inline void Cilk_lreturn_spbags(StackFrame *f, volatile StackFrame *child)
{
  USE_SHARED_RO(parent)[f->p_set = Union(f->p_set, child->s_set)] = P_SET;
  USE_SHARED_RO(current_proc_id) = f->proc_id;
  if (((--USE_SHARED_RO(ref_count)[child->proc_id]) == 0) &&
      (f->p_set != child->proc_id)) {
    USE_SHARED_RO(ref_count)[USE_SHARED_RO(parent)[child->proc_id]]--;
    collect_proc_id(child->proc_id);
  }
}

/*
 * union the P-bag of the current procedure into its S-bag.
 * Cilk_sync_spbags is called in _AT_SYNC_FAST
 */
inline void Cilk_sync_spbags(StackFrame *f)
{
  USE_SHARED_RO(parent)[f->s_set = Union(f->p_set, f->s_set)] = S_SET;
  f->p_set = P_SET;
  f->inlet_id = NO_INLET_ID; /* after sync, a new inlet id is needed */
}

/*
 * set the flag to indicate entering an inlet
 */
void Cilk_enter_inlet(StackFrame *f)
{
  USE_SHARED_RO(in_inlet) = 1;
  USE_SHARED_RO(inlet_parent_frame) = f;
  Cilk_acquire_lock_check(USER_INLET_LOCK, FAKE_LOCK_NAME(USER_INLET_LOCK));
}

/*
 * set the flag to indicate leaving an inlet
 */
void Cilk_leave_inlet()
{
  USE_SHARED_RO(in_inlet) = 0;
  Cilk_release_lock_check(USER_INLET_LOCK);
}

/*
 * acquire lock routice called in the Cilk program (not Cilk runtime system).
 */
inline void Cilk_nd_lock(Cilk_lockvar_basetype *lock, char *name)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return;
#endif

  Cilk_acquire_lock_check(lock, name);
}

/*
 * release lock routice called in the Cilk program (not Cilk runtime system).
 */
inline void Cilk_nd_unlock(Cilk_lockvar_basetype *lock)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return;
#endif

  Cilk_release_lock_check(lock);
}

/*
 * warning because spawn in critical section
 */
inline void Cilk_locked_spawn_warn()
{
  static int warned = 0;
  if (!warned) {
    Cilk_dprintf("WARNING: Spawn encountered while holding user lock. "
		 "Some data races may be missed\n");
    warned = 1;
  }
}

inline void Cilk_publish(void *p, int sz)
{
#ifdef CILK_LIBRARY_VERSION
  if (USE_SHARED_RO(nd_library_inited) == 0) 
    return;
#endif

  if (USE_SHARED_RO(umbrella))
    clean_shadow_space_umbrella(HEAP_ACCESSOR(p), HEAP_LOCKS(p), sz);
  else
    clean_shadow_space_all_sets(HEAP_LOCKLISTS(p), sz);
}

/* 
 * initialize shared read-only variables 
 */
static void init_shared_ro()
{
  USE_SHARED_RO(current_proc_id) =S_SET;
  USE_SHARED_RO(next_proc_id) = FIRST_THREAD;
  USE_SHARED_RO(collected_proc_id) = 0;
  USE_SHARED_RO(filelist) = NULL;
  memset(USE_SHARED_RO(race_table), 0, RACE_HASH_SIZE*sizeof(RaceHashT *));

  USE_SHARED_RO(current_locks).next = USE_SHARED_RO(current_locks).prev =
    &(USE_SHARED_RO(current_locks));
  USE_SHARED_RO(current_locks).lock = (Cilk_lockvar_basetype *) 0;
  USE_SHARED_RO(num_current_locks) = 0;

  if (USE_SHARED_RO(all_sets)) {
    USE_SHARED_RO(current_lock_hash) = 0;
    USE_SHARED_RO(lock_hashed) = 0;
    if ((USE_SHARED_RO(lock_hash_table) =
	 (LockSetHashT **)malloc(sizeof(LockSetHashT *)*(LOCK_HASH_SIZE+1)))
	== NULL)
      Cilk_die("malloc error in init_shared_ro for lock_hash_table\n");
    USE_SHARED_RO(lock_hash_table_entry).size = 0;
    USE_SHARED_RO(lock_hash_table)[0] = &USE_SHARED_RO(lock_hash_table_entry);
  }
}

/* 
 * malloc space for parent and rank used in disjoint-set 
 */
static void grow_disjoint_set()
{
  int new_nd_arena = 2 * USE_SHARED_RO(nd_arena);
  DisjointSetMemberT *new_parent, *new_rank, *new_ref_count;
  
  if (USE_SHARED_RO(nd_arena) == MAX_PROC_NUMBER)
    Cilk_die("Too many procedure instances to store procedure IDs in shadow"
	     "space\n");
  
  if (new_nd_arena > MAX_PROC_NUMBER)
    new_nd_arena = MAX_PROC_NUMBER;

  if ((new_parent =
      (DisjointSetMemberT *)malloc(new_nd_arena*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");

  memcpy(new_parent, USE_SHARED_RO(parent),
	 USE_SHARED_RO(nd_arena) * sizeof(DisjointSetMemberT));
  free(USE_SHARED_RO(parent));
  USE_SHARED_RO(parent) = new_parent;
  
  if ((new_rank =
      (DisjointSetMemberT *)malloc(new_nd_arena*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");

  memcpy(new_rank, USE_SHARED_RO(rank),
	 USE_SHARED_RO(nd_arena) * sizeof(DisjointSetMemberT));
  free(USE_SHARED_RO(rank));
  USE_SHARED_RO(rank) = new_rank;

  if ((new_ref_count =
      (DisjointSetMemberT *)malloc(new_nd_arena*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");

  memcpy(new_ref_count, USE_SHARED_RO(ref_count),
	 USE_SHARED_RO(nd_arena) * sizeof(DisjointSetMemberT));
  free(USE_SHARED_RO(ref_count));
  USE_SHARED_RO(ref_count) = new_ref_count;

  USE_SHARED_RO(nd_arena) = new_nd_arena;

}

static void init_disjoint_set()
{
  USE_SHARED_RO(nd_arena) = DEFAULT_ND_HEAPSIZE;
  
  if ((USE_SHARED_RO(parent) =
       (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				    sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");
  
  if ((USE_SHARED_RO(rank) = 
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");
  
  if ((USE_SHARED_RO(ref_count) = 
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc failure for Nondeterminator's disjoint set structures\n");

  /* put the FREE_ID into a separate P-bag */
  USE_SHARED_RO(parent)[FREE_ID] = P_SET;
  USE_SHARED_RO(rank)[FREE_ID] = 0;
  USE_SHARED_RO(ref_count)[FREE_ID] = (USE_SHARED_RO(heap_size) >> 1) + 1;

  USE_SHARED_RO(ref_count)[S_SET] = (USE_SHARED_RO(static_size) >> 1) + 1;
}

/* 
 * set up the reader and writer shadow space 
 */
static void init_shadow_space()
{
  void *p;
  size_t static_size, heap_size, page_size;
  size_t start_static_arena, start_heap_arena;
  void *start_static_accessor, *start_heap_accessor;
  char *shadow_space_fname = "/tmp/_Cilk_shadow_space";
  char *shadow_space_info_fname = "/tmp/_Cilk_shadow_space_info";
  int shadow_space_fd, shadow_space_info_fd;
  int junk = 42;
  size_t i;

  /* align the starting address of static and heap regions */
  page_size = Cilk_get_page_size();
  start_static_arena = ((size_t) STATIC_ARENA_START) / page_size * page_size;
  start_heap_arena = ((size_t) USE_SHARED_RO(heap_arena_start)) /
                     page_size * page_size;

  /* get the size of static and heap regions, round to page size */
  static_size = (size_t) STATIC_ARENA_END - start_static_arena;
  static_size = (size_t) ((static_size / page_size * page_size) +
                          (((static_size % page_size) == 0) ? 0:page_size));
  heap_size = (size_t) USE_SHARED_RO(heap_arena_end) - start_heap_arena;
  heap_size = (size_t) ((heap_size / page_size * page_size) +
                          (((heap_size % page_size) == 0) ? 0:page_size));

  /* calculate some sizes so that the address macros only require one 
     arithmetic operation */
  USE_SHARED_RO(static_size) = static_size;
  USE_SHARED_RO(heap_size) = heap_size;
  USE_SHARED_RO(both_size) = heap_size + static_size;     /* for SHADOW_INFO */
  USE_SHARED_RO(both_size2) = 2 * USE_SHARED_RO(both_size);/* for SHADOW_VAR */
  USE_SHARED_RO(both_plus_heap_size) = USE_SHARED_RO(both_size) + heap_size; 
                                                        /* for HEAP_ACCESSOR */
  
  USE_SHARED_RO(static_locks_offset) =
     start_heap_arena - USE_SHARED_RO(both_size) - start_static_arena;
  USE_SHARED_RO(static_accessor_offset) = 
    USE_SHARED_RO(static_locks_offset) - USE_SHARED_RO(both_size);
     
  /* get starting addresses of shadow spaces */
  start_static_accessor = STATIC_ACCESSOR(start_static_arena);
  start_heap_accessor = HEAP_ACCESSOR(start_heap_arena);

#if ND_VERBOSE
  printf("static 0x%p (align to 0x%x) --- 0x%p", 
	 STATIC_ARENA_START, start_static_arena, STATIC_ARENA_END);
  printf("\n    locks    0x%p", STATIC_LOCKS(start_static_arena));
  printf("\n    accessor 0x%p", start_static_accessor);
  printf("\n    accessor info 0x%p, accessor var 0x%p",
	 SHADOW_INFO(start_static_accessor), 
	 SHADOW_VAR(start_static_accessor));
  printf("\nheap 0x%p (align to 0x%x) --- 0x%p",
	 USE_SHARED_RO(heap_arena_start), start_heap_arena,
	 USE_SHARED_RO(heap_arena_end));
  printf("\n    locks    0x%p", HEAP_LOCKS(start_heap_arena));
  printf("\n    accessor 0x%p", start_heap_accessor);
  printf("\n    accessor info 0x%p, accessor var 0x%p",
	 SHADOW_INFO(start_heap_accessor), 
	 SHADOW_VAR(start_heap_accessor));
  printf("\n");
  printf("End of data segment 0x%p\n", &_edata);
#endif	       

  /* minimal check for overlap */
  CILK_CHECK(((int *)STATIC_LOCKS(start_static_arena) > (int *)&_edata),
	     ("Total space required too large\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(((int *) SHADOW_VAR(start_static_accessor) > &_edata),
	       ("Total space required too large\n"
		"Please run with smaller heap size (by -heap option)\n"));

  /* this represents the order we expect */
  /* SAV | HAV | SAI | HAI | SA | HA | SL | HL | H */
  CILK_ASSERT(SHADOW_VAR(start_static_accessor) < 
	      SHADOW_VAR(start_heap_accessor));
  CILK_ASSERT((void *)SHADOW_VAR(start_heap_accessor) <
	      (void *)SHADOW_INFO(start_static_accessor));
  CILK_ASSERT(SHADOW_INFO(start_static_accessor) < 
	      SHADOW_INFO(start_heap_accessor));
  CILK_ASSERT((void *)SHADOW_INFO(start_heap_accessor) < 
	      (void *)start_static_accessor);
  CILK_ASSERT(start_static_accessor < 
	      start_heap_accessor);
  CILK_ASSERT(start_heap_accessor < 
	      (void *)STATIC_LOCKS(start_static_arena));
  CILK_ASSERT((void *)STATIC_LOCKS(start_static_arena) < 
	      (void *)HEAP_LOCKS(start_heap_arena));
  /* CILK_ASSERT((void *)HEAP_LOCKS(start_heap_arena) < 
                 (void *)start_heap_arena); */

  /* create shadow space for ACCESSOR and LOCKS */
  shadow_space_fd = open(shadow_space_fname, O_RDWR | O_CREAT, 0600);
  CILK_CHECK(shadow_space_fd >= 0, ("Open %s error\n", shadow_space_fname));
  unlink(shadow_space_fname);
  if (USE_SHARED_RO(umbrella)) {
    CILK_CHECK(lseek(shadow_space_fd, 
		     USE_SHARED_RO(both_size2) - sizeof(int), SEEK_SET) ==
	       USE_SHARED_RO(both_size2) - sizeof(int),
	       ("error in init_shadow_space to lseek shadow_space_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    CILK_CHECK(write(shadow_space_fd, &junk, sizeof(int)) == sizeof(int),
	       ("error in init_shadow_space to write shadow_space_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    p = mmap((void *)start_static_accessor, USE_SHARED_RO(both_size2), 
	     PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, 
	     shadow_space_fd, 0);
  }
  else { /* ALL_SETS */
    CILK_CHECK(lseek(shadow_space_fd, 
		     USE_SHARED_RO(both_size) - sizeof(int), SEEK_SET) ==
	       USE_SHARED_RO(both_size) - sizeof(int),
	       ("error in init_shadow_space to lseek shadow_space_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    CILK_CHECK(write(shadow_space_fd, &junk, sizeof(int)) == sizeof(int),
	       ("error in init_shadow_space to write shadow_space_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    p = mmap((void *)STATIC_LOCKS(start_static_arena), 
	     USE_SHARED_RO(both_size), PROT_READ | PROT_WRITE, 
	     MAP_SHARED | MAP_FIXED, shadow_space_fd, 0);
  }
  if (p == MAP_FAILED) {
    perror("mmap");
    Cilk_die("mmap error in init_shadow_space for reader/writer shadow "
	     "spaces\n");
  }
  
  /* create shadow space for INFO and VAR */
  if (USE_SHARED_RO(umbrella)) {
    shadow_space_info_fd = open(shadow_space_info_fname, O_RDWR | O_CREAT, 
				0600);
    CILK_CHECK(shadow_space_info_fd >= 0, 
	       ("Open %s error\n", shadow_space_info_fname));
    unlink(shadow_space_info_fname);
    CILK_CHECK(lseek(shadow_space_info_fd, 
		     USE_SHARED_RO(both_size2) - sizeof(int), SEEK_SET) ==
	       USE_SHARED_RO(both_size2) - sizeof(int),
	       ("error in init_shadow_space to lseek shadow_space_info_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    CILK_CHECK(write(shadow_space_info_fd, &junk, sizeof(int)) == sizeof(int),
	       ("error in init_shadow_space to write shadow_space_info_fd!\n"
		"Please run with smaller heap size (by -heap option)\n"));
    p = mmap((void *)SHADOW_VAR(start_static_accessor), 
	     USE_SHARED_RO(both_size2), PROT_READ | PROT_WRITE, 
	     MAP_SHARED | MAP_FIXED, shadow_space_info_fd, 0);
    if (p == MAP_FAILED)
      Cilk_die("mmap error in init_shadow_space for reader/writer info shadow "
	       "spaces\n");
  }

  if (USE_SHARED_RO(umbrella)) {
    for (i = 0; i < (static_size >> 2); i++)
      *((DisjointSetMemberT *)start_static_accessor + i) = S_SET;
  }
  else { /* ALL_SETS */
    for (i = 0; i < (static_size >> 2); i++)
      *((DisjointSetMemberT *) STATIC_LOCKLISTS(start_static_arena) + i) = 
	S_SET;
  }
}

#if CILK_MANAGE_SHARED_SEGMENT
/*
 * If linker segments are used, a machine-specific version
 * must be provided in the architecture-specific file
 */

struct Cilk_shared_variables *Cilk_shared_vars;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_map_shared_segment(void)
{
     Cilk_shared_vars = 
	  Cilk_malloc_fixed(sizeof (struct Cilk_shared_variables));
     CILK_CHECK(Cilk_shared_vars,
		("Can't allocate shared variables!\n"));
}
#endif

#if CILK_MANAGE_PRIVATE_SEGMENT
/*
 * this is really useless, since if Cilk_private_vars
 * is really private, then all the other variables can
 * be private as well.  But at least the mechanism is
 * here, so that if Cilk_private_vars can be put in
 * some register or special location, everything works.
 */
CILK_DEFINE_REALLY_PRIVATE(struct Cilk_private_variables *, Cilk_private_vars);

static struct Cilk_private_variables dummy_private;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_map_private_segment(void)
{
     /* 
      * we use a dummy value because malloc() collects some statistics
      * that use private variables --- aaargh
      */
     Cilk_private_vars = &dummy_private;
     Self = -1;
     Cilk_private_vars = 
	  Cilk_malloc_fixed(sizeof (struct Cilk_private_variables));
     CILK_CHECK(Cilk_private_vars,
		("Can't allocate shared variables!\n"));
}

#endif

/*
 * initialize data strutures for Nondeterminator (independent from rts)
 * used in cmdline.c. malloc.c, internal-malloc.c
 */
void Cilk_global_init(Cilk_options *options)
{
  INIT_SHARED_RO(active_size, options->nproc);
  INIT_SHARED_RO(dump_core, options->dump_core);
  INIT_SHARED_RO(stackdepth, options->stackdepth);
  INIT_SHARED_RO(infofile, (FILE *)0);
  INIT_SHARED_RO(assertion_failed_msg,
		 "Assertion failed: %s line %d file %s\n"
		 "This is either a Cilk bug, or your program\n"
		 "has corrupted some Cilk internal data structure.\n");
  INIT_SHARED_RO(stack_overflow_msg,
		 "Cilk runtime system: Stack overflow.\n"
		 "Your program may have infinite recursion.\n");

  /*set up malloc() arena */
  Cilk_init_arena(options->heap);
 
  /* setup the address space */
  Cilk_map_shared_segment();
  Cilk_map_private_segment();
 
  /* set weird self value */
  Self = -1;
 
  /* architecture-specific module */
  Cilk_arch_specific_init();

  /* initialize the caches */
  INIT_SHARED_RO(caches, Cilk_malloc_fixed(sizeof(ClosureCache)));
  CILK_CHECK(USE_SHARED_RO(caches), ("failed to allocate caches\n"));
  
  USE_SHARED_RO(caches)[0].stack =
    Cilk_malloc_fixed(USE_SHARED_RO(stackdepth) * sizeof(StackFrame *));
  CILK_CHECK(USE_SHARED_RO(caches)[0].stack, ("failed to allocate stack\n"));
}

void nondeterminator_init(Cilk_options *options)
{
  Cilk_global_init(options);

  Cilk_internal_malloc_global_init();

  Self = 0;

  /* run initialization hooks for library functions */
  Cilk_run_hooks(Cilk_init_global_hooks);
  /* run initialization hooks for library functions */
  Cilk_run_hooks(Cilk_init_per_process_hooks);

  CILK_CHECK(PROC_ID_BIT + OPERATOR_BIT + TYPE_BIT <= 32,
	     ("Operator type for each shadow location need at least 6 bit"));

  init_shared_ro();
  init_shadow_space();
  init_disjoint_set();

  USE_PRIVATE(cache) = USE_SHARED_RO(caches);
  USE_PRIVATE(cache)->head = 
    (volatile StackFrame**)(USE_PRIVATE(cache)->stack);
  USE_PRIVATE(cache)->tail = 
    (volatile StackFrame**)(USE_PRIVATE(cache)->stack) + 1;

}

inline void Cilk_enter_state(int state) {}
inline void Cilk_exit_state(int state) {}

#ifndef CILK_LIBRARY_VERSION

#undef main

extern int cilk_main(int argc, char *argv[]);

/*
 * startup routine
 */
int main(int argc, char *argv[])
{
  Cilk_options options;

  if (Cilk_parse_command_line(&options, &argc, argv))
     return 0;

  nondeterminator_init(&options);

  return cilk_main(argc, argv);
}
#else /* CILK_LIBRARY_VERSION */
void Cilk_init(Cilk_options *options)
{
  nondeterminator_init(options);
}

void Cilk_finish(void)
{
  USE_SHARED_RO(nd_library_inited) = 0;
}

int Cilk_run(int (*foo)(void*), void *arg)
{
  USE_SHARED_RO(nd_library_inited) = 1;
  return foo(arg);
}
#endif

#endif /* CILK_ND */
