/*
 * $Id: arch-irix-threads.h,v 1.2 1998/06/01 14:32:43 fengmd Exp $
 */
/*
 * $Log: arch-irix-threads.h,v $
 * Revision 1.2  1998/06/01 14:32:43  fengmd
 * Merge changes from the main chunk.
 *
 * Revision 1.1.1.1  1998/03/19 04:43:35  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.1.1.2.4  1997/09/20 20:49:14  astark
 * Merged main branch Cilk_lock fixes
 *
 * Revision 1.1.1.1.2.3  1997/08/29 07:40:06  fengmd
 * Merge malloc/free checker, race report hashing, garbage collection.
 *
 * Revision 1.3  1997/08/25 21:20:04  randall
 * Added user-level locks.
 *
 * Revision 1.2  1997/03/01 17:42:27  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#define CACHE_LINE 128

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

#if (!CILK_IRIX_THREADS)
#error CILK_IRIX_THREADS not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/

#include <abi_mutex.h>

/* a lock variable */
typedef union {
     volatile int memory;
     abilock_t os;
} Cilk_mutex;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return sysconf(_SC_NPROC_ONLN);
}

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     {  }

/***********************************************************\
 * Various types of membars
\***********************************************************/

/*
 *  If your hardware is sequentially consistent, you can define CILK_SYNC 
 *  to be a NOP.
 *  If your hardware has relaxed consistency, use the CILK_SYNC macro below.
 */
#if 1
#define CILK_SYNC asm volatile ("sync":::"memory")
#else
#define CILK_SYNC asm volatile ("":::"memory")
#endif

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
  CILK_SYNC;
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
  CILK_SYNC;
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
  CILK_SYNC;
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/time.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef long long Cilk_time;

static inline Cilk_time Cilk_get_time(void)
{
     /*
      * not implemented!  We couldn't find a fast user-level high-resolution
      * timer. 
      */
     return 0;
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return t / 1e6;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     struct timeval tv;
     gettimeofday(&tv);
     return tv.tv_sec * (long long) 1000000 + tv.tv_usec;
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return t / 1e6;
}

/***********************************************************\
 * Fast user locks (spin locks)
\***********************************************************/

typedef volatile int Cilk_lockvar_basetype;
typedef Cilk_lockvar_basetype Cilk_lockvar[1];

static inline void Cilk_lock_init(Cilk_lockvar_basetype *v)
{
  *v = 0;
  Cilk_membar_StoreStore();
}

static inline void Cilk_lock(Cilk_lockvar_basetype *v)
{
  int tmp;
  
  do {
    do {
      asm volatile ("ll %0, (%1)":"=r" (tmp):"r"(v));
    } while (tmp != 0);
    
    tmp = 1;
    asm volatile ("sc %0, (%1)":"=r" (tmp):"r"(v), "0"(tmp));
  } while (tmp == 0);
  CILK_SYNC;
}

static inline void Cilk_unlock(Cilk_lockvar_basetype *v)
{
  Cilk_membar_StoreStore();
  *v = 0;
}
